/** \file CereConn.h
    \brief  CereConn class
    
    This file defines the main class to provide a nice C++ interface to CbSdk,
    encapsulating more multithreaded processing, and alo providing Python-friendly methods.
    
© Copyright 2018  - Jonas B. Zimmermann. All Rights Reserved.

    \author Jonas B. Zimmermann 
    \date 2018-08-23

**/

#ifndef CERECONN_H_BE1FC5EA
#define CERECONN_H_BE1FC5EA

#include <condition_variable>
#include <exception>
#include <memory>
#include <mutex>
#include <string>
#include "spdlog/spdlog.h"
#include "folly/ProducerConsumerQueue.h"
#include "cbsdk.h"
#include "CereRingBuffer.h"
#include "sre/SpikeRateEstimator.h"
#include "sbpe/SpikeBandPowerEstimator.h"

//! Main CereConn namespace
namespace cc
{

struct CereException : public std::exception
{
    const char * what () const throw ()
    {
        return "cbSDK function did not complete successfully";
    }
};


class CereConn
{
public:
    using SGroupBoolA = std::array<bool, constants::NSampleGroups>;
    enum class State
    {
        Invalid         = -1,
        Closed          = 0,
        Idle,
        Recording
    };
    enum class Instruction
    {
        Invalid         = -1,
        Open            = 0,
        Record,
        Idle,
        Close
    };

    enum class ReturnCode
    {
        
        Error           = -1,       //!< General error
        Success         = 0,        //!< returned upon success
        NotRecording,               //!< CereConn is not recording and can therefore not perform the requested operation
        NoValidTimestamp,           //!< CereConn could not get a valid timestamp from NSP.
        InvalidSpikeRateBuffer,     //!< CereConn does not have a valid spike rate buffer
        InvalidSpikeRateEstimator,  //!< CereConn does not have a valid spike rate estimator
        CbSdkError,                 //!< CbSdk returned an error
        InvalidSpikeBandPowerBuffer,//!< CereConn does not have a valid spike band power buffer
        InvalidSpikeBandPowerEstimator,  //!< CereConn does not have a valid spike band power estimator
    };

    /**\brief Construct a CereConn instance
    \param instance instrument instance of the cbSdk library. Allows multiple instruments in parallel
    \param withSRE if `true`, Spike Rate Estimator will be automatically set up and started
    \param withSBPE if `true`, Spike Band Power Estimator will be automatically set up and started
    **/    
    CereConn (uint32_t instance = 0, bool withSRE = false, bool withSBPE = false);
    virtual ~CereConn ();


    // convenience methods for Python
    
    /** \brief Open connection to NSP
    
        This method is asynchronous. The worker thread calls `cbOpen()`.
    
        \sa ccRecord(), ccIdle(), ccClose(), getState()
    **/
    void ccOpen();
    /** \brief Start processing of incoming data
    
        This method is asynchronous. The worker thread calls `cbStartRecording()`.
    
        Callbacks for spike and continuous data will be registered
    
        \sa ccOpen(), ccIdle(), ccClose(), getState()
    **/
    void ccRecord();
    
    /** \brief Stop processing of incoming data
    
        This method is asynchronous. The worker thread calls `cbStopRecording()`.
    
        Callbacks for spike and continuous data will be unregistered
    
        \sa ccOpen(), ccRecord(), ccClose(), getState()
    **/
    void ccIdle();
    
    /** \brief Close connection to NSP
        
        This method is asynchronous. The worker thread calls `cbClose()`.
        
        \sa ccOpen(), ccRecord(), ccIdle(), getState()
    **/
    void ccClose();

    /** \brief Return current state of CereConn
    
        \return State the current state
        \sa ccOpen(), ccRecord(), ccIdle(), ccClose()
    **/
    State getState() const;
    
    /** \brief read sample group data from internal buffers
        
        Copies sample group (continuous) data from internal ring buffers
        to `sgl` and flushes that buffer. The NSP's time at the end of that
        operation is written to `cbTime`. This timestamp does not have to coincide
        with the last data sample in the buffer.
        
        \param [out] sgl SampleGroupList object to which samples will be written
        \param [out] cbTime timestamp of last package received at the end of the copying operation.
        \return ReturnCode success or failure
        \sa setEnabledSampleGroups(), setSampleGroup()
    **/
    ReturnCode getSampleGroupData(SampleGroupList &sgl, uint32_t &cbTime);

    /** \brief send comment to NSP, to be included in data stream
    
        Constructs a comment package and sends it to the NSP.
        The first 127 characters of `comment` are taken, together with a 32bit colour
        value `rgba`, for <red, green, blue, alpha>.
    
        \param [in] comment comment string
        \param [in] rgba RGBA value
        \param [in] charset Character set (0 - ANSI, 1 - UTF16, 255 - NeuroMotive ANSI) (from `cbhwlib.h`)
        \return ReturnCode success or failure
        \sa getCommentData()
    **/
    ReturnCode sendComment(const std::string comment, const uint32_t rgba=0, const uint8_t charset=0);

    /** \brief read comment data internal buffers
    
        Copies comment data from internal ring buffers
        to `clist` and flushes that buffer. The NSP's time at the end of that
        operation is written to `cbTime`. This timestamp does not have to coincide
        with the last comment in the buffer.
    
        \param [out] clist `CommentList` object to which samples will be written
        \param [out] cbTime timestamp of last package received at the end of the copying operation.
        \return ReturnCode success or failure
        \sa sendComment(), getDigInEvData()
    **/
    ReturnCode getCommentData(CommentList &clist, uint32_t &cbTime);

    /** \brief read comment data internal buffers

        Copies comment data from internal ring buffers
        to `clist` and flushes that buffer. The NSP's time at the end of that
        operation is written to `cbTime`. This timestamp does not have to coincide
        with the last comment in the buffer.

        \param [out] clist `CommentList` object to which samples will be written
        \param [out] cbTime timestamp of last package received at the end of the copying operation.
        \param [out] overflowCount stores the number of overflow items in comment buffer since last call
        \return ReturnCode success or failure
        \sa sendComment(), getDigInEvData()
    **/
    ReturnCode getCommentData(CommentList &clist, uint32_t &cbTime, size_t &overflowCount);

    /** \brief read digitalInEvent data internal buffers

        Copies digitalInEvent data from internal ring buffers
        to `dievlist` and flushes that buffer. The NSP's time at the end of that
        operation is written to `cbTime`. This timestamp does not have to coincide
        with the last digitalInEvent in the buffer.

        \param [out] dievlist `DigInEvList` object to which samples will be written
        \param [out] cbTime timestamp of last package received at the end of the copying operation.
        \return ReturnCode success or failure
        \sa getCommentData()
    **/
    ReturnCode getDigInEvData(DigInEvList &dievlist, uint32_t &cbTime);

    /** \brief read digitalInEvent data internal buffers

        Copies digitalInEvent data from internal ring buffers
        to `dievlist` and flushes that buffer. The NSP's time at the end of that
        operation is written to `cbTime`. This timestamp does not have to coincide
        with the last digitalInEvent in the buffer.

        \param [out] dievlist `DigInEvList` object to which samples will be written
        \param [out] cbTime timestamp of last package received at the end of the copying operation.
        \param [out] overflowCount stores the number of overflow items in digital event buffer since last call
        \return ReturnCode success or failure
        \sa getCommentData()
    **/
    ReturnCode getDigInEvData(DigInEvList &dievlist, uint32_t &cbTime, size_t &overflowCount);

    /** \brief set which sample groups are enabled, i.e. will be buffered for later consumption
        
        Sample groups not enabled here will not be processed at all.
        
        \param [in] sgroups_enabled `SGroupBoolA` a reference to an array of bools of the length of sampling groups, where `true` indicates that this sampling group will be processed.
        \sa setSampleGroup(), getSampleGroupData()
    **/
    void setEnabledSampleGroups(const SGroupBoolA &sgroups_enabled);
    
    /** \brief set for each particular sampling group whether it will be processed or not
                
        \param [in] group number of sampling group (usually 1 – 6)
        \param [in] state `true`if sample group should be processed, `false` otherwise
        \sa getSampleGroup(), setEnabledSampleGroups(), getSampleGroupData()
    **/
    void setSampleGroup(const size_t group, const bool state=true);

    /** \brief Get recording state of sampling group
            
        \param [in] group number of sampling group (usually 1 – 6)
        \return `true`if sample group is being processed, `false` otherwise
        \sa setSampleGroup(), setEnabledSampleGroups(), getSampleGroupData()
    **/
    bool getSampleGroup(const size_t group) const;
    
    /** \name Common Average Referencing for sample groups
        \attention setCARChannels() and disableCAR() are applied both to the continuous sample
        buffer and to the spike band estimator (if enabled). Also note that SBPE will not be
        updated if it is set up after CAR.
    **/
    /**@{**/
    /** \brief Sets channnels for Common Average Reference for sample group
        \param group sample group index (one-based!)
        \param carChannelList list of channel IDs to be used for CAR
        \return `true` if CAR successfully set up, `false` otherwise
        \sa SampleGroupRingBufferCollectionT::setCARChannels(size_t group, ChannelNumList_t &&carChannelList), disableCAR()
    **/
    bool setCARChannels(const size_t group, const ChannelNumList_t &carChannelList);
    
    /** \brief Sets channnels for Common Average Reference for sample group
        \param group sample group index (one-based!)
        \param carChannelList list of channel IDs to be used for CAR
        \return `true` if CAR successfully set up, `false` otherwise
        \sa SampleGroupRingBufferCollectionT::setCARChannels(size_t group, const ChannelNumList_t &carChannelList), disableCAR(),
    **/
    bool setCARChannels(const size_t group, ChannelNumList_t &&carChannelList);
    
    /** \brief Disables CAR for sample group
        \param group sample group index (one-based!)
        \sa setCARChannels(size_t group, ChannelNumList_t &&carChannelList),
        setCARChannels(size_t group, const ChannelNumList_t &carChannelList),
        SampleGroupRingBufferCollectionT::disableCAR()
    **/
    void disableCAR(const size_t group);
    /**@}**/
    
    /** \brief Return a sample group's sampling period
        \param group 1-based channel ID
        \return ticks per sample
        \sa getSampleGroupFrequency()
    **/
    unsigned int getSampleGroupPeriod(const size_t group) const {return constants::SampleGroupPeriods[group - 1];}
    
    /** \brief Return a sample group's sampling frequency
        \param group 1-based channel ID
        \return samples per second
        \sa getSampleGroupPeriod()
    **/
    unsigned int getSampleGroupFrequency(const size_t group) const {return constants::SampleGroups[group - 1];}

    /** \brief returns current time of Cerebus system
        
        \param [out] cbTime holds timestamp after successful execution
        \return ReturnCode success or failure
    **/
    ReturnCode getCbTime(uint32_t &cbTime);

    /** \brief print overflow count of continuous sample buffers
        
        The overflow counters will be printed to the logs.
        
    **/
    void printBufferState() const;

    /** \name Spike Rate Estimator control **/
    /**@{**/
    /** \brief Starts spike rate estimator processing thread
                
        \sa spikeRateEstimatorStop()
    **/
    void spikeRateEstimatorStart();

    /** \brief Stops spike rate estimator processing thread
            
        \sa spikeRateEstimatorStart()
    **/
    void spikeRateEstimatorStop();
    
    /** \brief Set a list of channel-unit pairs for which to calculate spike rates
        \sa cc::sre::SpikeRateEstimator::setChannelUnitList()
    **/
    void setSpikeRateEstimatorChannelUnitList(const sre::SpikeRateEstimator::ch_un_list_type & ch_un_list);

    /** \brief Enable spike rate calculation for all channel-unit combinations
        \sa cc::sre::SpikeRateEstimator::fillChannelUnitList()
    **/
    void fillSpikeRateEstimatorChannelUnitList(const sre::SpikeRateEstimator::size_type n_channels, const sre::SpikeRateEstimator::size_type n_units);
    
    /** \brief Get the current mapping of channel-units to spike rates
        \sa cc::sre::SpikeRateEstimator::getChannelUnitMapping()
    **/
    std::pair<sre::SpikeRateEstimator::ch_un_list_type, sre::SpikeRateEstimator::ch_un_map_type> getSpikeRateEstimatorChannelUnitMapping() const;
    
    /** \brief Get a list of current channel-unit pairs for which spike rates are computed
        \sa cc::sre::SpikeRateEstimator::getChannelUnitList()
    **/
    sre::SpikeRateEstimator::ch_un_list_type getSpikeRateEstimatorChannelUnitList() const;
    
    /** \brief Set interval for repeated calculation of spike rates
        \param new_interval new interval in sre::SpikeRateEstimator::rate_interval_type
        \sa cc::sre::SpikeRateEstimator::setSpikeRateLoopInterval()
    **/
    void setSpikeRateEstimatorLoopInterval(const sre::SpikeRateEstimator::rate_interval_type & new_interval);
    
    /** \brief Get interval for repeated calculation of spike rates
        \sa cc::sre::SpikeRateEstimator::getSpikeRateLoopInterval()
    **/
    sre::SpikeRateEstimator::rate_interval_type getSpikeRateEstimatorLoopInterval() const;
    
    /** \brief Set interval for repeated calculation of spike rates
        \param new_interval new interval in milliseconds
        \sa cc::sre::SpikeRateEstimator::setSpikeRateLoopInterval()
    **/
    void setSpikeRateEstimatorLoopIntervalMS(const unsigned int & new_interval);
    
    /** \brief Get interval for repeated calculation of spike rates
        \return interval in milliseconds
        \sa cc::sre::SpikeRateEstimator::getSpikeRateLoopInterval()
    **/
    unsigned int getSpikeRateEstimatorLoopIntervalMS() const;
    
    /** \brief Bind a function to perform spike rate estimation
    
        Any callable conforming to sre::SpikeRateEstimator::rate_callback_type can
        be passed as a callback.
    
        \param[in] rcb callback calculating the spike rate, given binned spike counts.
        \sa spikeRateEstimatorId(), spikeRateEstimatorBoxcar(), spikeRateEstimatorExp(),
            cc::sre::SpikeRateEstimator::bindCallback(), sre::SpikeRateEstimator::rate_callback_type
    **/
    void bindSpikeRateEstimatorCallback(const sre::SpikeRateEstimator::rate_callback_type& rcb);
    
    /** \brief Use identity spike rate estimator (i.e. latest bin's count)

        \sa cc::sre::r_estimate_id(), spikeRateEstimatorBoxcar(), spikeRateEstimatorExp(),
            bindSpikeRateEstimatorCallback()
    **/
    void spikeRateEstimatorId();
    
    /** \brief Use boxcar spike rate estimator

        \sa cc::sre::r_estimate_with_boxcar(), spikeRateEstimatorId(), spikeRateEstimatorExp(),
            bindSpikeRateEstimatorCallback()
    **/
    void spikeRateEstimatorBoxcar(const size_t length);
    
    /** \brief Use exponential decay spike rate estimator

        \sa cc::sre::r_estimate_with_exp_decay(), spikeRateEstimatorBoxcar(), spikeRateEstimatorId(),
            bindSpikeRateEstimatorCallback()
    **/
    void spikeRateEstimatorExp(const double decay_factor, const size_t max_length = 5);
    
    /** \brief Get buffered Spike Rate data
    
    When Spike Rate Estimation is enabled, this function should be called
    periodically to retrieve the calculated data.
    
    \param[out] srl variable to store spike rate data
    \param[out] cbTime stores NSP timestamp at call time
    **/
    CereConn::ReturnCode getSpikeRateData(sre::SpikeRateEstimator::rate_list_type &srl, uint32_t &cbTime);

    /** \brief Get buffered Spike Rate data

        When Spike Rate Estimation is enabled, this function should be called
        periodically to retrieve the calculated data.

        \param[out] srl variable to store spike rate data
        \param[out] cbTime stores NSP timestamp at call time
        \param [out] overflowCount stores the number of overflow items in spike rate buffer since last call
    **/
    CereConn::ReturnCode getSpikeRateData(sre::SpikeRateEstimator::rate_list_type &srl, uint32_t &cbTime, size_t &overflowCount);
    /**@}**/

    /** \name Spike Band Power Estimator control **/
    /**@{**/
    /** \brief Starts spike band power estimator processing thread
            
        \sa spikeBandPowerEstimatorStop(), cc::sbpe::SpikeBandPowerEstimator::signalStart()
    **/
    void spikeBandPowerEstimatorStart();
    
    /** \brief Stops spike band power estimator processing thread
        
        \sa spikeBandPowerEstimatorStart(), cc::sbpe::SpikeBandPowerEstimator::signalStop()
    **/
    void spikeBandPowerEstimatorStop();
    
    /** \brief Select channels for spike band power estimation
        
        NB this is usually from sample group 6 (i.e. unfiltered raw data)
    
        \param ch_list a list of (0-based) channel ids
        \sa fillSpikeBandPowerEstimatorChannelList(), getSpikeBandPowerEstimatorChannelList(), getSpikeBandPowerEstimatorChannelMapping(), cc::sbpe::SpikeBandPowerEstimator::setChannelList()
    **/
    void setSpikeBandPowerEstimatorChannelList(const sbpe::SpikeBandPowerEstimator::ch_list_type & ch_list);
    
    /** \brief Select all channels for spike band power estimation
        
        All channels in (0..n_ch-1) will be used.
        
        \param n_channels an integer channel id
        \sa setSpikeBandPowerEstimatorChannelList(), getSpikeBandPowerEstimatorChannelMapping()
            getSpikeBandPowerEstimatorChannelList(), cc::sbpe::SpikeBandPowerEstimator::fillChannelList()
    **/
    void fillSpikeBandPowerEstimatorChannelList(const sbpe::SpikeBandPowerEstimator::size_type n_channels);
    
    /** \brief Get channel to spike band power value map
        
        \return a std::pair with list of channels in their order of occurrence and
            a map of channels to index in the spike band power data arrays
        \sa getSpikeBandPowerEstimatorChannelList(), 
            setSpikeBandPowerEstimatorChannelList(), fillSpikeBandPowerEstimatorChannelList(),
            cc::sbpe::SpikeBandPowerEstimator::getChannelMapping()
    **/
    std::pair<sbpe::SpikeBandPowerEstimator::ch_list_type, sbpe::SpikeBandPowerEstimator::ch_map_type> getSpikeBandPowerEstimatorChannelMapping() const;
    
    /** \brief Get channel list for which spike band power is calculated
    
        \return list of channels in their order of occurrence
        \sa getSpikeBandPowerEstimatorChannelMapping(), 
            setSpikeBandPowerEstimatorChannelList(), fillSpikeBandPowerEstimatorChannelList(),
            cc::sbpe::SpikeBandPowerEstimator::getChannelList()
    **/
    sbpe::SpikeBandPowerEstimator::ch_list_type getSpikeBandPowerEstimatorChannelList() const;
    
    /** \brief Set the time interval at which spike band power is calculated
        
        \param new_interval interval at which spike band power is calculated
        \sa cc::sbpe::setLoopInterval(), getSpikeBandPowerEstimatorLoopInterval(),
            setSpikeBandPowerEstimatorLoopIntervalMS(), getSpikeBandPowerEstimatorLoopIntervalMS()
    **/
    void setSpikeBandPowerEstimatorLoopInterval(const sbpe::SpikeBandPowerEstimator::sbp_interval_type & new_interval);

    /** \brief Get the time interval at which spike band power is calculated
    
        \return interval at which spike band power is calculated
        \sa cc::sbpe::SpikeBandPowerEstimator::getLoopInterval(), setSpikeBandPowerEstimatorLoopInterval(),
            setSpikeBandPowerEstimatorLoopIntervalMS(), getSpikeBandPowerEstimatorLoopIntervalMS()
    **/
    sbpe::SpikeBandPowerEstimator::sbp_interval_type getSpikeBandPowerEstimatorLoopInterval() const;
    
    /** \brief Set the time interval at which spike band power is calculated
    
        \param new_interval interval at which spike band power is calculated, in milliseconds
        \sa cc::sbpe::SpikeBandPowerEstimator::setLoopInterval(), getSpikeBandPowerEstimatorLoopIntervalMS(),
            setSpikeBandPowerEstimatorLoopInterval(), getSpikeBandPowerEstimatorLoopInterval()
    **/
    void setSpikeBandPowerEstimatorLoopIntervalMS(const unsigned int & new_interval);

    /** \brief Get the time interval at which spike band power is calculated

        \return interval at which spike band power is calculated, in milliseconds
        \sa cc::sbpe::SpikeBandPowerEstimator::getLoopInterval(), setSpikeBandPowerEstimatorLoopIntervalMS(),
            setSpikeBandPowerEstimatorLoopInterval(), getSpikeBandPowerEstimatorLoopInterval()
    **/
    unsigned int getSpikeBandPowerEstimatorLoopIntervalMS() const;
    
    /** \brief Set which sample group packets will be accepted for spike band power

        The default is *6*, i.e. raw data samples
        
        \param  group number of the accepted sample group
        \sa getSpikeBandPowerEstimatorAcceptedSampleGroup(), cc::sbpe::SpikeBandPowerEstimator::setAcceptedSampleGroup()
    **/
    void setSpikeBandPowerEstimatorAcceptedSampleGroup(const size_t group);
    
    /** \brief Get the sample group number for spike band power processing

        \return the currently set accepted sample group.
        \sa setSpikeBandPowerEstimatorAcceptedSampleGroup(), cc::sbpe::SpikeBandPowerEstimator::setAcceptedSampleGroup()
    **/
    size_t getSpikeBandPowerEstimatorAcceptedSampleGroup() const;
    
    /**\brief Set number of samples the RMS part of the IIR/RMS filter integrates over (for spike band power)
        
        The number of samples governs the target rate of the SpikeBandPowerEstimator.
        For example, if the sample rate of the input sample group is 30'000 samples/s
        and SpikeBandPowerEstimator generates an RMS value every 600 samples (as is the default),
        then the RMS will be calculated for every 20 milliseconds of data.
        
        \param n_samples number of samples
        \sa getSpikeBandPowerEstimatorIIRFilterNSamples()
    **/
    void setSpikeBandPowerEstimatorIIRFilterNSamples(const size_t n_samples);
    
    /**\brief Get number of samples the RMS part of the IIR/RMS filter integrates over (for spike band power)
        \return number of samples
        \sa setSpikeBandPowerEstimatorIIRFilterNSamples()
    **/
    size_t getSpikeBandPowerEstimatorIIRFilterNSamples() const;

    /**\brief Set number of bins the RMS part of the IIR/RMS filter averages over (for spike band power)
    
        This parameter determines how many bins are averaged before the square root is taken.
        Together with the number of samples set by setSpikeBandPowerEstimatorIIRFilterNSamples()
        this determines the temporal features of the resulting RMS signal.
        By default, this is 1, which means that only the RMS of the latest bin is returned.
        
        However, at sampling rate of 30'000 kS/s and if n_samples is 600 (as by default),
        setting this parameter to 50 would mean that each time the RMS value is updated (every
        20 ms), the whole history of 1s is used to calculate the RMS. This will produce
        a smoother signal at the expense of delays.
    
        \param n_bins number of bins to average for RMS
        \sa getSpikeBandPowerEstimatorAvgNBins()
    **/
    void setSpikeBandPowerEstimatorAvgNBins(const size_t n_bins);

    /**\brief Get number of bins the RMS part of the IIR/RMS filter averages over (for spike band power)
        \return number of bins
        \sa setSpikeBandPowerEstimatorAvgNBins()
    **/
    size_t getSpikeBandPowerEstimatorAvgNBins() const;

    
    /** \brief Set IIR filter coefficients for spike band power estimation
        \param[in] b input coefficients
        \param[in] a feedback coefficients
        \sa getSpikeBandPowerEstimatorIIRFilterCoeffs(), cc::sbpe::SpikeBandPowerEstimator::setIIRFilterNSamples()
    **/
    void setSpikeBandPowerEstimatorIIRFilterCoeffs(const sbpe::SpikeBandPowerEstimator::iir_filt_coeff_type &b, const sbpe::SpikeBandPowerEstimator::iir_filt_coeff_type &a);
    
    /** \brief Get IIR filter coefficients for spike band power estimation
        \return a std::pair of coefficient containers, the first item representing the
            input coefficients, the second item representing the feedback coefficients
        \sa setSpikeBandPowerEstimatorIIRFilterCoeffs(), cc::sbpe::SpikeBandPowerEstimator::getIIRFilterCoeffs()
    **/
    std::pair<sbpe::SpikeBandPowerEstimator::iir_filt_coeff_type, sbpe::SpikeBandPowerEstimator::iir_filt_coeff_type> getSpikeBandPowerEstimatorIIRFilterCoeffs() const;
    
    /** \brief Get buffered Spike Band Power data
    
    When Spike Band Power Estimation is enabled, this function should be called
    periodically to retrieve the calculated data.
    
    \param[out] sbpl variable to store spike band power data
    \param[out] cbTime stores NSP timestamp at call time
    **/
    CereConn::ReturnCode getSpikeBandPowerData(sbpe::SpikeBandPowerEstimator::sbp_list_type &sbpl, uint32_t &cbTime);

    /** \brief Get buffered Spike Band Power data

    When Spike Band Power Estimation is enabled, this function should be called
    periodically to retrieve the calculated data.

    \param[out] sbpl variable to store spike band power data
    \param[out] cbTime stores NSP timestamp at call time
    \param[out] overflowCount stores the number of overflow items in spike band power buffer since last call
    **/
    CereConn::ReturnCode getSpikeBandPowerData(sbpe::SpikeBandPowerEstimator::sbp_list_type &sbpl, uint32_t &cbTime, size_t &overflowCount);

    /**@}**/

    /** \name cbSDK callback functions
    
        These members are called (via their static cousins) by cbSDK to pass on data.
    **/
    /**@{**/
    void cbSdkCallbackContinuous(uint32_t nInstance, const cbSdkPktType type, const void* pPkt);
    void cbSdkCallbackSpike(uint32_t nInstance, const cbSdkPktType type, const void* pPkt);
    void cbSdkCallbackComment(uint32_t nInstance, const cbSdkPktType type, const void* pPkt);
    void cbSdkCallbackDigInEv(uint32_t nInstance, const cbSdkPktType type, const void* pPkt);
    void cbSdkCallbackGroupInfo(uint32_t nInstance, const cbSdkPktType type, const void* pPkt);
    /**@}**/

private:
    void spikeRateEstimatorSetUp();
    void spikeBandPowerEstimatorSetUp();
    
    static void cbSdkCallbackContinuous(uint32_t nInstance, const cbSdkPktType type, const void* pPkt, void* pCallbackData);
    static void cbSdkCallbackSpike(uint32_t nInstance, const cbSdkPktType type, const void* pPkt, void* pCallbackData);
    static void cbSdkCallbackComment(uint32_t nInstance, const cbSdkPktType type, const void* pPkt, void* pCallbackData);
    static void cbSdkCallbackDigInEv(uint32_t nInstance, const cbSdkPktType type, const void* pPkt, void* pCallbackData);
    static void cbSdkCallbackGroupInfo(uint32_t nInstance, const cbSdkPktType type, const void* pPkt, void* pCallbackData);
    
    /** \brief Get channel list for sample group from cbSDK
    **/
    void updateGroupChannelList(const uint32_t group);
    /** \brief Update all channel lists from cbSDK
    **/
    void updateGroupChannelLists();
    
    
    folly::ProducerConsumerQueue<Instruction> m_msgQ;
    cbSdkConnectionType m_conType;
    uint32_t m_instance;
    State m_state;
    std::shared_ptr<spdlog::logger> m_console;
    std::shared_ptr<spdlog::logger> m_acq;

    std::thread m_thread;
    void threadFun();
    void startThread();

    void sendMessage(Instruction instruction);

    void cbOpen();
    void cbStartRecording();

    void cbStopRecording();
    void cbClose();

    std::unique_ptr<SampleGroupRingBufferCollection> m_pRb;
    std::unique_ptr<SpikeRingBuffer> m_pSRb;
    std::unique_ptr<CommentRingBuffer> m_pCRb;
    std::unique_ptr<DigInEvRingBuffer> m_pDIRb;
    
    // store rates in ring buffer
    std::shared_ptr<sre::SpikeRateEstimator::rate_ring_buffer_type> mp_rateRB;
    std::unique_ptr<sre::SpikeRateEstimator> mp_sre;

    std::shared_ptr<sbpe::SpikeBandPowerEstimator::sbp_ring_buffer_type> mp_sbpRB;
    std::unique_ptr<sbpe::SpikeBandPowerEstimator> mp_sbpe;

    std::atomic<bool> m_keep_going;
    std::mutex m_msg_lock;
    std::condition_variable m_msg_queue_not_empty;

    // store information about which sample groups to buffer
    SGroupBoolA m_sample_groups_enabled;
    
    // keep track of channels used for continuous recordings
    std::array<bool,constants::NSampleGroups>               m_sgroup_changed;
    std::array<ChannelNumList_t, constants::NSampleGroups>  m_sgroup_chlist;
    
    std::atomic<bool> m_with_SRE;
    std::atomic<bool> m_with_SBPE;
};

}

#endif /* end of include guard: CERECONN_H_BE1FC5EA */
