/*! \file cbsdk.h
    \brief  Mock up library to simulate cbSdk
    
*/
#ifndef CBSDK_H_347FBB8B
#define CBSDK_H_347FBB8B
#include <cstddef>
#include "cbhwlib.h"

#ifdef STATIC_CBSDK_LINK
#undef CBSDK_EXPORTS
#endif

#ifdef WIN32
// Windows shared library
#ifdef CBSDK_EXPORTS
    #define CBSDKAPI __declspec(dllexport)
#elif ! defined(STATIC_CBSDK_LINK)
    #define CBSDKAPI __declspec(dllimport)
#else
    #define CBSDKAPI
#endif
#else
// Non Windows shared library
#ifdef CBSDK_EXPORTS
    #define CBSDKAPI __attribute__ ((visibility ("default")))
#else
    #define CBSDKAPI
#endif
#endif

/// cbSdk return values
typedef enum _cbSdkResult
{
    CBSDKRESULT_WARNCONVERT            =     3, ///< If file conversion is needed
    CBSDKRESULT_WARNCLOSED             =     2, ///< Library is already closed
    CBSDKRESULT_WARNOPEN               =     1, ///< Library is already opened
    CBSDKRESULT_SUCCESS                =     0, ///< Successful operation
    CBSDKRESULT_NOTIMPLEMENTED         =    -1, ///< Not implemented
    CBSDKRESULT_UNKNOWN                =    -2, ///< Unknown error
    CBSDKRESULT_INVALIDPARAM           =    -3, ///< Invalid parameter
    CBSDKRESULT_CLOSED                 =    -4, ///< Interface is closed cannot do this operation
    CBSDKRESULT_OPEN                   =    -5, ///< Interface is open cannot do this operation
    CBSDKRESULT_NULLPTR                =    -6, ///< Null pointer
    CBSDKRESULT_ERROPENCENTRAL         =    -7, ///< Unable to open Central interface
    CBSDKRESULT_ERROPENUDP             =    -8, ///< Unable to open UDP interface (might happen if default)
    CBSDKRESULT_ERROPENUDPPORT         =    -9, ///< Unable to open UDP port
    CBSDKRESULT_ERRMEMORYTRIAL         =   -10, ///< Unable to allocate RAM for trial cache data
    CBSDKRESULT_ERROPENUDPTHREAD       =   -11, ///< Unable to open UDP timer thread
    CBSDKRESULT_ERROPENCENTRALTHREAD   =   -12, ///< Unable to open Central communication thread
    CBSDKRESULT_INVALIDCHANNEL         =   -13, ///< Invalid channel number
    CBSDKRESULT_INVALIDCOMMENT         =   -14, ///< Comment too long or invalid
    CBSDKRESULT_INVALIDFILENAME        =   -15, ///< Filename too long or invalid
    CBSDKRESULT_INVALIDCALLBACKTYPE    =   -16, ///< Invalid callback type
    CBSDKRESULT_CALLBACKREGFAILED      =   -17, ///< Callback register/unregister failed
    CBSDKRESULT_ERRCONFIG              =   -18, ///< Trying to run an unconfigured method
    CBSDKRESULT_INVALIDTRACKABLE       =   -19, ///< Invalid trackable id, or trackable not present
    CBSDKRESULT_INVALIDVIDEOSRC        =   -20, ///< Invalid video source id, or video source not present
    CBSDKRESULT_ERROPENFILE            =   -21, ///< Cannot open file
    CBSDKRESULT_ERRFORMATFILE          =   -22, ///< Wrong file format
    CBSDKRESULT_OPTERRUDP              =   -23, ///< Socket option error (possibly permission issue)
    CBSDKRESULT_MEMERRUDP              =   -24, ///< Socket memory assignment error
    CBSDKRESULT_INVALIDINST            =   -25, ///< Invalid range or instrument address
    CBSDKRESULT_ERRMEMORY              =   -26, ///< library memory allocation error
    CBSDKRESULT_ERRINIT                =   -27, ///< Library initialization error
    CBSDKRESULT_TIMEOUT                =   -28, ///< Conection timeout error
    CBSDKRESULT_BUSY                   =   -29, ///< Resource is busy
    CBSDKRESULT_ERROFFLINE             =   -30, ///< Instrument is offline
    CBSDKRESULT_INSTOUTDATED           =   -31, ///<  The instrument runs an outdated protocol version
    CBSDKRESULT_LIBOUTDATED            =   -32, ///<  The library is outdated
} cbSdkResult;

/// cbSdk Connection Type (Central, UDP, other)
typedef enum _cbSdkConnectionType
{
    CBSDKCONNECTION_DEFAULT = 0, ///< Try Central then UDP
    CBSDKCONNECTION_CENTRAL,     ///< Use Central
    CBSDKCONNECTION_UDP,         ///< Use UDP
    CBSDKCONNECTION_CLOSED,      ///< Closed
    CBSDKCONNECTION_COUNT ///< Allways the last value (Unknown)
} cbSdkConnectionType;

/// Connection information
typedef struct _cbSdkConnection
{
    _cbSdkConnection()
    {
        nInPort = cbNET_UDP_PORT_BCAST;
        nOutPort = cbNET_UDP_PORT_CNT;
        #ifdef __APPLE__
			nRecBufSize = (6 * 1024 * 1024); // Despite setting kern.ipc.maxsockbuf=8388608, 8MB is still too much.
        #else
			nRecBufSize = (8 * 1024 * 1024); // 8MB default needed for best performance
        #endif
        szInIP = "";
        szOutIP = "";
    }
    int nInPort;  ///< Client port number
    int nOutPort; ///< Instrument port number
    int nRecBufSize; ///< Receive buffer size (0 to ignore altogether)
    LPCSTR szInIP;  ///< Client IPv4 address
    LPCSTR szOutIP; ///< Instrument IPv4 address
} cbSdkConnection;

/// Packet type information
typedef enum _cbSdkPktType
{
    cbSdkPkt_PACKETLOST = 0, ///< will be received only by the first registered callback
                             ///< data points to cbSdkPktLostEvent
    cbSdkPkt_INSTINFO,       ///< data points to cbSdkInstInfo
    cbSdkPkt_SPIKE,          ///< data points to cbPKT_SPK
    cbSdkPkt_DIGITAL,        ///< data points to cbPKT_DINP
    cbSdkPkt_SERIAL,         ///< data points to cbPKT_DINP
    cbSdkPkt_CONTINUOUS,     ///< data points to cbPKT_GROUP
    cbSdkPkt_TRACKING,       ///< data points to cbPKT_VIDEOTRACK
    cbSdkPkt_COMMENT,        ///< data points to cbPKT_COMMENT
    cbSdkPkt_GROUPINFO,      ///< data points to cbPKT_GROUPINFO
    cbSdkPkt_CHANINFO,       ///< data points to cbPKT_CHANINFO
    cbSdkPkt_FILECFG,        ///< data points to cbPKT_FILECFG
    cbSdkPkt_POLL,           ///< data points to cbPKT_POLL
    cbSdkPkt_SYNCH,          ///< data points to cbPKT_VIDEOSYNCH
    cbSdkPkt_NM,             ///< data points to cbPKT_NM
    cbSdkPkt_CCF,            ///< data points to cbSdkCCFEvent
    cbSdkPkt_IMPEDANCE,      ///< data points to cbPKT_IMPEDANCE
    cbSdkPkt_SYSHEARTBEAT,   ///< data points to cbPKT_SYSHEARTBEAT
    cbSdkPkt_LOG,		     ///< data points to cbPKT_LOG
    cbSdkPkt_COUNT ///< Always the last value
} cbSdkPktType;

/// Type of events to monitor
typedef enum _cbSdkCallbackType
{
    CBSDKCALLBACK_ALL = 0,      ///< Monitor all events
    CBSDKCALLBACK_INSTINFO,     ///< Monitor instrument connection information
    CBSDKCALLBACK_SPIKE,        ///< Monitor spike events
    CBSDKCALLBACK_DIGITAL,      ///< Monitor digital input events
    CBSDKCALLBACK_SERIAL,       ///< Monitor serial input events
    CBSDKCALLBACK_CONTINUOUS,   ///< Monitor continuous events
    CBSDKCALLBACK_TRACKING,     ///< Monitor video tracking events
    CBSDKCALLBACK_COMMENT,      ///< Monitor comment or custom events
    CBSDKCALLBACK_GROUPINFO,    ///< Monitor channel group info events
    CBSDKCALLBACK_CHANINFO,     ///< Monitor channel info events
    CBSDKCALLBACK_FILECFG,      ///< Monitor file config events
    CBSDKCALLBACK_POLL,         ///< respond to poll
    CBSDKCALLBACK_SYNCH,        ///< Monitor video synchronizarion events
    CBSDKCALLBACK_NM,           ///< Monitor NeuroMotive events
    CBSDKCALLBACK_CCF,          ///< Monitor CCF events
    CBSDKCALLBACK_IMPEDENCE,    ///< Monitor impedence events
    CBSDKCALLBACK_SYSHEARTBEAT, ///< Monitor system heartbeats (100 times a second)
    CBSDKCALLBACK_LOG,		    ///< Monitor system heartbeats (100 times a second)
    CBSDKCALLBACK_COUNT  ///< Always the last value
} cbSdkCallbackType;


CBSDKAPI    cbSdkResult cbSdkOpen(uint32_t nInstance,
                                  cbSdkConnectionType conType = CBSDKCONNECTION_DEFAULT,
                                  cbSdkConnection con = cbSdkConnection());


CBSDKAPI    cbSdkResult cbSdkGetTime(uint32_t nInstance, uint32_t * cbtime);

typedef void (* cbSdkCallback)(uint32_t nInstance, const cbSdkPktType type, const void* pEventData, void* pCallbackData);

CBSDKAPI    cbSdkResult cbSdkRegisterCallback(uint32_t nInstance, cbSdkCallbackType callbacktype, cbSdkCallback pCallbackFn, void* pCallbackData);

CBSDKAPI    cbSdkResult cbSdkUnRegisterCallback(uint32_t nInstance, cbSdkCallbackType callbacktype);

CBSDKAPI    cbSdkResult cbSdkClose(uint32_t nInstance);

CBSDKAPI    cbSdkResult cbSdkSetComment(uint32_t nInstance, uint32_t rgba, uint8_t charset, const char * comment = NULL);

CBSDKAPI    cbSdkResult cbSdkGetSampleGroupInfo(uint32_t nInstance, uint32_t proc, uint32_t group, char *label, uint32_t *period, uint32_t *length);

CBSDKAPI    cbSdkResult cbSdkGetSampleGroupList(uint32_t nInstance, uint32_t proc, uint32_t group, uint32_t *length, uint32_t *list);


#endif /* end of include guard: CBSDK_H_347FBB8B */
