/*! \file RingBuffer.h
    \brief A generic Ring buffer
*/
#ifndef RingBufferT_H_860EE4F5
#define RingBufferT_H_860EE4F5
#include <array>
#include <atomic>
#include <cstdint>
#include <mutex>
#include <vector>

namespace cc
{
/** \brief  A generic ring buffer implementation
    
    A ring buffer implementation, see https://kjellkod.wordpress.com/2012/11/28/c-debt-paid-in-full-wait-free-lock-free-queue/
    for explanations. This is mostly lock-free, single-producer, single-consumer, for most uses.
    The reader's head pointer is lock-protected, however, to allow the consumer to call
    reset() to reset the ring buffer.
    
    © Copyright 2018  - Jonas B. Zimmermann. All Rights Reserved.

    \author Jonas B. Zimmermann 
    \date 2018-08-30

**/
template<typename T, size_t Size>
class RingBufferT
{
public:
    using BufType = T;
    using VecT = std::vector<T>;
    
    RingBufferT () : m_tail(0), m_head(0), m_overflow_count(0) {};
    virtual ~RingBufferT () {};

    /** \brief Size of ring buffer
        
        Returns maximum number of elements the ring buffer can store
        
        \return size of ring buffer
        \sa wasFull(), wasEmpty()
    **/
    constexpr size_t size() const {return Size;}
    
    /** \brief push element
    
        Add a new element to RingBuffer. Only to be called by Producer thread
    
        \param [in] item to be added
        \return `true` if item was successfully added, `false` otherwise
        \sa pop()
    **/
    bool push(const T& item);
    
    /** \brief Reset buffer
        
        Resets buffer by setting head and tail to point to same element.
        
    **/
    void reset();
    
    /** \brief reads and removes oldest element from buffer
        
        \param [out] item removed element will be written here
        \return returns `true` if succesful, `false` otherwise
        \sa consume(), push()
    **/
    bool pop(T& item);
    
    /** \brief Reads all available data into vector-of-T
        
        \param [out] vec reference to vector of T
        \return returns `true` if items were available, `false` otherwise
        \sa pop()
    **/
    bool consume(VecT& vec);

    /** \brief Reads all available data and overflow count

        \param[out] vec reference to vector of T
        \param[out] overflow_count will store overflow count of ring buffer
        \return returns `true` if items were available, `false` otherwise
        \sa pop()
    **/
    bool consume(VecT& vec, size_t& overflow_count);

    /** \brief Checks whether buffer was empty at time this function was called
    
        \return returns `true` if empty
        \sa wasFull()
    **/
    bool wasEmpty() const;

    /** \brief Checks whether buffer was full at time this function was called

        \return returns `true` if full
        \sa wasEmpty()
    **/
    bool wasFull() const;
    bool isLockFree() const;
    
    /** \brief Gets how many times pushes failed because of overflow

        \return `size_t` overflow count
    **/
    size_t getOverflowCount() const { return m_overflow_count; };

private:
    enum { Capacity = Size + 1 };
    
    size_t increment(size_t idx) const;

    std::atomic <size_t>    m_tail;  // tail(input) index
    T                       m_array[Capacity];
    std::atomic<size_t>     m_head; // head(output) index
    
    std::mutex              m_head_mutex;   // we need this mutex so we can reset
                                            //the buffer from the writer thread
protected:
    std::atomic<size_t>     m_overflow_count;
};

// Push on tail
template<typename T, size_t Size>
bool RingBufferT<T, Size>::push(const T& item)
{	
  const auto current_tail = m_tail.load();
  const auto next_tail = increment(current_tail);
  if(next_tail != m_head.load())
  {
    m_array[current_tail] = item;
    m_tail.store(next_tail);
    return true;
  }
  ++ m_overflow_count;
  return false;  // full queue
}

// Without the lock, this is not thread safe! If Producer calls this while Consumer is in pop(),
// the state would be undefined.
template<typename T, size_t Size>
void RingBufferT<T, Size>::reset()
{
    std::lock_guard<std::mutex> lock(m_head_mutex);
    m_tail.store(0);
    m_head.store(0);
}

// Pop by Consumer can only update the head
template<typename T, size_t Size>
bool RingBufferT<T, Size>::pop(T& item)
{
    std::lock_guard<std::mutex> lock(m_head_mutex);
    
    const auto current_head = m_head.load();
    if(current_head == m_tail.load())
    {
        return false;   // empty queue
    }

    item = m_array[current_head];
    m_head.store(increment(current_head));
    return true;
}

template<typename T, size_t Size>
bool RingBufferT<T, Size>::consume(VecT &vec)
{
    m_overflow_count = 0;
    vec.resize(0);
    T item;
    bool retval = false;
    while (pop(item))
    {
        vec.push_back(item);
        retval = true;
    }
    return retval;
}

template<typename T, size_t Size>
bool RingBufferT<T, Size>::consume(VecT& vec, size_t& overflow_count)
{
    overflow_count = m_overflow_count;
    return consume(vec);
}

template<typename T, size_t Size>
bool RingBufferT<T, Size>::wasEmpty() const
{
    return (m_head.load() == m_tail.load());
}

// snapshot with acceptance that this comparison is not atomic
// (*) Used by clients or test, since push() avoid double load overhead by not
// using wasFull()
template<typename T, size_t Size>
bool RingBufferT<T, Size>::wasFull() const
{
    const auto next_tail = increment(m_tail.load());
    return (next_tail == m_head.load());
}

template<typename T, size_t Size>
bool RingBufferT<T, Size>::isLockFree() const
{
  return (m_tail.is_lock_free() && m_head.is_lock_free());
}

template<typename T, size_t Size>
size_t RingBufferT<T, Size>::increment(size_t idx) const
{
    return (idx + 1) % Capacity;
}
}

#endif /* end of include guard: RingBufferT_H_860EE4F5 */
