/*! \file SampleGroupRingBuffer.h
    \brief  Ring buffer to store sample group (continuous) data
*/
#ifndef SAMPLEGROUPRINGBUFFERT_H_99FCB67F
#define SAMPLEGROUPRINGBUFFERT_H_99FCB67F
#include "RingBuffer.h"

#include <algorithm>
#include <array>
#include <iterator>
#include <memory>
#include "cbhwlib.h"

#include "CereConnBaseTypes.h"
#include "CereConnHelpers.h"
#include "SampleGroupList.h"
#include "car/CommonAverager.h"

namespace cc
{

/** \brief SampleGroupRingBufferT subclasses RingBufferT with special members for sample group data
    \tparam T type of data point
    \tparam U type of time stamps
    \tparam MaxChannels maximum number of channels
    \tparam Size length of buffer (maximum number of samples stored)
**/
template<typename T, typename U, size_t MaxChannels, size_t Size, typename V = std::vector<T>>
class SampleGroupRingBufferT: public RingBufferT<V, Size>
{
public:
    using ContinuousDataSampleT = T; //!< data type of samples
    using ContinuousDataSampleVT = V; //!< type of item that will be stored in RingBufferT
    using ChannelListT = std::array<ChannelNumCont_t, MaxChannels>; //!< data type for storing channel numbers
    // using ChannelListDT = std::vector<ChannelNumCont_t>;
    
    constexpr static size_t NMaxChannels = MaxChannels;
    
    SampleGroupRingBufferT() : m_last_sample_time(0), m_channel_list({{0}}), m_channel_count(0), m_sample_period(0) {};
    SampleGroupRingBufferT(size_t sample_period) : m_last_sample_time(0), m_channel_list({{0}}), m_channel_count(0), m_sample_period(sample_period) {};
    virtual ~SampleGroupRingBufferT() {};
    
    /**\name Adding data methods
        See https://solarianprogrammer.com/2016/11/28/cpp-passing-c-style-array-with-size-information-to-function/
        for explanation of `const T(&channel_list)[N]`
    **/
    /**@{**/
    
    /** \brief Add samples from a cbPKT_GROUP package
        
        Copies `data` field from a `cbPKT_GROUP` package to buffer. This function performs the following
        checks:
            - test if buffer has free space (if not, this method fails and overflow count is increased)
            - test if package timestamp has increased (fail if not)
            - test if package has enough data. Note that the package only indicates data lenght in words,
                but data points are 16 bit integers. That means that this buffer may expect \f$2n\f$ channels,
                but the package only provides \f$2n - 1\f$ valid samples. The operation is still safe, but the
                junk data would not be detected.
        
        \param[in] pkt_grp packet of continuous data
        \return `true` if data was successfully added to buffer
    **/
    bool addSamplesFromPacket(const cbPKT_GROUP *pkt_grp);
    
    /** \brief Add samples from raw data
    
        Copies raw data to buffer. This function performs the following
        checks:
            - test if buffer has free space (if not, this method fails and overflow count is increased)
            - test if timestamp has increased (fail if not)
            - test if package has enough data, i.e. maxdatalen should be greater than or equal to the
                length of the internal channel list
    
        \param timestamp NSP time stamp of data
        \param[in] data pointer to data items
        \param maxdatalen number of data items
        \return `true` if data was successfully added to buffer
    **/
    bool addSamplesFromPacket(U timestamp, const T * data, const unsigned int maxdatalen);
    /**@}**/
    /** \brief Set number of ticks (in terms of the underlying 30kHz sample clock) between samples in buffer
    \param sample_period number of ticks between samples
    **/
    void setSamplePeriod(size_t sample_period) { m_sample_period = sample_period; }

    /** \name Reset buffer
    
    These methods reset the buffer as well as the channel list.
    Optionally, new channel list can be provided.
    **/
    /**@{**/
    /** \brief Reset buffer
    This will clear channel list, common average referencing, and last sample time.
    **/
    void reset();
    
    /**\brief Reset buffer and provide new channel ID list
    NB: unlike the plain reset() function, reset(const TT(&channel_list)[N]), reset(const ChannelListT),
    and reset(const ChannelNumList_t) do not reset Common Average Referencing (if enabled). Instead,
    the CAR channel ID map will be updated to match the new channel list.
    \param channel_list list of channel IDs (1-based)
    \return `true` if common average referencing will be performed after reset
    **/
    template<typename TT, size_t N>
    bool reset(const TT(&channel_list)[N]);
    bool reset(const ChannelListT &channel_list);
    bool reset(const ChannelNumList_t &channel_list);
    /**@}**/
    
    /** \brief Get timestamp of last sample in buffer
    \return timestamp
    **/
    U getLastSampleTime() const {return m_last_sample_time.load();}
    
    /** \brief Consume buffer
    
    Gets samples and channel information from buffer
    
    \param [out] sgc structure to hold buffer data
    \return `true` if there were samples returned
    **/
    bool consume(SampleGroupChunkT<T, U, MaxChannels, Size> &sgc);
    
    /**\brief Return number of channels**/
    size_t getChannelCount() const {return m_channel_count;}
    
    /**\brief Return number of ticks per sample period**/
    size_t getSamplePeriod() const {return m_sample_period;}
    
    /**\brief Return const_iterator to first item of channel list **/
    typename ChannelListT::const_iterator ch_list_cbegin() const {return m_channel_list.cbegin();}
    
    /**\brief Return const_iterator to end of channel list **/
    // Note for implementation: m_channel_list is an array; but its valid entries may be fewer
    // than its capacity, hence the unusual way of calculating the end.
    typename ChannelListT::const_iterator ch_list_cend() const {return m_channel_list.cbegin() + m_channel_count;}
    
    /** \name Common Average Referencing
    
    Control Common Average Referenncing for channels in this sample group. When channels
    are selected with setCARChannels(), each time samples are added with addSamplesFromPacket(),
    the samples from the selected channels are averaged and then subtracted from each of the
    channels. See car::CommonAverager for details.
    Channel IDs passed to setCARChannels() are mapped to the indices in the sample group's
    channel ID list (the same list set with the reset(channel_list) functions). If the there
    are matching channels in the sample group's channel list, CAR will be performed on every
    newly received sample. If there are no matching channels, no CAR will be performed.
    
    **/
    /**@{**/
    /** \brief Sets channnels for Common Average Reference
    \param carChannelList list of channel IDs to be used for CAR
    \return `true` if CAR successfully set up, `false` otherwise
    \sa setCARChannels(ChannelNumList_t &&carChannelList), disableCAR()
    **/
    bool setCARChannels(const ChannelNumList_t &carChannelList);
    
    /** \brief Sets channnels for Common Average Reference
    \param carChannelList list of channel IDs to be used for CAR. May move argument.
    \return `true` if CAR successfully set up, `false` otherwise
    \sa setCARChannels(const ChannelNumList_t &carChannelList), disableCAR()
    **/
    bool setCARChannels(ChannelNumList_t &&carChannelList);

    /** \brief Disables CAR
    
    \sa setCARChannels(ChannelNumList_t &&carChannelList), setCARChannels(const ChannelNumList_t &carChannelList)
    **/
    void disableCAR();
    /**@}**/
protected:
    
    std::atomic<U>              m_last_sample_time;
    ChannelListT                m_channel_list;
    size_t                      m_channel_count;
    size_t                      m_sample_period;
    std::atomic<bool>           m_channel_changed; //!< stores if channel list has changed since last consume
    std::mutex                  m_consume_mutex;
    
    std::mutex                  m_car_mutex;
    ChannelNumList_t            m_car_channels;
    
    std::unique_ptr<CommonAverager<T,V>>    mp_car;
    
private:
#ifdef CC_TESTING_DEBUG
    FRIEND_TEST(ContRBTest, SetCAR);
#endif
};
template<typename T, typename U, size_t MaxChannels, size_t Size, typename V>
constexpr size_t SampleGroupRingBufferT<T, U, MaxChannels, Size, V>::NMaxChannels;

template<typename T, typename U, size_t MaxChannels, size_t Size, typename V>
bool SampleGroupRingBufferT<T, U, MaxChannels, Size, V>::addSamplesFromPacket(const cbPKT_GROUP *pkt_grp)
{
    if (this->wasFull())
    {
        ++ (this->m_overflow_count);
        return false;  // full queue
    }
    // "Channel list length must match length of data snippet
    // pkt_grp.dlen gives length in quadlets, i.e. 2 int16 samples
    if (m_channel_count > pkt_grp->dlen * 2)
    {
        return false;
    }
    if (pkt_grp->time <= m_last_sample_time.load())
    {
        return false;
    }
    // TODO: Make the push operation move, not copy
    V cds(m_channel_count);
    std::copy(std::begin(pkt_grp->data), std::begin(pkt_grp->data) + m_channel_count, cds.begin());
    if (mp_car)
    {
        (*mp_car)(cds);
    }
    RingBufferT<V, Size>::push(std::move(cds));
    m_last_sample_time.store(pkt_grp->time);
    return true;
}

template<typename T, typename U, size_t MaxChannels, size_t Size, typename V>
bool SampleGroupRingBufferT<T, U, MaxChannels, Size, V>::addSamplesFromPacket(U timestamp, const T * data, const unsigned int maxdatalen)
{
    // Attempting to read more data than is available
    if (this->wasFull())
    {
        ++ (this->m_overflow_count);
        return false;  // full queue
    }
    if (m_channel_count > maxdatalen)
    {
        return false;
    }
    if (timestamp <= m_last_sample_time.load())
    {
        return false;
    }
    V cds(m_channel_count);
    std::copy(data, data + m_channel_count, cds.begin());
    if (mp_car)
    {
        (*mp_car)(cds);
    }
    RingBufferT<V, Size>::push(std::move(cds));
    m_last_sample_time.store(timestamp);
    return true;
}

template<typename T, typename U, size_t MaxChannels, size_t Size, typename V>
void SampleGroupRingBufferT<T, U, MaxChannels, Size, V>::reset()
{
    std::lock_guard<std::mutex> lock(m_consume_mutex);
    RingBufferT<V, Size>::reset();
    m_last_sample_time.store(0);
    m_channel_list.fill(0);
    m_channel_count = 0;
    m_channel_changed = true;
    mp_car = nullptr;
    m_car_channels.resize(0);
}

template<typename T, typename U, size_t MaxChannels, size_t Size, typename V>
template<typename TT, size_t N>
bool SampleGroupRingBufferT<T, U, MaxChannels, Size, V>::reset(const TT(&channel_list)[N])
{
    bool has_car = mp_car != nullptr;
    ChannelNumList_t tmp_car_ch_list(std::move(m_car_channels));
    static_assert(MaxChannels == N, "Channel list length must match MaxChannels");
    reset();
    std::copy(std::cbegin(channel_list), std::end(channel_list), m_channel_list.begin());
    m_channel_count = std::count_if(m_channel_list.begin(), m_channel_list.end(), [](size_t ch){return ch != 0;});
    if (has_car)
    {
        return setCARChannels(std::move(tmp_car_ch_list));
    }
    return false;
}

template<typename T, typename U, size_t MaxChannels, size_t Size, typename V>
bool SampleGroupRingBufferT<T, U, MaxChannels, Size, V>::reset(const ChannelListT &channel_list)
{
    bool has_car = mp_car != nullptr;
    ChannelNumList_t tmp_car_ch_list(std::move(m_car_channels));
    reset();
    std::copy(channel_list.cbegin(), channel_list.end(), m_channel_list.begin());
    m_channel_count = std::count_if(m_channel_list.begin(), m_channel_list.end(), [](size_t ch){return ch != 0;});
    if (has_car)
    {
        return setCARChannels(std::move(tmp_car_ch_list));
    }
    return false;
}

template<typename T, typename U, size_t MaxChannels, size_t Size, typename V>
bool SampleGroupRingBufferT<T, U, MaxChannels, Size, V>::reset(const ChannelNumList_t &channel_list)
{
    bool has_car = mp_car != nullptr;
    ChannelNumList_t tmp_car_ch_list(std::move(m_car_channels));
    reset();
    m_channel_list.fill(0);
    std::copy(channel_list.cbegin(), channel_list.end(), m_channel_list.begin());
    m_channel_count = channel_list.size();
    if (has_car)
    {
        return setCARChannels(std::move(tmp_car_ch_list));
    }
    return false;
}


template<typename T, typename U, size_t MaxChannels, size_t Size, typename V>
bool SampleGroupRingBufferT<T, U, MaxChannels, Size, V>::consume(SampleGroupChunkT<T, U, MaxChannels, Size> &sgc)
{
    sgc.buffer.resize(0);
    sgc.channel_ids.resize(0);
    
    sgc.channel_ids.insert(sgc.channel_ids.end(), m_channel_list.begin(), m_channel_list.begin() + m_channel_count);
    
    V item;
    bool retval = false;
    std::lock_guard<std::mutex> lock(m_consume_mutex);
    auto it_sgc_buf =  sgc.buffer.end();
    while (this->pop(item))
    {
        sgc.buffer.insert(it_sgc_buf, item.begin(), item.begin() + m_channel_count);
        it_sgc_buf = sgc.buffer.end();
        retval = true;
    }
    
    sgc.last_timestamp = m_last_sample_time.load();
    sgc.sample_period = m_sample_period;
    sgc.channel_ids_changed = m_channel_changed;
    sgc.overflow_count = this->m_overflow_count;
    this->m_overflow_count = 0;
    m_channel_changed = false;
    return retval;
}

template<typename T, typename U, size_t MaxChannels, size_t Size, typename V>
bool SampleGroupRingBufferT<T, U, MaxChannels, Size, V>::setCARChannels(const ChannelNumList_t &carChannelList)
{
    ChannelNumList_t tmpChannelList(carChannelList);
    return setCARChannels(std::move(tmpChannelList));
}

template<typename T, typename U, size_t MaxChannels, size_t Size, typename V>
bool SampleGroupRingBufferT<T, U, MaxChannels, Size, V>::setCARChannels(ChannelNumList_t &&carChannelList)
{
    std::lock_guard<std::mutex> lock(m_car_mutex);
    m_car_channels = std::move(carChannelList);
    ChannelNumList_t car_channel_map;
    
    size_t i_in_list = 0;
    
    auto begin = m_car_channels.cbegin();
    auto end = m_car_channels.cend();
    
    while (begin != end)
    {
        auto it = std::find(m_channel_list.cbegin(), m_channel_list.cend(), *begin); // BlackRock channels are 1 based
        if (it != m_channel_list.cend())
        {
            auto i_in_channel_list = it - m_channel_list.cbegin();
            car_channel_map.push_back(i_in_channel_list);
        }
        ++ i_in_list;
        ++ begin;
    }
    if (car_channel_map.size() > 0)
    {
        mp_car = std::make_unique<CommonAverager<T, V>>(std::move(car_channel_map));
        return true;
    }
    return false;
}

template<typename T, typename U, size_t MaxChannels, size_t Size, typename V>
void SampleGroupRingBufferT<T, U, MaxChannels, Size, V>::disableCAR()
{
    mp_car = nullptr;
}


/** \brief Unites a number of sample group buffers together for easier access
    \tparam T type of data point
    \tparam U type of time stamps
    \tparam MaxChannels maximum number of channels
    \tparam Size length of buffer (maximum number of samples stored)
    \tparam NSampleGroups number of sample groups to store
**/
template<typename T, typename U, size_t MaxChannels, size_t Size, size_t NSampleGroups, typename V = std::vector<T>>
class SampleGroupRingBufferCollectionT
{
public:
    SampleGroupRingBufferCollectionT(): m_sgrb(NSampleGroups) {};

    /** \name Add samples to buffer
    @{**/

    bool addSamplesFromPacket(const cbPKT_GROUP *pkt_grp);
    
    // newly refactored adding function
    bool addSamplesFromPacket(const size_t group, U timestamp, T * data, const unsigned int maxdatalen);
    /**@}**/
    
    /** \brief Consume all buffers
        \param [out] sgl sample group data are written to this collection of sample group chunks
    **/
    void consume(SampleGroupListT<T, U, MaxChannels, Size, NSampleGroups> &sgl);
    
    /** \brief Get ticks per sample period
    \param group sample group index (one-based!)
    \return number of ticks per sample period
    **/
    size_t getSamplePeriodForGroup(const size_t group) {return m_sgrb[group - 1].getSamplePeriod(); }
    
    /** \brief Set ticks per sample period
    \param group sample group index (one-based!)
    \param period ticks per sample period
    **/
    void setSamplePeriodForGroup(const size_t group, size_t period) { m_sgrb[group - 1].setSamplePeriod(period); }
    
    /** \brief Get count of overflow events, per sample group buffer
    \return an array containg overflow count per sample group
    **/
    std::array<size_t, NSampleGroups> getOverflowCount() const;
    
    /** \brief Reset a group's buffer by supplying a new channel list
    \param group sample group index (one-based!)
    \param channel_list new list of channel IDs in sample group
    \return `true` if common average referencing will be performed after reset
    **/
    bool reset(const size_t group, const ChannelNumList_t &channel_list);
    
    /** \name Common Average Referencing
    **/
    /**@{**/
    /** \brief Sets channnels for Common Average Reference for sample group
    \param group sample group index (one-based!)
    \param carChannelList list of channel IDs to be used for CAR
    \return `true` if CAR successfully set up, `false` otherwise
    \sa setCARChannels(size_t group, ChannelNumList_t &&carChannelList), disableCAR(),
    SampleGroupRingBufferT::setCARChannels()
    **/
    bool setCARChannels(const size_t group, const ChannelNumList_t &carChannelList);
    
    /** \brief Sets channnels for Common Average Reference for sample group
    \param group sample group index (one-based!)
    \param carChannelList list of channel IDs to be used for CAR
    \return `true` if CAR successfully set up, `false` otherwise
    \sa setCARChannels(size_t group, const ChannelNumList_t &carChannelList), disableCAR(),
    SampleGroupRingBufferT::setCARChannels()
    **/
    bool setCARChannels(const size_t group, ChannelNumList_t &&carChannelList);
    
    /** \brief Disables CAR for sample group
    \param group sample group index (one-based!)
    \sa setCARChannels(size_t group, ChannelNumList_t &&carChannelList),
    setCARChannels(size_t group, const ChannelNumList_t &carChannelList),
    SampleGroupRingBufferT::disableCAR()
    **/
    void disableCAR(const size_t group);
    /**@}**/
private:
    std::vector<SampleGroupRingBufferT<T, U, MaxChannels, Size, V>> m_sgrb;
    
};

template<typename T, typename U, size_t MaxChannels, size_t Size, size_t NSampleGroups, typename V>
bool SampleGroupRingBufferCollectionT<T, U, MaxChannels, Size, NSampleGroups, V>::addSamplesFromPacket(const cbPKT_GROUP *pkt_grp)
{
    return m_sgrb[pkt_grp->type - 1].addSamplesFromPacket(pkt_grp);
}

template<typename T, typename U, size_t MaxChannels, size_t Size, size_t NSampleGroups, typename V>
bool SampleGroupRingBufferCollectionT<T, U, MaxChannels, Size, NSampleGroups, V>::addSamplesFromPacket(const size_t group, U timestamp, T * data, const unsigned int maxdatalen)
{
    return m_sgrb[group - 1].addSamplesFromPacket(timestamp, data, maxdatalen);
}

template<typename T, typename U, size_t MaxChannels, size_t Size, size_t NSampleGroups, typename V>
void SampleGroupRingBufferCollectionT<T, U, MaxChannels, Size, NSampleGroups, V>::consume(SampleGroupListT<T, U, MaxChannels, Size, NSampleGroups> &sgl)
{
    auto i_sgrb = m_sgrb.begin();
    auto i_sgl = sgl.sgc.begin();
    
    for (; i_sgrb != m_sgrb.end() and i_sgl != sgl.sgc.end(); ++ i_sgrb, ++ i_sgl)
    {
        i_sgrb->consume(*i_sgl);
    }
    
}

template<typename T, typename U, size_t MaxChannels, size_t Size, size_t NSampleGroups, typename V>
std::array<size_t, NSampleGroups> SampleGroupRingBufferCollectionT<T, U, MaxChannels, Size, NSampleGroups, V>::getOverflowCount() const
{
    std::array<size_t, NSampleGroups> oc = {{0}};
    auto i_sgrb = m_sgrb.begin();
    auto i_oc = oc.begin();
    for (; i_sgrb != m_sgrb.end() and i_oc != oc.end(); ++ i_sgrb, ++ i_oc)
    {
        *i_oc = i_sgrb->getOverflowCount();
    }
    return oc;
}

template<typename T, typename U, size_t MaxChannels, size_t Size, size_t NSampleGroups, typename V>
bool SampleGroupRingBufferCollectionT<T, U, MaxChannels, Size, NSampleGroups, V>::reset(const size_t group, const ChannelNumList_t &channel_list)
{
    return m_sgrb[group - 1].reset(channel_list);
}

template<typename T, typename U, size_t MaxChannels, size_t Size, size_t NSampleGroups, typename V>
bool SampleGroupRingBufferCollectionT<T, U, MaxChannels, Size, NSampleGroups, V>::setCARChannels(const size_t group, const ChannelNumList_t &carChannelList)
{
    return m_sgrb[group - 1].setCARChannels(carChannelList);
}

template<typename T, typename U, size_t MaxChannels, size_t Size, size_t NSampleGroups, typename V>
bool SampleGroupRingBufferCollectionT<T, U, MaxChannels, Size, NSampleGroups, V>::setCARChannels(const size_t group, ChannelNumList_t &&carChannelList)
{
    return m_sgrb[group - 1].setCARChannels(carChannelList);
}

template<typename T, typename U, size_t MaxChannels, size_t Size, size_t NSampleGroups, typename V>
void SampleGroupRingBufferCollectionT<T, U, MaxChannels, Size, NSampleGroups, V>::disableCAR(const size_t group)
{
    m_sgrb[group - 1].disableCAR();
}


} // namespace cc


#endif /* end of include guard: SAMPLEGROUPRINGBUFFERT_H_99FCB67F */
