#ifndef SPIKEBANDPOWERESTIMATOR_H_5A7C0342
#define SPIKEBANDPOWERESTIMATOR_H_5A7C0342
/** \file SpikeBandPowerEstimator.h
    A class implementing efficient spike band power estimation for a set of channels
    
© Copyright 2018  - Jonas B. Zimmermann. All Rights Reserved.
    \author Jonas B. Zimmermann 
    \date 2018-11-28
**/

#include <chrono>
#include <cstdio>
#include <memory>
#include <queue>
#include <unordered_map>
#include "CereRingBuffer.h"
#include <armadillo>
#include "sigpack.h"

#include "spdlog/async.h"
#include "spdlog/sinks/stdout_color_sinks.h"
#include "spdlog/spdlog.h"


namespace cc
{
//! Namespace for Spike Band Power Estimation related classes
namespace sbpe
{
    using namespace std::chrono;
    using namespace std::literals;

/** \brief  FiltRMSSampler
    
    Class encapsulating IIR filter coefficients, filter buffers, as well as an RMS process.
    Main access is through operator().
        
**/
template<typename T, typename U, typename V>
class FiltRMSSampler
{
public:
    using sample_type = T;                              //!< input data type
    using rms_type = V;                                 //!< what type should spike band power be stored as
    using rms_container_type = std::vector<rms_type>;   //!< container for RMS data
    
    using iir_filter_in_type = T;                       //!< type for filter input values
    using iir_filter_coeff_type = U;                    //!< type for coefficients
    using iir_filter_coeff_container_type = arma::Col<iir_filter_coeff_type>; //!< type holding filter coefficients
    using iir_filter_out_type = V;                      //!< type for filter output
    
    using iir_filter_type = sp::IIR_filt<iir_filter_in_type, iir_filter_coeff_type, iir_filter_out_type>; //!< type of the IIR filter

    
    FiltRMSSampler (const size_t n_channels); //!< constructor for `n_channels` channels
    FiltRMSSampler (const size_t n_channels, const iir_filter_coeff_container_type b, const iir_filter_coeff_container_type a); //!< constructor for number of channels and filter coefficients
    FiltRMSSampler (const size_t n_channels, const size_t n_samples, const arma::Col<iir_filter_coeff_type> b, const arma::Col<iir_filter_coeff_type> a); //!< constructor for number of channels, number of samples, and filter coefficients
    virtual ~FiltRMSSampler ();

    void resize(const size_t n_channels);       //!< resize for `n_channels` channels. Will zero buffers
    
    void setNSamples(const size_t n_samples); //!< set number of samples used for an RMS cycle
    size_t getNSamples() const; //!< get number of samples used for an RMS cycle
    
    void setNAvgBins(const size_t n_bins); //!< set number of last cycles that will be averaged before taking the root. Set to 1 for no averaging.
    size_t getNAvgBins() const; //!< get number of last cycles to be averaged before taking the root.

    void setIIRFilterCoeffs(const iir_filter_coeff_container_type b, const iir_filter_coeff_container_type a); //!< set b, a IIR filter coefficients, which will affect all channels
    
    template<typename Iter>
    bool operator()(Iter _begin, Iter _end); //!< collect samples from iterator, add apply them to successive channels

    
    typename rms_container_type::const_iterator cbegin();
    typename rms_container_type::const_iterator cend();
    typename rms_container_type::iterator begin();
    typename rms_container_type::iterator end();

private:
    void _resize();
    void applyIIRFilterCoeffs();
    void zeroBuffers();

private:
    size_t                                      m_n_channels;
    size_t                                      m_n_samples;
    
    std::vector<size_t>                         m_sample_counts;
    size_t                                      m_sample_count;
    
    std::vector<iir_filter_type>                m_iir_filters;
    iir_filter_coeff_container_type             m_iir_filter_coeff_b;      ///!< MA Filter coefficients
    iir_filter_coeff_container_type             m_iir_filter_coeff_a;      ///!< AR Filter coefficients
    
    rms_container_type                          m_ms_buffer;
    rms_container_type                          m_rms_out_buffer;
    
    size_t                                      m_n_average;
    std::unique_ptr<arma::Mat<V>>               mp_sbp_history;
    size_t                                      m_sbp_history_current_item;
};


/** \brief  SpikeBandPowerEstimator transforms raw samples int an estimate of power in the spike band
    
    Real-time estimation of spike band power.

    Data flow, calling from CereConn:

    - Capture of raw packets in CereConn::cbSdkCallbackContinuous
        - if `pPktSgrp->group` == getAcceptedSampleGroup()
            - `p_sbpe->p_sgrp_rb->addSamplesFromPacket(pPktSgrp->time, ch_list_v, pPktSgrp->data)`
    
    To get data in other threads: `p_sgrp_rb->consume()`
        
**/
class SpikeBandPowerEstimator
{
#ifdef CC_TESTING_DEBUG
    FRIEND_TEST(SpikeBandPowerEstimatorBasic,AddSamples);
    FRIEND_TEST(SpikeBandPowerEstimatorBasic, AddContSamples);
    FRIEND_TEST(SpikeBandPowerEstimatorBasic, AddContSamplesWithChannelChange1);
    FRIEND_TEST(SpikeBandPowerEstimatorBasic, AddContSamplesWithChannelChange2);
    FRIEND_TEST(SpikeBandPowerEstimatorPerformance, FastEat);
    FRIEND_TEST(SpikeBandPowerEstimatorBasic, AcceptedSampleGroup);
    FRIEND_TEST(SpikeBandPowerEstimatorBasic, SelectSampleGroup);
    FRIEND_TEST(SpikeBandPowerEstimatorBasic, ChannelMapping);
    FRIEND_TEST(SpikeBandPowerEstimatorBasic, OverFlow);
    FRIEND_TEST(SpikeBandPowerEstimatorBasic, AvgBins);
#endif
private:
    // based on https://stackoverflow.com/a/38253266/1893154 
    using maxres_sys_or_steady =
        std::conditional<
            system_clock::period::den <= steady_clock::period::den,
            system_clock, steady_clock
          >::type;
public:
    using clock =
        std::conditional<
            high_resolution_clock::is_steady,
            high_resolution_clock, maxres_sys_or_steady
          >::type;                                                              //!< clock to be used for thread timing
    
    
public:
    using size_type = size_t;                      //!< size type to count elements with
    
    using timestamp_callback_type = std::function<Timestamp_t()>;               //!< type for callbacks to get a current time stamp
    
    using cont_sample_rb_type = SampleGroupRingBuffer;                          //!< type for storing continuous samples
    
    using sbp_type = double;                                                    //!< what type should spike band power be stored as
    
    using sbp_item_type = std::vector<sbp_type>;                                //!< type to capture spike rates for all units at one time point
    using sbp_ring_buffer_item_type = std::pair<Timestamp_t, sbp_item_type>;    //!< type to store rates together with a timestamp in a ring buffer
    using sbp_ring_buffer_type = RingBufferT<sbp_ring_buffer_item_type, 1000>;  //!< type of ring buffer to store spike band power values
    using sbp_list_type = sbp_ring_buffer_type::VecT;                           //!< type to store rates as read out from ring buffer
    using sbp_interval_type = std::chrono::milliseconds;                        //!< intervals for spike rate loops given in this unit
    
    using ch_type = ChannelNumCont_t;                                           //!< describe a channel by index
    using ch_list_type = ChannelNumList_t;                                      //!< a list of channel indices
    using ch_map_type = std::unordered_map<ch_type, size_type>;                 //!< a map of ch_type to indices in spike band power vector
    using ch_sgrp_to_ch_list_type = std::vector<std::pair<size_type, size_type>>; //! a list of mappings of indices in raw buffer to channel list
    
    using iir_filt_RMS_type = FiltRMSSampler<double, double, sbp_type>;         //!< type of object keeping track off IIR filter and RMS buffers
    using iir_filt_coeff_type = iir_filt_RMS_type::iir_filter_coeff_container_type;
    
    SpikeBandPowerEstimator();                                                  //!< default constructor, 1 channel, 300–6000Hz 2nd order Butterworth
                                                                                // bandpass (assuming 30kHz sampling rate), 600 samples, processing
                                                                                // loop in thread runs every 10ms
    SpikeBandPowerEstimator(size_type n_channels,
        sbp_interval_type t_sbp_loop             = 10ms);                       //!< set channels IDs 1--n_channels, in Blackrock counting
                                                                                // as well as processing loop time
    SpikeBandPowerEstimator(ch_list_type ch_list,                               //!< set channels IDs as list as well as processing loop time
        sbp_interval_type t_sbp_loop             //!< set calculation interval
            = 10ms);
    

    virtual ~SpikeBandPowerEstimator ();

    /** \name Starting and Stopping **/
    /**@{**/
    /** \brief Start spike band power processing
        
        Starts processing thread. Note that a SpikeBandPowerEstimator::sbp_ring_buffer_type has to be
        attached with attachSpikeBandPowerRB().
        
        \sa signalStop(), attachSpikeBandPowerRB()
    **/
    void signalStart();
    
    /** \brief Send processing thread a signal to stop
        
        This function is non-blocking.
        
        \sa signalStart()
    **/
    void signalStop();


    /** \brief Pass a pointer to a SpikeBandPowerEstimator::sbp_ring_buffer_type to store spike band power in
    
        Spike band power will only be calculated if SpikeBandPowerEstimator has
        a valid pointer to a SpikeBandPowerEstimator::sbp_ring_buffer_type.
    
        \param sbp_rb shared pointer to SpikeBandPowerEstimator::sbp_ring_buffer_type to be used to store spike band power
        \sa ringBufferInvalid()
    **/
    void attachSpikeBandPowerRB(std::shared_ptr<sbp_ring_buffer_type> sbp_rb);

    /** \brief determine if ring buffer is invalid
        
        This could be the case if no ring buffer has been attached with attachSpikeBandPowerRB(),
        or if it or internal buffers need to be reset (e.g. when channel count changes)
        
        \return true if ring buffer is invalid
        \sa attachSpikeBandPowerRB()
    **/
    bool ringBufferInvalid() const;
    /**@}**/

    /** \name Channels and Mappings
        These members determine of which channels spike band power will be calculated and how
        they map to the elements in the output ringbuffer.
    **/
    /**@{**/
    
    /** \brief set which sample group packets will be accepted
        
        The default is *6*, i.e. raw data samples
        
        \param  group number of the accepted sample group
        \sa getAcceptedSampleGroup()
    **/
    void setAcceptedSampleGroup(const size_t group);

    /** \brief get the sample group number for which samples will be processed

        \return the currently set accepted sample group.
        \sa setAcceptedSampleGroup()
    **/
    size_t getAcceptedSampleGroup() const;
    
    /** \brief Set a list of channels for which to calculate spike band power
    
        Note that Channel in SpikeBandPowerEstimator::ch_list_type has 1-based indices.
    
        \param  ch_list a vector of channel indices
        \sa fillChannelList(), getChannelMapping(), getChannelList()
    **/
    void setChannelList(const ch_list_type & ch_list);

    /** \brief Enable spike band power calculation for all channels
    
        \param  n_channels Number of channels, starting from 1, to be included in spike band power estimation
        \sa setChannelList(const ch_list_type & ch_list), getChannelList(), getChannelMapping()
    **/
    void fillChannelList(const size_type n_channels);


    /** \brief return the current mapping of channels to spike band power
    
        The current mapping of channels to the elements of the spike band power
        vector may not be ordered. This function provides both a look-up table
        for channels to indices in the spike band power vector, as well as a
        list of channels in the same order as in the SBP.

        \return a std::pair of SpikeBandPowerEstimator::ch_list_type and SpikeBandPowerEstimator::ch_map_type.
        \sa getChannelList(), setChannelList(), fillChannelList()
    **/
    std::pair<ch_list_type, ch_map_type> getChannelMapping() const;

    /** \brief get a list of current channels for which spike band power is computed
    
        The returned list is in the same order as the SBP for these channels
        as stored in the spike band power buffer.
    
        \return the current channel list
        \sa getChannelMapping(), setChannelList(), fillChannelList()
    **/
    ch_list_type getChannelList() const;
    
    /**@}**/
    
    /** \brief set interval for the spike band power estimator loop
        
        SpikeBandPowerEstimator will attempt to keep the time between beginnings
        of consecutive iterations of the processing loop constant at this
        interval.
        
        \param  new_interval new interval
        \sa getLoopInterval()
    **/
    void setLoopInterval(const sbp_interval_type & new_interval);
    
    /** \brief get current spike band power estimator loop time
        
        \return current interval
        \sa setLoopInterval()
    **/
    sbp_interval_type getLoopInterval() const;
    
    /** \brief set IIR filter coefficients
        \param[in] b input coefficients
        \param[in] a feedback coefficients
        \sa getIIRFilterCoeffs(), setIIRFilterNSamples()
    **/
    void setIIRFilterCoeffs(const iir_filt_RMS_type::iir_filter_coeff_container_type &b, const iir_filt_RMS_type::iir_filter_coeff_container_type &a);

    /** \brief get IIR filter coefficients
        \return a std::pair of coefficient containers, the first item representing the
            input coefficients, the second item representing the feedback coefficients
        \sa setIIRFilterCoeffs(), setIIRFilterNSamples()
    **/
    std::pair<iir_filt_RMS_type::iir_filter_coeff_container_type, iir_filt_RMS_type::iir_filter_coeff_container_type> getIIRFilterCoeffs() const;

    
    /**
        \brief set number of samples the RMS part of the IIR/RMS filter integrates over
        \sa getIIRFilterNSamples(), setIIRFilterCoeffs()
    **/
    void setIIRFilterNSamples(const size_t n_samples);
    
    /**
        \brief get number of samples the RMS part of the IIR/RMS filter integrates over
        \sa setIIRFilterNSamples()
    **/
    size_t getIIRFilterNSamples() const;

    /**
        \brief set number of bins the RMS part of the IIR/RMS filter averages over
        \sa getRMSNBins(), getIIRFilterNSamples()
    **/
    void setRMSNBins(const size_t n_bins);

    /**
        \brief get number of bins the RMS part of the IIR/RMS filter averages over
        \sa setRMSNBins()
    **/
    size_t getRMSNBins() const;


    /** \name Interaction with underlying RingBuffer **/
    /**@{**/
    /** \bief Resets ring buffer, sets new channel list
        
        Performs reset iff group is accepted by SBPE
    
        \param group sampling group
        \param[in] channel_list list of channel ids (1-based)
    **/
    bool reset(const size_t group, const ChannelNumList_t &channel_list);
    /** \bief Resets ring buffer, sets new channel list
        \param[in] channel_list list of channel ids (1-based)
    **/
    bool reset(const ChannelNumList_t &channel_list);
    
    /** \brief adds continuous samples
    
        This function assumes channel list has not changed
    
        \param timestamp NSP timestamp, extracted from data packet
        \param[in] data pointer to data; must be at least as long as channel_list
        \return true if operation succesful
        \sa addSamplesFromPacket()
    **/
    bool addSamplesFromPacket(const Timestamp_t timestamp, const SampleGroupRingBuffer::ContinuousDataSampleT * data, const unsigned int maxdatalen);
    
    /** \brief adds continuous samples
    
        This function assumes channel list has not changed
    
        \param group sample group
        \param timestamp NSP timestamp, extracted from data packet
        \param data pointer to data; must be at least as long as channel_list
        \return true if operation succesful
        \sa addSamplesFromPacket()
    **/
    bool addSamplesFromPacket(const size_t group, const Timestamp_t timestamp, const SampleGroupRingBuffer::ContinuousDataSampleT * data, const unsigned int maxdatalen);
    /** \brief Add samples from a cbPKT_GROUP package
        
        Copies `data` field from a `cbPKT_GROUP` package to buffer.

        \param [in] pkt_grp packet of continuous data
        \return true if operation succesful
        \sa addSamplesFromPacket()
    **/
    bool addSamplesFromPacket(const cbPKT_GROUP *pkt_grp);
    
    /** \brief Sets channnels for Common Average Reference for sample group
    \param group sample group index (one-based!)
    \param carChannelList list of channel IDs to be used for CAR
    \return `true` if CAR successfully set up, `false` otherwise
    \sa setCARChannels(size_t group, ChannelNumList_t &&carChannelList), disableCAR(),
    SampleGroupRingBufferT::setCARChannels()
    **/
    bool setCARChannels(const size_t group, const ChannelNumList_t &carChannelList);
    
    /** \brief Sets channnels for Common Average Reference for sample group
        \param group sample group index (one-based!)
        \param carChannelList list of channel IDs to be used for CAR
        \return `true` if CAR successfully set up, `false` otherwise
        \sa setCARChannels(size_t group, const ChannelNumList_t &carChannelList), disableCAR(),
        SampleGroupRingBufferT::setCARChannels()
    **/
    bool setCARChannels(const size_t group, ChannelNumList_t &&carChannelList);
    
    /** \brief Disables CAR for sample group
        \param group sample group index (one-based!)
        \sa setCARChannels(size_t group, ChannelNumList_t &&carChannelList),
        setCARChannels(size_t group, const ChannelNumList_t &carChannelList),
        SampleGroupRingBufferT::disableCAR()
    **/
    void disableCAR(const size_t group);
    /**@}**/
    
    /** \brief clone logger passed in argument and set it as SpikeBandPowerEstimator's logger
        
        \param logger the logger to be cloned
    **/
    void setLogger(std::shared_ptr<spdlog::logger> logger);

    /** \brief Bind a callback to retrieve a timestamp
    
        This is used to indicate the last timestamp until spikes could
        have possibly been included. This should be called by client
        code (e.g. in cc::CereConn), to enable link to a data acquisition system.
    
        \param tscb callback to return a cc::Timestamp_t
    **/
    void bindTimestampCallback(const timestamp_callback_type & tscb);
    
private:
    void updateChannelPairList(const ChannelNumList_t & channel_ids); //<! update mapping from sample buffer index to index in channel list / iir_filter / sbp_buffer
    
    template<typename Iter>
    void updateChannelPairList(Iter begin, Iter end);
    
    
    void stopProcessThread();
    void processThreadFun();

    bool applyFilters();

    void initLogger();
    
    void applyChannelList();
    inline size_type n_ch() const;

private:
    ch_list_type                                m_ch_list;
    ch_map_type                                 m_ch_map;
    ch_sgrp_to_ch_list_type                     m_ch_pair_v; //!< stores a list of pairs mapping position in the continuous sample ring buffer
                                                // to position in the channel list (i.e. in mp_sbp_rb)
    
    std::unique_ptr<cont_sample_rb_type>        mp_cont_data; //!< Ring buffer to store incoming raw data
    
    iir_filt_RMS_type::iir_filter_coeff_container_type m_iir_filt_coeff_b, m_iir_filt_coeff_a; //!< mp_iir_filt_rms IIR filter coefficients. We store them here because
                                                // changing them is not thread safe
    std::unique_ptr<iir_filt_RMS_type>          mp_iir_filt_rms; //!< in processing loop, raw data are fed to this, which calculates RMS
        
    std::shared_ptr<sbp_ring_buffer_type>       mp_sbp_rb; //!< stores a pointer to a spike band power ring buffer where RMS get stored
    
    
    sbp_interval_type                           m_t_sbp_loop;
    
    timestamp_callback_type                     m_timestamp_callback;

    
    std::shared_ptr<spdlog::logger>             mp_logger;
    
    std::thread                                 m_process_thread;
    std::atomic<bool>                           m_keep_going;
    std::atomic<bool>                           m_thread_loop_over_time_flag;
    clock::duration                             m_thread_loop_over_time;
    
    std::atomic<bool>                           m_sbp_rb_needs_reset;
    std::atomic<bool>                           m_iir_filt_need_reset;
    
    unsigned int                                m_accepted_sample_group;
};

} // namespace sbpe
} // namespace cc

#include "SpikeBandPowerEstimator_impl.tpp"

#endif /* end of include guard: SPIKEBANDPOWERESTIMATOR_H_5A7C0342 */
