#ifndef SPIKEBANDPOWERESTIMATOR_IMPL_TPP_4147A372
#define SPIKEBANDPOWERESTIMATOR_IMPL_TPP_4147A372

#include <cmath>

namespace cc
{
namespace sbpe
{

template<typename T, typename U, typename V>
FiltRMSSampler<T,U,V>::FiltRMSSampler (const size_t n_channels)
: m_n_channels(n_channels)
, m_n_samples(600)
, m_sample_count(0)
, m_iir_filter_coeff_b({1})
, m_iir_filter_coeff_a({1})
, m_n_average(1)
, mp_sbp_history(std::make_unique<arma::Mat<V>>(m_n_channels, m_n_average, arma::fill::zeros))
, m_sbp_history_current_item(0)
{
    _resize();
    applyIIRFilterCoeffs();
}

template<typename T, typename U, typename V>
FiltRMSSampler<T,U,V>::FiltRMSSampler (const size_t n_channels, const arma::Col<iir_filter_coeff_type> b, const arma::Col<iir_filter_coeff_type> a)
: m_n_channels(n_channels)
, m_n_samples(600)
, m_sample_count(0)
, m_iir_filter_coeff_b(b)
, m_iir_filter_coeff_a(a)
, m_n_average(1)
, mp_sbp_history(std::make_unique<arma::Mat<V>>(m_n_channels, m_n_average, arma::fill::zeros))
, m_sbp_history_current_item(0)
{
    _resize();
    applyIIRFilterCoeffs();
}

template<typename T, typename U, typename V>
FiltRMSSampler<T,U,V>::FiltRMSSampler (const size_t n_channels, const size_t n_samples, const arma::Col<iir_filter_coeff_type> b, const arma::Col<iir_filter_coeff_type> a)
: m_n_channels(n_channels)
, m_n_samples(n_samples)
, m_sample_count(0)
, m_iir_filter_coeff_b(b)
, m_iir_filter_coeff_a(a)
, m_n_average(1)
, mp_sbp_history(std::make_unique<arma::Mat<V>>(m_n_channels, m_n_average, arma::fill::zeros))
, m_sbp_history_current_item(0)
{
    _resize();
    applyIIRFilterCoeffs();
}


template<typename T, typename U, typename V>
FiltRMSSampler<T,U,V>::~FiltRMSSampler ()
{
}


template<typename T, typename U, typename V>
void FiltRMSSampler<T,U,V>::resize(const size_t n_channels)
{
    m_n_channels = n_channels;
    _resize();
}

template<typename T, typename U, typename V>
void FiltRMSSampler<T,U,V>::setNSamples(const size_t n_samples)
{
    m_n_samples = n_samples;
    zeroBuffers();
}

template<typename T, typename U, typename V>
size_t FiltRMSSampler<T,U,V>::getNSamples() const
{
    return m_n_samples;
}

template<typename T, typename U, typename V>
void FiltRMSSampler<T,U,V>::setNAvgBins(const size_t n_bins)
{
    m_n_average = n_bins;
    m_sbp_history_current_item = 0;
    mp_sbp_history->zeros(m_n_channels, m_n_average);
}

template<typename T, typename U, typename V>
size_t FiltRMSSampler<T,U,V>::getNAvgBins() const
{
    return m_n_average;
}


template<typename T, typename U, typename V>
void FiltRMSSampler<T,U,V>::applyIIRFilterCoeffs()
{
    std::for_each(m_iir_filters.begin(), m_iir_filters.end(), [this](iir_filter_type & iirf) {iirf.set_coeffs(m_iir_filter_coeff_b, m_iir_filter_coeff_a);});
}

template<typename T, typename U, typename V>
void FiltRMSSampler<T,U,V>::setIIRFilterCoeffs(const arma::Col<iir_filter_coeff_type> b, const arma::Col<iir_filter_coeff_type> a)
{
    m_iir_filter_coeff_b = b;
    m_iir_filter_coeff_a = a;
    applyIIRFilterCoeffs();
}

template<typename T, typename U, typename V>
template<typename Iter>
bool FiltRMSSampler<T,U,V>::operator()(Iter _begin, Iter _end)
{
    auto it_iirf = m_iir_filters.begin();
    auto it_ms = m_ms_buffer.begin();
    
    ++ m_sample_count;
    for (; _begin != _end and it_iirf != m_iir_filters.end() and it_ms != m_ms_buffer.end(); ++ _begin, ++ it_iirf, ++ it_ms)
    {
        auto fs = (*it_iirf)(*_begin);
        *it_ms *= ((V)m_sample_count - 1.0)/((V) m_sample_count);
        *it_ms += (1.0 / (V) m_sample_count) * fs * fs;
    }
    if (m_sample_count < m_n_samples)
    {
        return false;
    }
    
    it_ms = m_ms_buffer.begin();
    auto it_out_rms = m_rms_out_buffer.begin();
    
    if (m_n_average > 1)
    {
        auto it_ms_hist = mp_sbp_history->begin_col( m_sbp_history_current_item );
        for (; it_ms_hist != mp_sbp_history->end_col(m_sbp_history_current_item) and it_ms != m_ms_buffer.end(); ++ it_ms, ++ it_ms_hist)
        {
            *it_ms_hist = *it_ms;
            *it_ms = 0;
        }
    
        (++ m_sbp_history_current_item) %= m_n_average;
    
        arma::Mat<V> mean_ms = arma::mean(*mp_sbp_history, 1);
    
        auto it_mean_ms = mean_ms.begin();

        for (; it_out_rms != m_rms_out_buffer.end() and it_mean_ms != mean_ms.end(); ++ it_mean_ms, ++ it_out_rms)
        {
            *it_out_rms = sqrt(*it_mean_ms);
        }
    }
    else
    {
        for (; it_out_rms != m_rms_out_buffer.end() and it_ms != m_ms_buffer.end(); ++ it_ms, ++ it_out_rms)
        {
            *it_out_rms = sqrt(*it_ms);
            *it_ms = 0;
        }
        
    }
    m_sample_count = 0;
    return true;
}


template<typename T, typename U, typename V>
typename FiltRMSSampler<T,U,V>::rms_container_type::const_iterator FiltRMSSampler<T,U,V>::cbegin()
{
    return m_rms_out_buffer.cbegin();
}

template<typename T, typename U, typename V>
typename FiltRMSSampler<T,U,V>::rms_container_type::const_iterator FiltRMSSampler<T,U,V>::cend()
{
    return m_rms_out_buffer.cend();
}

template<typename T, typename U, typename V>
typename FiltRMSSampler<T,U,V>::rms_container_type::iterator FiltRMSSampler<T,U,V>::begin()
{
    return m_rms_out_buffer.begin();
}

template<typename T, typename U, typename V>
typename FiltRMSSampler<T,U,V>::rms_container_type::iterator FiltRMSSampler<T,U,V>::end()
{
    return m_rms_out_buffer.end();
}


template<typename T, typename U, typename V>
void FiltRMSSampler<T,U,V>::_resize()
{
    m_sample_counts.resize(m_n_channels);
    
    m_iir_filters.resize(m_n_channels);
    applyIIRFilterCoeffs();
    
    m_ms_buffer.resize(m_n_channels);
    m_rms_out_buffer.resize(m_n_channels);
    zeroBuffers();
    mp_sbp_history->zeros(m_n_channels, m_n_average);
}

template<typename T, typename U, typename V>
void FiltRMSSampler<T,U,V>::zeroBuffers()
{
    m_sample_count = 0;
    std::fill(m_sample_counts.begin(), m_sample_counts.end(), 0);
    std::fill(m_ms_buffer.begin(), m_ms_buffer.end(), 0);
    std::fill(m_rms_out_buffer.begin(), m_rms_out_buffer.end(), 0);
}

SpikeBandPowerEstimator::size_type SpikeBandPowerEstimator::n_ch() const
{
    return m_ch_list.size();
}

template<typename Iter>
void SpikeBandPowerEstimator::updateChannelPairList(Iter begin, Iter end)
{
    m_ch_pair_v.clear();
    size_t i_in_list = 0;
    std::ostringstream oss;
    
    while (begin != end)
    {
        ch_map_type::iterator it = m_ch_map.find((*begin)); // BlackRock channels are 1 based
        if (it != m_ch_map.end())
        {
            m_ch_pair_v.push_back(std::make_pair(i_in_list, it->second));

            oss << " [ch_id[" << i_in_list << "] = " << *begin << " -> " << it->second << "]";
        }
        else
        {
            oss << " [ch_id[" << i_in_list << "] = " << *begin << " -> none]";
        }
        ++ i_in_list;
        ++ begin;
    }
    mp_logger->debug("updateChannelPairList: {0}", oss.str());
}

} // namespace sbpe
} // namespace cc


#endif /* end of include guard: SPIKEBANDPOWERESTIMATOR_IMPL_TPP_4147A372 */
