/*! \file SpikeRateEstimationMethods.h
    \brief  Free functions to turn binned spike count data into rates
*/
#ifndef SPIKERATEESTIMATIONMETHODS_H_8A98AEF4
#define SPIKERATEESTIMATIONMETHODS_H_8A98AEF4

namespace cc
{
namespace sre
{

/** \brief rate estimation using single last value in bin (identity)
    
    \param  begin_it forward_iterator at start of sequence
    \param  end_it forward_iterator at end of sequence
    \return spike rate
    \sa r_estimate_id_w_cont()
**/
template<class FwdIt, typename T=double>
T r_estimate_id(const FwdIt& begin_it, const FwdIt& end_it);

/** \brief rate estimation using single last value (identity)
    \sa r_estimate_id()
**/
template<class Container, typename T=double>
T r_estimate_id_w_cont(const Container& c);

/** \brief Rate estimation with exponential decay.
    
    Exponential decay. Sums up values in sequence, where the item at the n-th position
    is multiplied with the (n-1)th power of decay_factor. The value is then normalized
    by the sum of all factors. The optional parameter max_length limits the number of
    items taken into account.

    I.e. the spike rate calculates as
        
        \f[r = c_0 + d * c_1 + d^2 * c_2 + ... + d^{n-1} * c_{n-1}\f]
        
    where \f$c_i\f$ are spike counts \f$i\f$ steps in the past, \f$d\f$ is the `decay_factor`,
    and \f$n\f$ is `max_length`. The result is then normalized by \f$\sum_{i=0}^{n-1}d^{i}\f$

    
    \param [in] begin_it forward_iterator at start of sequence
    \param [in] end_it forward_iterator at end of sequence
    \param [in] decay_factor factor whose power will be multiplied with elements in sequence. Should satisfy 0 < decay_factor < 1
    \param max_length maximum length of history to consider for each entry
    \return spike rate
    \sa r_estimate_id_w_cont()
**/
template<class FwdIt, typename T=double>
T r_estimate_with_exp_decay(const FwdIt& begin_it, const FwdIt& end_it, const T& decay_factor, const size_t max_length = 5);

/** \brief spike rate estimation with exponential decay, for a container
    \sa r_estimate_id()
**/
template<class Container, typename T=double>
T r_estimate_with_exp_decay_w_cont(const Container& c, const T& decay_factor, const size_t max_length = 5);

/** \brief Spike rate estimation with boxcar window

    Spike rates will be estimated by averaging over the last length bins.

    \param [in] begin_it forward_iterator at start of sequence
    \param [in] end_it forward_iterator at end of sequence
    \param length number of bins to average over
    \return spike rate
**/
template<class FwdIt, typename T=double>
T r_estimate_with_boxcar(const FwdIt& begin_it, const FwdIt& end_it, const size_t length);

template<class Container, typename T=double>
T r_estimate_with_boxcar_w_cont(const Container& c, const size_t length);

} // namespace sre
} // namespace cc

#include "SpikeRateEstimationMethods_impl.tpp"

#endif /* end of include guard: SPIKERATEESTIMATIONMETHODS_H_8A98AEF4 */
