/*! \file SpikeRateEstimator.h
    \brief  Declares a class to perform spike rate estimation
            SpikeRateEstimator
*/

#ifndef SPIKERATEESTIMATOR_H_C643D24C
#define SPIKERATEESTIMATOR_H_C643D24C
#include <chrono>
#include <cstdio>
#include <memory>
#include <queue>
#include <unordered_map>
#include "CereRingBuffer.h"

#include "spdlog/async.h"
#include "spdlog/sinks/stdout_color_sinks.h"
#include "spdlog/spdlog.h"
#include "folly/ProducerConsumerQueue.h"

#include "rb/RingBuffer.h"
#include "UnitHist.h"
#include "SpikeRateEstimationMethods.h"

namespace cc
{
//! Namespace for Spike Rate Estimation related classes
namespace sre
{
    using namespace std::chrono;
    using namespace std::literals;

struct spike
{
    spike()
    : timestamp(0)
    , channel(0)
    , unit(0)
    {
    }
    
    spike(Timestamp_t t, Channel_t c, Unit_t u)
    : timestamp(t)
    , channel(c)
    , unit(u)
    {
    }
    
    spike(const cbPKT_SPK & pkt_spk)
    : timestamp(pkt_spk.time)
    , channel(pkt_spk.chid)
    , unit(pkt_spk.unit)
    {
    }
    
    Timestamp_t      timestamp;         //!< timestamp of NSP
    Channel_t        channel;           //!< channel id (1-based)
    Unit_t           unit;              //!< unit id (0-based)
};


/** \brief  SpikeRateEstimator
    
    Class collecting spikes (as timestamp, channel, unit), and periodically
    collecting history and performing rate estimation.
    
© Copyright 2018  - Jonas B. Zimmermann. All Rights Reserved.

    \author Jonas B. Zimmermann 
    \date 2018-10-25
**/
class SpikeRateEstimator
{
#ifdef CC_TESTING_DEBUG
    friend class SpikeRateEstimatorBase_Create_Test;
    friend class SpikeRateEstimatorBasic_NormFactor_Test;
    friend class SpikeRateEstimatorBasic_AddSpikesToQueue_Test;
    friend class SpikeRateEstimatorBasic_ChannelUnitArithmetic_Test;
    friend class SpikeRateEstimatorBasic_UnitHistSize_Test;
    friend class SpikeRateEstimatorBasic_ApplyCallbackWithoutConfig_Test;
    friend class SpikeRateEstimatorBasic_ConsumeQueue_Test;
    friend class SpikeRateEstimatorBasic_SetChannelUnits_Test;
    friend class SpikeRateEstimatorBasic_GetChannelUnits_Test;
    friend class SpikeRateEstimatorBasic_ResizeQueue_Test;
    friend class SpikeRateEstimatorBasic_ValidateSpikes_Test;
    friend class SpikeRateEstimatorBasic_ValidateSpikeRate_Test;
    friend class SpikeRateEstimatorBasic_TimestampCallbackNotSet_Test;
    friend class SpikeRateEstimatorBasic_TimestampCallback_Test;
#endif
private:
    // based on https://stackoverflow.com/a/38253266/1893154 
    using maxres_sys_or_steady =
        std::conditional<
            system_clock::period::den <= steady_clock::period::den,
            system_clock, steady_clock
          >::type;
public:
    using clock =
        std::conditional<
            high_resolution_clock::is_steady,
            high_resolution_clock, maxres_sys_or_steady
          >::type;                                                              //!< clock to be used for thread timing
    
    
public:
    using unit_hist_type = UnitHist<unsigned int>;                              //!< type to store a unit's spike count history in
    using unit_hist_container_type = std::vector<unit_hist_type>;               //!< container for `unit_hist_type`s, to store many units
    using size_type = unit_hist_container_type::size_type;                      //!< size type to count elements with
    
    using timestamp_callback_type = std::function<Timestamp_t()>;               //!< type for callbacks to get a current time stamp
    
    using rate_type = double;                                                   //!< what type should spike rates be stored as
    using rate_item_type = std::vector<rate_type>;                              //!< type to capture spike rates for all units at one time point
    using rate_ring_buffer_item_type = std::pair<Timestamp_t, rate_item_type>;  //!< type to store rates together with a timestamp in a ring buffer
    using rate_ring_buffer_type = RingBufferT<rate_ring_buffer_item_type, 1000>;//!< type of ring buffer to store rates
    using rate_list_type = rate_ring_buffer_type::VecT;                         //!< type to store rates as read out from ring buffer
    using rate_interval_type = std::chrono::milliseconds;                       //!< intervals for spike rate loops given in this unit
    using rate_callback_type = std::function<rate_type(const unit_hist_type&)>; //!< type for callbacks to calculate spike rates

    using spike_queue_type = folly::ProducerConsumerQueue<spike>;               //!< type to cache spikes before processing
    
    using ch_un_type = std::pair<size_type, size_type>;                         //!< describe a channel-unit pair
    using ch_un_list_type = std::vector<ch_un_type>;                            //!< a list of channel-unit pairs
    
private:
    struct ChUnHash
    {
        std::size_t operator()(const ch_un_type & p) const noexcept
        {
            std::size_t h1 = std::hash<size_type>{}(p.first);
            std::size_t h2 = std::hash<size_type>{}(p.second);
            return h1 ^ (h2 << 1);
        }
    };
public:
    using ch_un_map_type = std::unordered_map<ch_un_type, size_type, ChUnHash>; //!< a map of ch_un_type to indices in rate vector
    
    SpikeRateEstimator();
    SpikeRateEstimator(size_type n_channels,
        size_type n_units,
        size_type history_length                    = 2'000,
        rate_interval_type t_spike_loop             = 10ms);
    SpikeRateEstimator(ch_un_list_type ch_un_list,
        size_type history_length                    = 2'000,
        rate_interval_type t_spike_loop             = 10ms);
    

    virtual ~SpikeRateEstimator ();

    /** \brief adds a spike to the waiting queue
        
        This function adds a spike that is compatible with respect to channel and
        unit number to the waiting queue.
        Incompatible spikes will not be added.
        
        \param  spk spike object
        \return `true` if spike was added, `false` otherwise
        \sa
    **/
    template<typename T=spike>
    bool addSpike(T &&spk);

    /** \brief adds a spike to the waiting queue
    
        This function adds a spike that is compatible with respect to channel and
        unit number to the waiting queue.
        Incompatible spikes will not be added.
    
        \param  spk spike object
        \return `true` if spike was added, `false` otherwise
        \sa
    **/
    template<typename T=spike>
    bool addSpike(const T& spk);
    
    
    /** \brief adds a spike to the waiting queue

        This function adds a spike that is compatible with respect to channel and
        unit number to the waiting queue.
        Incompatible spikes will not be added.

        \param  channel channel number
        \param  unit unit number
        \return `true` if spike was added, `false` otherwise
        \sa
    **/
    bool addSpike(const Channel_t channel, const Unit_t unit);

    /** \brief bind a call back function used to calculate spike rates
        
        The callback takes a vector of spike counts and turns them into
        spike rates. See SpikeRateEstimationMethods.h for examples.
        
        \param  rcb a callable object
        \sa
    **/
    void bindCallback(const rate_callback_type& rcb);

    /** \brief pass a pointer to a SpikeRateEstimator::rate_ring_buffer_type to store spike rates in
        
        Spike rates will only be calculated if SpikeRateEstimator has
        a valid pointer to a SpikeRateEstimator::rate_ring_buffer_type.
        
        \param sr_rb shared pointer to SpikeRateEstimator::rate_ring_buffer_type to be used to store spike rates
        \sa
    **/
    void attachSpikeRateRB(std::shared_ptr<rate_ring_buffer_type> sr_rb);

    /** \brief Start processing thread
        
        If a processing thread is already running, it will be stopped.
        Spikes in queue as well as spike rate ringbuffer will be cleared.
        
        \sa
    **/
    void signalStart();
    
    /** \brief Send processing thread a 'stop' signal
        
        \sa
    **/
    void signalStop();

    /** \brief enable or disable spike rate normalization
        
        If enabled, spike rates will be normalized to spikes
        per second. Otherwise, rates will be based on counts/bin
        
        \param  enabled state of spike rate normalization
        \sa
    **/
    void setSpikeRateNormalization(bool enabled);

    /** \brief get state of spike rate normalization
        
        \return `true` if spike rate normalization is enabled, `false` otherwise
        \sa
    **/
    bool getSpikeRateNormalization() const;

    void set_logger(std::shared_ptr<spdlog::logger> logger);
    
    /** \brief Set a list of channel-unit pairs for which to calculate spike rates
        
        Only spikes on channel-unit pairs in the list passed to this function will
        be considered for spike rate calculation.
        Note that Channel IDs are 1-based (in line with Blackrock usage)
        and Units in SpikeRateEstimator::ch_un_list_type have 0-based indices.
        
        \param  ch_un_list a vector of channel-unit pairs
        \sa fillChannelUnitList(), getChannelUnitMapping(), getChannelUnitList()
    **/
    void setChannelUnitList(const ch_un_list_type & ch_un_list);
    
    /** \brief Enable spike rate calculation for all channel-unit combinations
        
        Sets all channel-unit combinations for spike rate estimation.
        I.e. channels-units (1, 0), (1, 1), ..., (1, n_units - 1), (2, 0),
        (2, 1), (2, n_units - 1), ..., (n_channels, n_units - 1)
        will be added.
        
        \param  n_channels Number of channels, starting from 1, to be included in spike rate estimation
        \param  n_units Number of units per channel, starting from 0
        \sa setChannelUnitList(const ch_un_list_type & ch_un_list), getChannelUnitList(), getChannelUnitMapping()
    **/
    void fillChannelUnitList(const size_type n_channels, const size_type n_units);
    
    
    /** \brief return the current mapping of channel-units to spike rates
        
        The current mapping of channel-units to the elements of the spike
        rates may not be ordered. This function provides both a look-up table
        for channel-units to indices in the spike rate vector, as well as a
        list of channel-units in the same order as in the rates.
        
        \return a std::pair of SpikeRateEstimator::ch_un_list_type and SpikeRateEstimator::ch_un_map_type.
        \sa getChannelUnitList(), setChannelUnitList(), fillChannelUnitList()
    **/
    std::pair<ch_un_list_type, ch_un_map_type> getChannelUnitMapping() const;
    
    /** \brief get a list of current channel-unit pairs for which spike rates are computed
        
        The returned list is in the same order as the rates for these channel-unit pairs
        as stored in the spike rate buffer.
        
        \return the current channel-unit list
        \sa getChannelUnitMapping(), setChannelUnitList(), fillChannelUnitList()
    **/
    ch_un_list_type getChannelUnitList() const;

    /** \brief set interval for repeated calculation of spike rates
        
        The internal processing thread will aim to keep this interval as tightly as possible.
        
        \param new_interval an interval between repeated starts of the spike processing function
        \sa getSpikeRateLoopInterval()
    **/
    void setSpikeRateLoopInterval(const rate_interval_type & new_interval);
    
    /** \brief get current spike rate processing interval
        
        \return current spike rate processing interval
        \sa setSpikeRateLoopInterval()
    **/
    rate_interval_type getSpikeRateLoopInterval() const;
    
    /** \brief bind a callback to retrieve a timestamp
        
        this is used to indicate the last timestamp until spikes could
        have possibly been included. This should be called by client
        code (e.g. in cc::CereConn), to enable link to a data acquisition system.
        
        \param tscb callback to return a cc::Timestamp_t
        \sa
    **/
    void bindTimestampCallback(const timestamp_callback_type & tscb);
    
    
    /** \brief checks validity of ring buffer
        
        \return false, if buffer is valid
    **/
    bool ringBufferInvalid() const;
    
private:
    /** \brief validate a spike
    
        checks if the spike object conforms to capacity of the SpikeRateEstimator
    
        \param  spk spike object to be checked
        \return `true` if spike is compatible with this object, `false` otherwise.
    **/
    template<typename T=spike>
    inline bool validate_spk(const T& spk);
    
    /** \brief process spikes in queue and calculate rates
    
        This function is to be called periodically to read in the spike
        queue, update the unit history, and to calculate spike rates.
    **/
    void processSpikes();
    
    size_type consumeQueue();
    void applyCallback();

    void stopProcessThread();
    void processThreadFun();

    inline bool ch_u_offset(size_type ch, size_type u, size_type & unit_index);
    inline size_type n_ch_u();

    void init_logger();
    
    void applyChannelUnitList();
    

    void updateRateNormalizationFactor();
    void assignSpikeQueue();
    
private:
    ch_un_list_type                             m_ch_un_list;
    ch_un_map_type                              m_ch_un_map;
    const size_type                             m_history_length;
    
    rate_interval_type                          m_t_spike_loop;
    rate_type                                   m_rate_norm_factor;
    std::atomic<bool>                           m_normalize_rates;
        
    std::unique_ptr<spike_queue_type>           mp_sq;
    std::unique_ptr<unit_hist_container_type>   mp_unit_hist_container;

    rate_callback_type                          m_rate_callback;
    timestamp_callback_type                     m_timestamp_callback;

    std::shared_ptr<rate_ring_buffer_type>      mp_spkr_rb; //!< stores a pointer to a spike rate ring buffer
    
    std::shared_ptr<spdlog::logger>             mp_logger;
    
    std::thread                                 m_process_thread;
    std::atomic<bool>                           m_keep_going;
    std::atomic<bool>                           m_thread_loop_over_time_flag;
    clock::duration                             m_thread_loop_over_time;
    
    std::atomic<bool>                           m_spkr_rb_needs_reset;
};

} // namespace sre
} // namespace cc

#include "SpikeRateEstimator_impl.tpp"

#endif /* end of include guard: SPIKERATEESTIMATOR_H_C643D24C */
