/*! \file UnitHist.h
    \brief Declaration of UnitHist class
    
*/

#ifndef UNITHIST_H_AB7D64A8
#define UNITHIST_H_AB7D64A8

#include <algorithm>
#include <iterator>
#include <stdexcept>
#include <vector>

#include <cstdio>
#include "spdlog/fmt/ostr.h" // must be included


namespace cc
{
namespace sre
{

/** \brief  UnitHist
    
    Maintains a history of values in a circular container (std::vector by default).

    
© Copyright 2018  - Jonas B. Zimmermann. All Rights Reserved.

**/
template <typename T,
          class Container = std::vector<T>>
class UnitHist
{
public:
    using value_type            = typename Container::value_type;
    using reference             = typename Container::reference;
    using const_reference       = typename Container::const_reference;
    using size_type             = typename Container::size_type;
    using container_type        = Container;
    using container_iterator    = typename Container::iterator;
    
    class const_iterator
    {
        friend class UnitHist;
    public:
        using difference_type   = typename Container::difference_type;
        using value_type        = typename Container::value_type;
        using reference         = typename UnitHist::const_reference;
        using pointer           = const typename Container::pointer;
        using iterator_category = std::bidirectional_iterator_tag;
        using container_iterator        = typename Container::iterator;
        using container_const_iterator  = typename Container::const_iterator;

        const_iterator () = default;
        const_iterator (const const_iterator& other);
        ~const_iterator();
    private:
        const_iterator(const UnitHist& cont, container_const_iterator start);
        const_iterator(const UnitHist& cont, container_const_iterator start, bool iterated);
    public:
        const_iterator& operator=(const const_iterator& other);

        bool operator==(const const_iterator&) const;
        bool operator!=(const const_iterator&) const;

        const_iterator& operator++();
        const_iterator operator++(int);
        const_iterator& operator--();
        const_iterator operator--(int);

        const_iterator& operator+=(difference_type);
        const_iterator operator+(difference_type) const;

        friend const_iterator operator+ (difference_type add, const const_iterator& rhs)
        {
            auto it(rhs);
            return it + add;
        }

        const_iterator& operator-=(difference_type);
        const_iterator operator-(difference_type) const;
        difference_type operator-(const_iterator) const;

        reference operator*() const;

#ifdef CC_TESTING_DEBUG
        void print_debug();
#endif
        
    private:
        bool compatible(const const_iterator& other) const;

        const UnitHist & m_cont;
        container_const_iterator m_pos;
        bool m_iterated;
    };
    using const_reverse_iterator = std::reverse_iterator<const_iterator>;

public:
    /** \brief creates a UnitHist with capacity for c_size elements
        
        \param  c_size capacity of UnitHist
        \return UnitHist
        \sa
    **/
    UnitHist (const size_type c_size);

    /** \brief creates a UnitHist by copying elements from container
            
        The resulting UnitHist will have the same size as c,
        the first element of c will be the oldest one in UnitHist.
            
        \param  c container whose elements will be copied
        \return UnitHist
    **/
    UnitHist(const Container& c);

    /** \brief creates a UnitHist by moving elements from container
        
        The resulting UnitHist will have the same size as c,
        the first element of c will be the oldest one in UnitHist.
        
        \param  c container whose elements will be moved
        \return UnitHist
    **/
    explicit UnitHist(Container&& c = Container());

    /** \brief creates a UnitHist copying another UnitHist
        
        
        \param  other UnitHist to be copied
        \return UnitHist
    **/
    UnitHist(const UnitHist& other);

    /** \brief creates a UnitHist moving another UnitHist
    
        \param  other UnitHist to be moved
        \return UnitHist
    **/
    UnitHist(UnitHist&& other);

    UnitHist& operator=(const UnitHist&) = delete;
    UnitHist& operator=(UnitHist&&) = delete;

    bool operator==(const UnitHist&) const;
    bool operator!=(const UnitHist&) const;

    /** \brief increments current cell
        
        The value at the current cell will be incremented.
        
        \return a reference to the object
        \sa add_to_current(), push()
    **/
    inline UnitHist& operator++();
    
    size_type size() const;
    
    /** \brief Set cell to value and advance
        
        Sets the current cell to value x and advances current cell by one.
        
        \param x value to set cell to
        \sa push()
    **/
    void push(const value_type& x);

    /** \brief Set cell to value and advance
    
        Sets the current cell to value x and advances current cell by one.
    
        \param  x value to set cell to
        \sa push()
    **/
    void push(value_type&& x);

    UnitHist& operator<<(const value_type & x);
    UnitHist& operator<<(value_type&& x);
    
    /** \brief steps insertion pointer
        
        Steps the insertion pointer. Clears the newly pointed at cell.
        
        \sa push(), operator++()
    **/
    void step();

    /** \brief Adds value to current cell
        
        Adds value to current cell, without overwriting or stepping.
        
        \param x templated value to be added
        \sa operator++()
    **/
    void add_to_current(const value_type& x);
    
    /** \brief Adds value to current cell
    
        Adds value to current cell, without overwriting or stepping.
    
        \author Jonas B. Zimmermann 
        \date 2018-10-25
        \param x templated value to be added
        \sa operator++()
    **/
    void add_to_current(value_type&& x);
    
    /** \brief returns true if there are any events counted during the history
        
        \return true if there are events in history
    **/
    bool has_spikes() const;

    /** \brief gets const_iterator pointing to the first recorded event in history
        
        \return const_iterator pointing to the first recorded event in history
        \sa cbegin(), end()
    **/
    const_iterator begin() const;

    /** \brief gets const_iterator pointing to the first recorded event in history
    
        \return const_iterator pointing to the first recorded event in history
        \sa begin(), cend()
    **/
    const_iterator cbegin() const;

    /** \brief gets const_iterator pointing to cell one after the last recorded event in history
    
        \return const_iterator pointing to cell one after the last recorded event in history
        \sa begin(), cend()
    **/
    const_iterator end() const;

    /** \brief gets const_iterator pointing to cell one after the last recorded event in history

        \return const_iterator pointing to cell one after the last recorded event in history
        \sa cbegin(), end()
    **/
    const_iterator cend() const;
    
    /** \brief gets const_reverse_iterator pointing to the last recorded event in history

        \return const_iterator pointing to cell of the last recorded event in history
        \sa crbegin(), rend()
    **/
    const_reverse_iterator rbegin() const;

    /** \brief gets const_reverse_iterator pointing to the last recorded event in history

        \return const_iterator pointing to cell of the last recorded event in history
        \sa crend(), rbegin()
    **/
    const_reverse_iterator crbegin() const;

    /** \brief gets const_reverse_iterator pointing to one cell before the first recorded event in history

        \return const_iterator pointing to one cell before the first recorded event in history
        \sa rbegin(), crend()
    **/
    const_reverse_iterator rend() const;

    /** \brief gets const_reverse_iterator pointing to one cell before the first recorded event in history

        \return const_iterator pointing to one cell before the first recorded event in history
        \sa rend(), cbegin()
    **/
    const_reverse_iterator crend() const;

#ifdef CC_TESTING_DEBUG
    void print_debug() const;
#endif
private:
    void _step();
protected:
    const size_type m_size;
    const size_type m_container_size;
    Container m_c;
    container_iterator m_c_it;
    size_type m_zero_spikes;
};


template <typename T,
          class Container>
std::ostream& operator<<(std::ostream& os, const UnitHist<T, Container>& c);

} // namespace sre
} // namespace cc

#include "UnitHist_impl.tpp"

#endif /* end of include guard: UNITHIST_H_AB7D64A8 */
