# distutils: language = c++
# cython: language_level=3
##@file cere_conn.pyx
# @package cere_conn Python glue for cc::CereConn
# 
# User facing is mostly the definition of the class CereConn,
# which wraps an instance of cc::CereConn. In addition, some
# enumerations used by members of CereConn are defined.

"""@package cere_conn – Python glue for cc::CereConn

User facing is mostly the definition of the class CereConn,
which wraps an instance of cc::CereConn. In addition, some
enumerations used by members of CereConn are defined.

"""
include "cere_conn_version.pxi"
##@cond
from libc.stdint cimport uint8_t, int16_t, uint32_t
from libcpp.string cimport string
from libcpp cimport bool
from libcpp.vector cimport vector
from libcpp.pair cimport pair
cimport cython
from cython.operator cimport dereference as deref

cimport cere_conn_cython  as cc

from cere_conn_cython cimport SpikeRateEstimator as sre
from cere_conn_cython cimport SpikeBandPowerEstimator as sbpe

import numpy as np
cimport numpy as np

cimport arma_cy as ac

cpdef enum Instruction:
    ccI_Invalid         = -1,
    ccI_Open,
    ccI_Record,
    ccI_Idle,
    ccI_Close

cpdef enum State:
    ccS_Invalid         = -1,
    ccS_Closed,
    ccS_Idle,
    ccS_Recording

cpdef enum ReturnCode:
    ccR_Error         = -1,
    ccR_Success,
    ccR_NotRecording,
    ccR_NoValidTimestamp,
    ccR_InvalidSpikeRateBuffer,
    ccR_InvalidSpikeRateEstimator,
    ccR_CbSdkError,
    ccR_InvalidSpikeBandPowerBuffer,
    ccR_InvalidSpikeBandPowerEstimator

##@endcond

@cython.embedsignature(True)
cdef class CereConn:
    """!This is a wrapper for the cc::CereConn C++ class.
    
    General operation:
        
        ck = CereConn()
        ck.send_open()
        ck.send_record()
        
    and later, call periodically:
    
        ck.get_cont_data()
        ck.get_comment_data()
        ck.get_spike_rate_data()
        ck.get_spike_band_power_data()
        ck.get_digin_ev_data()
    
    etc.
    
    Note that by default, continuous channels, Spike Rate Estimation (SRE), and Spike Band Power Estimation (SBPE)
    are disabled. SRE and SBPE can be enabled when constructing the CereConn object, by passing
    `ck = CereConn(withSRE=True)` and `ck = CereConn(withSBPE=True)`, respectively (of course, these can eb combined).
    In order to enable collection of continuous data, use `set_sample_group_state()` or `set_sample_group_states()`.
    
    Note that sample groups are 1-based (1-6, usually), as are channel IDs.
    """
    ##@cond
    cdef cc.CereConn *thisobj
    cdef cc.SampleGroupList *sgl
    cdef cc.CommentList *cl
    cdef cc.DigInEvList *diel
    cdef sre.rate_list_type *rateList
    cdef sbpe.sbp_list_type *sbpList
    cdef public kwargs
    ##@endcond

    @cython.embedsignature
    def __cinit__(self, int instance=0, bool withSRE=False, bool withSBPE=False, int sbpeGroup=6, int sbpePeriod=20, **kwargs):
        """!@brief Sets up a CereConn wrapper object
        @param instance the NSP instrument instance, usually 0
        @param withSRE if `True`, Spike Rate Estimation will be automatically started when send_record() is called
        @param withSBPE if `True`, Spike Band Power Estimator will be automatically started when send_record() is called
        @param sbpeGroup accepted sampling group of Spike Band Power Estimator. Only has an effect if `withSBPE` is `True`.
            This can be later changed with set_spike_band_power_estimator_use_sample_group()
            (see also cc::CereConn::setSpikeBandPowerEstimatorAcceptedSampleGroup())
        @param sbpePeriod sets both SBPE loop interval and number of samples included in RMS. Only has an effect if `withSBPE` is `True`.
            This can be later changed with set_spike_band_power_estimator_loop_interval_ms() and
            set_spike_band_power_estimator_integrated_samples()
            (see also cc::CereConn::setSpikeBandPowerEstimatorLoopIntervalMS(), cc::CereConn::setSpikeBandPowerEstimatorIIRFilterNSamples())
        
        """
        self.kwargs = kwargs
        self.thisobj = new cc.CereConn(instance, withSRE, withSBPE)
        self.sgl = new cc.SampleGroupList()
        self.cl = new cc.CommentList()
        self.cl.reserve(100)
        self.diel = new cc.DigInEvList()
        self.diel.reserve(100)
        
        self.rateList = new sre.rate_list_type()
        self.rateList.reserve(100)
        self.sbpList = new sbpe.sbp_list_type()
        self.sbpList.reserve(100)
        self.kwargs['instance'] = instance
        self.kwargs['withSRE'] = withSRE
        self.kwargs['withSBPE'] = withSBPE
        self.kwargs['sbpeGroup'] = sbpeGroup
        self.kwargs['sbpePeriod'] = sbpePeriod
        if withSBPE:
            self.thisobj.setSpikeBandPowerEstimatorAcceptedSampleGroup(sbpeGroup)
            sbpeSFreq = self.thisobj.getSampleGroupFrequency(self.thisobj.getSpikeBandPowerEstimatorAcceptedSampleGroup())
            sbpeNPer = int(sbpeSFreq * sbpePeriod / 1000)
            self.thisobj.setSpikeBandPowerEstimatorLoopIntervalMS(sbpePeriod)
            self.thisobj.setSpikeBandPowerEstimatorIIRFilterNSamples(sbpeNPer)
            
    def __init__(self, instance=0, withSRE=False, withSBPE=False, sbpeGroup=6, sbpePeriod=20, **kwargs):
        pass

    def __dealloc___(self):
        del self.thisobj
        del self.sgl
        del self.cl
        del self.rateList
        del self.sbpList

    ## @name General Operations
    ##@{
    def send_open(self):
        """!@brief Opens connection to NSP via the cbSDK library

        After this operation, which may take up to a second to complete,
        CereConn will be in the Idle state.
        
        This operation is non-blocking.
        """
        self.thisobj.ccOpen()

    def send_idle(self):
        """!Sets CereConn to Idle state
        
        Recording is stopped, but connection with NSP via cbSDK remains open.
        
        This operation is non-blocking.
        """
        self.thisobj.ccIdle()

    def send_record(self):
        """!Starts recording
        
        Spike band power and spike rate estimation processes are started, continuous,
        spike, event, and comment data are stored in their respective buffers for
        periodic retrieval.
        
        This operation is non-blocking.
        """
        self.thisobj.ccRecord()

    def send_close(self):
        """!Closes connection to NSP and cbSDK"""
        self.thisobj.ccClose()

    def print_buffer_state(self):
        """!Calls cc::CereConn::printBufferState()"""
        self.thisobj.printBufferState()

    def get_state(self):
        """!Return current state of CereConn.
        
        @return One of:
        
            ccS_Invalid,
            ccS_Closed,
            ccS_Idle,
            ccS_Recording
        """
        return <State> self.thisobj.getState()

    def get_cb_time(self):
        """!Get current NSP time (timestamp of latest packet received by cbSDK)
    
        @return dictionary, with keys:
                - `ts`: NSP timestamp
                - `rc`: return code
        """
        cdef uint32_t cbTime = 0
        rc = self.thisobj.getCbTime(cbTime)
        outdata = {'ts': cbTime, 'rc': <ReturnCode> rc}
    
        return outdata
    ##@}
    
    ## @name Continuous and event data
    ##@{
    def get_cont_data(self, sample_groups=None):
        """!Get continuous data from buffers
        
        Read out sample group buffers, i.e. continuous data as transmitted from NSP.
        
        @param sample_groups a list of sample groups, as numbers
                            - 1: 500Hz
                            - 2: 1kHz
                            - 3: 2kHz
                            - 4: 10kHz
                            - 5: 30kHz
                            - 6: 30kHz (raw data)
        @return a dictionary with the following keys:
            - `ts`: NSP timestamp at time this function was called
            - `sample_groups`: Dictionary, containing data for the sample groups, i.e.
              - `ch`:   list of channels as numpy.array
              - `data`: an samples-by-channels numpy.int16 matrix
              - `last_time`: the NSP timestamp of the last sample (only set if there are samples)
              - 'oc': overflow count since last call (i.e. how many samples have been lost
                because buffer overflowed)
            - `rc`:  return code
                
        """
        cdef uint32_t cbTime = 0
        cdef size_t overflow_count = 0
        cdef Py_ssize_t n_channels, num_samples, n_s_p_c
        cdef double time_covered
        cdef int sg
        
        if sample_groups is None:
            sample_groups = range(1, cc.ccon.NSampleGroups + 1)

        # fetch new continuous data
        rc = self.thisobj.getSampleGroupData(self.sgl[0], cbTime)
        cdef dict outdata = {'ts': cbTime, 'sample_groups': {}, 'rc': <ReturnCode>rc}

        cdef int16_t *vec_ptr
        cdef int16_t[::1] vec_view

        cdef uint32_t *ch_vec_ptr
        cdef uint32_t[::1] ch_vec_view

        for sg_i in sample_groups:
            sg = sg_i - 1
            n_channels = self.sgl.sgc[sg].channel_ids.size()
            num_samples = self.sgl.sgc[sg].buffer.size()
            last_time = self.sgl.sgc[sg].last_timestamp
            
            n_s_p_c = int(num_samples / (n_channels if n_channels else 1))
            time_covered = self.sgl.sgc[sg].sample_period / 30.0 * n_s_p_c;

            if n_channels == 0:
                outdata['sample_groups'][sg_i] = {'ch': [], 'data': np.empty((0, 0), dtype=np.int16, order='C'), 'last_time':last_time, 'oc':0}
                continue

            ch_vec_ptr = self.sgl.sgc[sg].channel_ids.data()
            ch_vec_view = <uint32_t[:n_channels]>ch_vec_ptr
            channel_ids = np.asarray(ch_vec_view)

            if num_samples == 0:
                outdata['sample_groups'][sg_i] = {'ch': channel_ids, 'data': np.empty((0, n_channels), dtype=np.int16, order='C'), 'last_time': last_time, 'oc':0}
                continue

            vec_ptr = self.sgl.sgc[sg].buffer.data()
            vec_view = <int16_t[:num_samples]>vec_ptr

            sample_array = np.asarray(vec_view)
            sample_array = np.reshape(sample_array, (-1, n_channels))

            outdata['sample_groups'][sg_i] = {'ch': channel_ids, 'data':sample_array, 'last_time':last_time, 'oc':self.sgl.sgc[sg].overflow_count}

        return outdata

    def send_comment(self, comment, uint32_t rgba=0, uint8_t charset=0):
        """!Send a text comment to NSP
        
        @param comment a string
        @param rgba a `uint32_t`, used to represent a colour or arbitrary data
        @param charset charset of string; reserved
        @return dictionary with key
            - `rc`: return code
        """
        cdef string cpp_comment = comment.encode('UTF-8')
        rc = self.thisobj.sendComment(cpp_comment, <uint32_t> rgba, <uint8_t> charset)
        return {'rc': <ReturnCode> rc}

    def get_comment_data(self):
        """!Retrieve comments from buffer
        
        @return dictionary with keys:
            - `ts`: NSP timestamp at the time this function was called
            - `rc`: return code
            - 'oc': overflow count since last call (i.e. how many comments
                have been lost because buffer overflowed)
            - `comments`: list of dictionaries with keys:
              - `ts`: NSP timestamp of comment
              - `flag`: a flag
              - `data`: uint32_t data associated with comment
              - `text`: the comment's text.
        """
        cdef uint32_t cbTime = 0
        cdef size_t overflow_count = 0
        rc = self.thisobj.getCommentData(self.cl[0], cbTime, overflow_count)
        outdata = {'ts': cbTime, 'comments': [], 'rc': <ReturnCode> rc, 'oc':overflow_count}
        for c in self.cl[0]:
            outdata['comments'].append({'ts':c.timestamp, 'flag':c.flag, 'data':c.data, 'text':c.comment})

        return outdata

    def get_digin_ev_data(self):
        """!Request buffered digital input and serial events.
        
        @return a dictionary containing the following fields:
          - `ts`: timestamp of current NSP time at the time the requested data
                  were retrieved from buffers
          - `rc`: return code
          - 'oc': overflow count since last call (i.e. how many digital events
                have been lost because buffer overflowed)

          - `digin_events`: list of dictionaries representing individual events,
                          each having the following structure:
            - `ts`:       NSP time when event was received
            - `source`:   'DigIn' if signal was received from front plate DB37 conenctor
                          'Serial' if signal came
                          'Unknown' if signal source is unknown (indicates unexpected condition)
            - `data`:     uint32_t value associated with event. For Serial data, only the
                          lowest byte is used. For DigIn, it may be a 8 or 16 bit value,
                          depending on the settings in Central.
        """
    
        cdef uint32_t cbTime = 0
        cdef size_t overflow_count = 0
        rc = self.thisobj.getDigInEvData(self.diel[0], cbTime, overflow_count)
        outdata = {'ts': cbTime, 'digin_events': [], 'rc': <ReturnCode> rc, 'oc':overflow_count}
        for die in self.diel[0]:
            this_source = "Unknown"
            if <cc.Source>die.source == cc.Source.dies_DigIn:
                this_source = "DigIn"
            elif <cc.Source>die.source == cc.Source.dies_Serial:
                this_source = "Serial"
            outdata['digin_events'].append({'ts':die.timestamp, 'source':this_source, 'data':die.data})

        return outdata

    def set_sample_group_state(self, group, enable=True):
        """!Enable or disable processing of a sample group
        
        @param group one of the sample groups, as integer (1-6, usually). See get_cont_data()
        @param enable set `True` if sample group should be processed
        @sa cc::CereConn::setSampleGroup(), get_cont_data(), get_sample_group_state()
        """
        cdef bool new_state = enable
        cdef size_t sgroup = group
        self.thisobj.setSampleGroup(sgroup, new_state)

    def get_sample_group_state(self, group):
        """!Get processing state of a sample group
    
        @param group one of the sample groups, as integer (1-6, usually). See get_cont_data()
        @return `True` if sample group is being processed
        @sa cc::CereConn::getSampleGroup(), get_cont_data(), set_sample_group_state()
        """
        cdef size_t sgroup = group
        cdef bool state = self.thisobj.getSampleGroup(sgroup)
        return state

    def set_sample_group_states(self, groups=None):
        """!Enable or disable processing of sample groups
    
        @param groups list of booleans, one per sample group. If less than cc::constants::NSampleGroups, the state
            of the remaining groups will not be changed. If more than cc::constants::NSampleGroups values, those
            will be ignored.
        @return `True` on success
        @sa cc::CereConn::setSampleGroup(), get_cont_data(), get_sample_group_state(), get_sample_group_states(), set_sample_group_state()
        """
        if groups is None:
            return False
        cdef bool new_state
        cdef size_t sgroup
        for gixx, state in enumerate(groups):
            new_state = state
            sgroup = gixx + 1
            if sgroup > cc.ccon.NSampleGroups:
                break
            self.thisobj.setSampleGroup(sgroup, new_state)
        return True

    def get_sample_group_states(self):
        """!Get processing state of all sample groups

        @return List of cc::constants::NSampleGroups booleans, where `True` means that sample group is being processed.
        @sa cc::CereConn::getSampleGroup(), get_cont_data(), set_sample_group_state() set_sample_group_states(), get_sample_group_state()
        """
        states = []
        cdef size_t sgroup
        cdef bool state
        for gid in range(1, cc.ccon.NSampleGroups + 1):
            sgroup = gid
            state = self.thisobj.getSampleGroup(sgroup)
            states.append(state)
        return states

    def set_car_channels(self, size_t group, cc.ChannelNumList_t channel_list):
        """!Set Common Average Referencing channels per sample group
        @param group sample group (1-based)
        @param channel_list list of channel IDs whose average will be subtracted from each channel's
            sample in this group
        @return `True` if CAR was set up successfully
        @sa cc::SampleGroupRingBufferT::setCARChannels(), disable_car()
        """
        return self.thisobj.setCARChannels(group, channel_list)
    
    def disable_car(self, size_t group):
        
        """!Disable Common Average Referencing for sample group
        @param group sample group (1-based)
        @sa cc::SampleGroupRingBufferT::disableCAR(), set_car_channels()
        """
        self.thisobj.disableCAR(group)

    ##@}
    
    ## @name Spike Rate Estimator related functions
    ##@{
    def start_spike_rate_estimator(self):
        """!Start spike rate estimation processing in separate thread
        @sa cc::CereConn::spikeRateEstimatorStart()
        """
        self.thisobj.spikeRateEstimatorStart()

    def stop_spike_rate_estimator(self):
        """!Stop spike rate estimation processing in separate thread
        @sa cc::CereConn::spikeRateEstimatorStop()"""
        self.thisobj.spikeRateEstimatorStop()
        
    def set_spike_rate_estimator_ch_u_list(self, sre.ch_un_list_type ch_u_list):
        """!Set channel-units for spike rate estimation
        
        @param ch_u_list list of pairs, where the first element of a pair
            represents the (1-based) channel id, the second element represents
            the unit (0 for unsorted, 1 for Unit 1, etc.)
        @sa cc::CereConn::setSpikeRateEstimatorChannelUnitList()
        """
        self.thisobj.setSpikeRateEstimatorChannelUnitList(ch_u_list)

    def fill_spike_rate_estimator_ch_u_list(self, int n_ch, int n_u):
        """!Enable all channel-unit combinations for spike rate estimation
        
        Channels-units (1, 0), (1, 1), ..., (1, n_u - 1), (2, 0), (2, 1), (2, n_u - 1), ..., (n_ch, n_u - 1)
        will be enabled.
        
        @param n_ch number of channels
        @param n_u number of units
        @sa cc::CereConn::fillSpikeRateEstimatorChannelUnitList()
        """
        self.thisobj.fillSpikeRateEstimatorChannelUnitList(n_ch, n_u)

    def get_spike_rate_estimator_ch_u_map(self):
        """!Get mapping of channel-units to spike rate values
        
        @return a dictionary with keys:
            - `list`: contains the channel-unit pairs in their order of occurrence in the spike rates matrix
            - `map`: contains a dictionary mapping channel-units to index in the spike band power lists.
        @sa cc::CereConn::getSpikeRateEstimatorChannelUnitList()
        """
        ch_u_list = self.thisobj.getSpikeRateEstimatorChannelUnitList()
        outdata = {'list': ch_u_list}
        ch_u_map = {}
        for idx, pair in enumerate(ch_u_list):
            ch_u_map[pair] = idx
        outdata['map'] = ch_u_map
        return outdata

    def set_spike_rate_estimator_loop_interval_ms(self, unsigned int new_interval):
        """!Set the time interval at which spike rates are calculated
    
        @param new_interval interval in milliseconds
        @sa cc::CereConn::setSpikeRateEstimatorLoopIntervalMS()
        """
        self.thisobj.setSpikeRateEstimatorLoopIntervalMS(new_interval)
    
    def get_spike_rate_estimator_loop_interval_ms(self):
        """!Get the time interval at which spike rates are calculated
    
        @return integer representing the interval in milliseconds
        @sa cc::CereConn::getSpikeRateEstimatorLoopIntervalMS()
        """
        return self.thisobj.getSpikeRateEstimatorLoopIntervalMS()
    
    def set_spike_rate_estimation_method_id(self):
        """!Use identity function for spike rate estimation
        
        Spike rate will be estimated by taking the spike count in the last bin.
        @sa cc::CereConn::spikeRateEstimatorId(), cc::sre::r_estimate_id()
        """
        self.thisobj.spikeRateEstimatorId()
    
    def set_spike_rate_estimation_method_boxcar(self, size_t length):
        """!Use boxcar averaging to estimate spike rates
        
        Spike rates will be estimated by averaging over the last length bins.
        
        @param length number of bins to average over
        @sa cc::CereConn::spikeRateEstimatorBoxcar(), cc::sre::r_estimate_with_boxcar()
        """
        self.thisobj.spikeRateEstimatorBoxcar(length)

    def set_spike_rate_estimation_method_exponential(self, double decay_factor, size_t max_length=5):
        """!Use exponential averaging to estimate spike rates
    
        Spike rates from earlier bins have exponentially lower weight.
        I.e. the spike rate calculates as
            
            \f[r = c_0 + d \cdot c_1 + d^2 \cdot c_2 + ... + d^{n-1} \cdot c_{n-1}\f]
            
        where \f$c_i\f$ are spike counts \f$i\f$ steps in the past, \f$d\f$ is the `decay_factor`,
        and \f$n\f$ is `max_length`. The result is then normalized by \f$\sum_{i=0}^{n-1}d^{i}\f$
    
        @param decay_factor double between 0 and 1
        @param max_length integer
        @sa cc:CereConn:spikeRateEstimatorExp(), cc::sre::r_estimate_with_exp_decay()
        """
        self.thisobj.spikeRateEstimatorExp(decay_factor, max_length)

    def get_spike_rate_data(self):
        """!Get buffered Spike Rate data
    
        When Spike Rate Estimation is enabled, this function should be called
        periodically to retrieve the calculated data.
    
        @return a dictionary with keys:
            - `ts`: NSP timestamp at which this function was called
            - `rc`: return code
            - 'oc': overflow count since last call (i.e. how many spike rate items
                have been lost because buffer overflowed)
            - `rates`: represents spike rates as numpy.array with m rows
                and n columns, where m is number of time points and n is
                number of channel-units
            - `rate_ts`: NSP timestamps as 1-dimensional numpy.array of length m
        @sa cc::CereConn::getSpikeRateData()
        """
        cdef uint32_t cbTime = 0
    
        cdef Py_ssize_t n_times, n_ch
        cdef size_t oc
        cdef ReturnCode rc = self.thisobj.getSpikeRateData(self.rateList[0], cbTime, oc)
        
        cdef double[:,:] rate_arr
        cdef double *r_vec_ptr
        cdef double[:] r_vec_view
        
        outdata = {'ts': cbTime, 'rc': <ReturnCode>rc, 'oc': oc}
        
        if rc == ccR_Success:
            
            n_times = self.rateList[0].size()
            n_ch = self.rateList[0][0].second.size()
            
            rate_arr = np.empty((n_times, n_ch), dtype=np.double, order='C')
            
            ts_array = np.empty((n_times, ), dtype=np.uint32, order='C')
            
            for i in range(n_times):
                r_vec_ptr = self.rateList[0][i].second.data()
                r_vec_view = <double[:n_ch]>r_vec_ptr
                
                rate_arr[i, :] = r_vec_view
                
                ts_array[i] = self.rateList[0][i].first
                
            outdata['rates'] = np.asarray(rate_arr)
            outdata['rate_ts'] = ts_array
        else:
            outdata['rates'] = np.empty((0, 0), dtype=np.double, order='C')
            outdata['rate_ts'] = np.empty((0, ), dtype=np.uint32, order='C')
            
        return outdata
    ##@}
    
    ## @name Spike Band Power Estimator related functions
    ##@{
    def start_spike_band_power_estimator(self):
        """!Start spike band power estimation processing in separate thread
        @sa cc::CereConn::spikeBandPowerEstimatorStart()
        """
        self.thisobj.spikeBandPowerEstimatorStart()

    def stop_spike_band_power_estimator(self):
        """!Stop spike band power estimation processing in separate thread
        @sa cc::CereConn::spikeBandPowerEstimatorStop()
        """
        self.thisobj.spikeBandPowerEstimatorStop()
    
    def set_spike_band_power_estimator_ch_list(self, sbpe.ch_list_type ch_list):
        """!Select channels for spike band power estimation
        
        NB this is usually from sample group 6 (i.e. unfiltered raw data)
        
        @param ch_list a list of (1-based) channel ids
        @sa cc::CereConn::setSpikeBandPowerEstimatorChannelList(),
            fill_spike_band_power_estimator_ch_list(), get_spike_band_power_estimator_ch_map()
        """
        self.thisobj.setSpikeBandPowerEstimatorChannelList(ch_list)

    def fill_spike_band_power_estimator_ch_list(self, int n_ch):
        """!Select all channels for spike band power estimation
        
        All channels in `[1, ..., n_ch]` will be used.
        
        @param n_ch an integer channel id
        @sa cc::CereConn::fillSpikeBandPowerEstimatorChannelList(),
            set_spike_band_power_estimator_ch_list(), get_spike_band_power_estimator_ch_map()
        """
        self.thisobj.fillSpikeBandPowerEstimatorChannelList(n_ch)

    def get_spike_band_power_estimator_ch_map(self):
        """!Get channel to spike band power value map
        
        @return a dictionary with keys:
            - `list`: contains the 1-based channel IDs in their order of occurrence
            - `map`: contains a dictionary mapping 1-based channel id
                to index (0-based) in the spike band power lists.
        @sa cc::CereConn::getSpikeBandPowerEstimatorChannelList(), 
            set_spike_band_power_estimator_ch_list(), fill_spike_band_power_estimator_ch_list()
        """
        ch_list = self.thisobj.getSpikeBandPowerEstimatorChannelList()
        outdata = {'list': ch_list}
        ch_map = {}
        for idx, ch in enumerate(ch_list):
            ch_map[ch] = idx
        outdata['map'] = ch_map
        return outdata

    def get_spike_band_power_estimator_loop_interval_ms(self):
        """!Get the time interval at which spike band power is calculated
        
        @return integer representing the interval in milliseconds
        @sa cc::CereConn::getSpikeBandPowerEstimatorLoopIntervalMS()
        """
        return self.thisobj.getSpikeBandPowerEstimatorLoopIntervalMS()
        
    def set_spike_band_power_estimator_loop_interval_ms(self, unsigned int new_interval):
        """!Set the time interval at which spike band power is calculated
        
        @param new_interval interval in milliseconds
        @sa cc::CereConn::setSpikeBandPowerEstimatorLoopIntervalMS()
        """
        self.thisobj.setSpikeBandPowerEstimatorLoopIntervalMS(new_interval)

    def set_spike_band_power_estimator_filter_coefficients(self, b, a):
        """!Set coefficients for the IIR filter used by the Spike Band Power Estimator.
        
        @param b represents the input coefficients, a numpy.array
        @param a the feedback coefficients, numpy.array, of the same size as b
        @sa cc::CereConn::setSpikeBandPowerEstimatorIIRFilterCoeffs()
        """
        cdef ac.vec b_vec, a_vec
        b_vec = numpy_to_vec_d(b)
        a_vec = numpy_to_vec_d(a)
        self.thisobj.setSpikeBandPowerEstimatorIIRFilterCoeffs(b_vec, a_vec)

    def get_spike_band_power_estimator_filter_coefficients(self):
        """!Get coefficients for the IIR filter used by the Spike Band Power Estimator.
        
        @return dictionary with keys:
            - `b`: input coefficients
            - `a`: output coefficients
            The coefficients are represented as a numpy.array vector.
        @sa cc::CereConn::getSpikeBandPowerEstimatorIIRFilterCoeffs()
        """
        coeff_pair = self.thisobj.getSpikeBandPowerEstimatorIIRFilterCoeffs()
        rv = {}
        rv['b'] = vec_to_numpy(coeff_pair.first, None)
        rv['a'] = vec_to_numpy(coeff_pair.second, None)
        return rv

    def set_spike_band_power_estimator_integrated_samples(self, unsigned int n_samples):
        """!Set number of samples the RMS part of the IIR/RMS filter integrates over
        
        @param n_samples number of samples. The represented time depends on the sampling
                rate of the input data
        @sa cc::CereConn::setSpikeBandPowerEstimatorIIRFilterNSamples()
        """
        self.thisobj.setSpikeBandPowerEstimatorIIRFilterNSamples(n_samples)

    def get_spike_band_power_estimator_integrated_samples(self):
        """!Get number of samples the RMS part of the IIR/RMS filter integrates over
        
        @return number of samples, an integer. The represented time depends on the sampling
            rate of the input data
        @sa cc::CereConn::getSpikeBandPowerEstimatorIIRFilterNSamples()
        """
        return self.thisobj.getSpikeBandPowerEstimatorIIRFilterNSamples()

    def set_spike_band_power_estimator_avg_bins(self, unsigned int n_bins):
        """!Set number of bins the RMS part of the IIR/RMS filter averages over
    
        @param n_bins number of bins. The represented time depends on the sampling
                rate of the input data and the number of samples integrated over
        @sa cc::CereConn::setSpikeBandPowerEstimatorAvgNBins(), set_spike_band_power_estimator_integrated_samples()
        """
        self.thisobj.setSpikeBandPowerEstimatorAvgNBins(n_bins)

    def get_spike_band_power_estimator_avg_bins(self):
        """!Get number of bins the RMS part of the IIR/RMS filter averages over
    
        @return number of bins, an integer. The represented time depends on the sampling
            rate of the input data and the number of samples integrated over
        @sa cc::CereConn::getSpikeBandPowerEstimatorAvgNBins(), get_spike_band_power_estimator_integrated_samples()
        """
        return self.thisobj.getSpikeBandPowerEstimatorAvgNBins()


    def set_spike_band_power_estimator_use_sample_group(self, unsigned int sample_group):
        """!Set the sampling group used as input data for SBP estimation
        
        @param sample_group integer representing sampling group. See
                get_cont_data() for a description
        @sa cc::CereConn::setSpikeBandPowerEstimatorAcceptedSampleGroup()
        """
        self.thisobj.setSpikeBandPowerEstimatorAcceptedSampleGroup(sample_group)

    def get_spike_band_power_estimator_use_sample_group(self):
        """!Get the sampling group used as input data for SBP estimation
        @return sampling group integer (see get_cont_data())
        @sa cc::CereConn::getSpikeBandPowerEstimatorAcceptedSampleGroup()
        """
        return self.thisobj.getSpikeBandPowerEstimatorAcceptedSampleGroup()

    def get_spike_band_power_data(self):
        """!Get buffered Spike Band Power data
        
        When Spike Band Power Estimation is enabled, this function should be called
        periodically to retrieve the calculated data.
        
        @return a dictionary with keys:
            - `ts`: NSP timestamp at which this function was called
            - `rc`: return code
            - 'oc': overflow count since last call (i.e. how many spike band power items
                have been lost because buffer overflowed)
            - `sbp`: represents spike band power as numpy.array with m rows
                and n columns, where m is number of time points and n is
                number of channels
            - `rate_ts`: NSP timestamps a 1-dimensional numpy.array of length m
        @sa cc::CereConn::getSpikeBandPowerData()
        """
        cdef uint32_t cbTime = 0
        cdef size_t overflow_count = 0
        cdef Py_ssize_t n_times, n_ch
        cdef ReturnCode rc = self.thisobj.getSpikeBandPowerData(self.sbpList[0], cbTime, overflow_count)
    
        cdef double[:,:] sbp_arr
        cdef double *r_vec_ptr
        cdef double[:] r_vec_view
    
        outdata = {'ts': cbTime, 'rc': <ReturnCode>rc, 'oc': overflow_count}
        n_times = self.sbpList[0].size()
        if rc == ccR_Success and n_times > 0:
            
            n_ch = self.sbpList[0][0].second.size()
            sbp_arr = np.empty((n_times, n_ch), dtype=np.double, order='C')
        
            ts_array = np.empty((n_times, ), dtype=np.uint32, order='C')
        
            for i in range(n_times):
                r_vec_ptr = self.sbpList[0][i].second.data()
                r_vec_view = <double[:n_ch]>r_vec_ptr
            
                sbp_arr[i, :] = r_vec_view
            
                ts_array[i] = self.sbpList[0][i].first
            
            outdata['sbp'] = np.asarray(sbp_arr)
            outdata['rate_ts'] = ts_array
        else:
            outdata['sbp'] = np.empty((0, 0), dtype=np.double, order='C')
            outdata['rate_ts'] = np.empty((0, ), dtype=np.uint32, order='C')
        
        return outdata
    ##@}

##@cond
# armadillo bindings lifted from https://github.com/andrewcron/cy_armadillo
@cython.boundscheck(False)
cdef np.ndarray[np.double_t, ndim=1] vec_to_numpy(const ac.vec & X, np.ndarray[np.double_t, ndim=1] D):
    cdef const double * Xptr = X.memptr()

    if D is None:
        D = np.empty(X.n_elem, dtype=np.double)
    cdef double * Dptr = <double*> D.data
    for i in range(X.n_elem):
        Dptr[i] = Xptr[i]
    return D

cdef ac.vec * numpy_to_vec(np.ndarray[np.double_t, ndim=1] x):
    if not (x.flags.f_contiguous or x.flags.owndata):
        x = x.copy()
    cdef ac.vec *ar_p = new ac.vec(<double*> x.data, x.shape[0], False, True)
    return ar_p

cdef ac.vec numpy_to_vec_d(np.ndarray[np.double_t, ndim=1] x):
    cdef ac.vec *ar_p = numpy_to_vec(x)
    cdef ac.vec ar = deref(ar_p)
    del ar_p
    return ar


def main():
    ck = CereConn()
    ck.print_buffer_state()
    ck.send_open()
    ck.print_buffer_state()
    ck.send_record()


if __name__ == '__main__':
    main()
##@endcond