#!/usr/bin/env python3
## @package test_01
# Spike Band Power Test 01
#
# @author Jonas Zimmermann
#
# This test will open a connection to an NSP, and it will retrieve continuous data and calculate spike band power.

import time
import cere_conn as cc

# get CereConn object
ck = cc.CereConn(withSBPE=True)
ck.set_spike_band_power_estimator_use_sample_group(6)
ck.set_spike_band_power_estimator_integrated_samples(600)

ck.send_open()

# Wait until connection is established
t = time.time()
while ck.get_state() != cc.ccS_Idle:
    time.sleep(0.005)
print("It took {:5.3f}s to open CereConn\n".format(time.time() - t))

# NEW: spike band power estimator has to be started explicitly, unless
# CereConn is called with argument withSBPE=True
#ck.start_spike_band_power_estimator()

# start recording
ck.send_record()


t = time.time()
while ck.get_state() != cc.ccS_Recording:
    time.sleep(0.05)
print("It took {:5.3f}s to start CereConn recording\n".format(time.time() - t))


# at least with NPlayServer, we need to wait for things to settle
time.sleep(.5)


ch_list = [1, 2, 7, 8, 9, 10, 11]

print("Setting channel list to {}".format(ch_list))
ck.set_spike_band_power_estimator_ch_list(ch_list)

ch_map = ck.get_spike_band_power_estimator_ch_map()
print("Channel map is now: {}".format(ch_map))

bin_width = 0.5
run_time = 5.0
print("We will check spike band power data every {} seconds, and run for {} seconds".format(bin_width, run_time))


print("""Read out current spike band power data. The return code should be non-zero, indicating that the data are not valid.
This is because the spike band power estimator has not had a chance to process samples with the new channel map""")
cd = ck.get_spike_band_power_data()
print(cd)

intvl = ck.get_spike_band_power_estimator_loop_interval_ms()
print("Spike band power loop interval is {}ms".format(intvl))

time.sleep(.5)
print("""Read out current spike band power data. The return code should be zero now, indicating that the data are valid.""")
cd = ck.get_spike_band_power_data()
print(cd)


ck.set_spike_band_power_estimator_loop_interval_ms(42)
intvl = ck.get_spike_band_power_estimator_loop_interval_ms()
print("Spike band power loop interval is now {}ms".format(intvl))

print("""Read out current spike band power data. The return code should be non-zero, indicating that the data are not valid.
This is because the spike band power estimator has not had a chance to process samples with the new loop interval.""")
cd = ck.get_spike_band_power_data()
print(cd)

time.sleep(.5)
print("""Read out current spike band power data. The return code should be zero now, indicating that the data are valid.""")
cd = ck.get_spike_band_power_data()
print(cd)


# keep stats...
lasttime = time.time()
starttime = lasttime
times = []
samples = {}

start_ts = cd['ts']
last_ts = start_ts

# run data acquisition for a while
while (time.time() - starttime < run_time):
    cd = ck.get_spike_band_power_data()
    print(cd)
    print("NSP Time since start {}s. Time since last loop {}s".format((cd['ts'] - start_ts) / 30000.0, (cd['ts'] - last_ts) / 30000.0))
    last_ts = cd['ts']

    ctime = time.time()
    times.append( ctime - lasttime)
    lasttime = ctime
    
    print("System time since last loop: {}s".format(times[-1]))
    
    time.sleep(bin_width)

print("Changing spike band power estimator loop again, to 10 ms:")
ck.set_spike_band_power_estimator_loop_interval_ms(10)

ch_list = [0]

print("Changing channel map again, to {}".format(ch_list))
ck.set_spike_band_power_estimator_ch_list(ch_list)

print("Sleeping, to give the spike band power estimator a chance to update the buffers")
time.sleep(bin_width)
print("Getting new spike band power data:")
cd = ck.get_spike_band_power_data()
print(cd)


print("Setting number of samples collected per RMS cycle to 42:")
ck.set_spike_band_power_estimator_integrated_samples(42)
time.sleep(bin_width)
print("Confirming number of samples collected: {}".format(ck.get_spike_band_power_estimator_integrated_samples()))
time.sleep(0.5)
print("Getting new spike band power data:")
cd = ck.get_spike_band_power_data()
print(cd)


ck.send_close()


print("Some stats: Bin width: {:4.3f}s".format(bin_width))

while ck.get_state() != cc.ccS_Closed:
    time.sleep(0.05)
