#!/usr/bin/env python3
## @package test_01
# Spike Rate Test 01
#
# @author Jonas Zimmermann
#
# This test will open a connection to an NSP, and it will retrieve spike events and calculate spike rate data.

import time
import cere_conn as cc

# get CereConn object
ck = cc.CereConn(withSRE=True)
ck.send_open()

# Wait until connection is established
t = time.time()
while ck.get_state() != cc.ccS_Idle:
    time.sleep(0.005)
print("It took {:5.3f}s to open CereConn\n".format(time.time() - t))

# NEW: spike rate estimator has to be started explicitly, unless
# CereConn is called with argument withSRE=True
#ck.start_spike_rate_estimator()

# start recording
ck.send_record()


t = time.time()
while ck.get_state() != cc.ccS_Recording:
    time.sleep(0.05)
print("It took {:5.3f}s to start CereConn recording\n".format(time.time() - t))


# at least with NPlayServer, we need to wait for things to settle
time.sleep(.5)


ch_u_list = [(1,1), (2,1), (1, 0), (2, 0) ]

print("Setting channel-unit list to {}".format(ch_u_list))
ck.set_spike_rate_estimator_ch_u_list(ch_u_list)

ch_map = ck.get_spike_rate_estimator_ch_u_map()
print("Channel map is now: {}".format(ch_map))

bin_width = 0.5
run_time = 5.0
print("We will check spike rate data every {} seconds, and run for {} seconds".format(bin_width, run_time))


print("""Read out current spike rate data. The return code should be non-zero, indicating that the rates are not valid.
This is because the spike rate estimator has not had a chance to process spikes with the new channel map""")
cd = ck.get_spike_rate_data()
print(cd)

intvl = ck.get_spike_rate_estimator_loop_interval_ms()
print("Spike rate loop interval is {}ms".format(intvl))

time.sleep(.5)
print("""Read out current spike rate data. The return code should be zero now, indicating that the rates are valid.""")
cd = ck.get_spike_rate_data()
print(cd)


ck.set_spike_rate_estimator_loop_interval_ms(42)
intvl = ck.get_spike_rate_estimator_loop_interval_ms()
print("Spike rate loop interval is now {}ms".format(intvl))

print("""Read out current spike rate data. The return code should be non-zero, indicating that the rates are not valid.
This is because the spike rate estimator has not had a chance to process spikes with the new loop interval.""")
cd = ck.get_spike_rate_data()
print(cd)

time.sleep(.5)
print("""Read out current spike rate data. The return code should be zero now, indicating that the rates are valid.""")
cd = ck.get_spike_rate_data()
print(cd)


# keep stats...
lasttime = time.time()
starttime = lasttime
times = []
samples = {}

start_ts = cd['ts']
last_ts = start_ts

# run data acquisition for a while
while (time.time() - starttime < run_time):
    cd = ck.get_spike_rate_data()
    print(cd)
    print("NSP Time since start {}s. Time since last loop {}s".format((cd['ts'] - start_ts) / 30000.0, (cd['ts'] - last_ts) / 30000.0))
    last_ts = cd['ts']

    ctime = time.time()
    times.append( ctime - lasttime)
    lasttime = ctime
    
    print("System time since last loop: {}s".format(times[-1]))
    
    time.sleep(bin_width)

print("Changing spike rate estimator loop again, to 10 ms:")
ck.set_spike_rate_estimator_loop_interval_ms(10)

ch_u_list = [(1, 0)]

print("Changing channel map again, to {}".format(ch_u_list))
ck.set_spike_rate_estimator_ch_u_list(ch_u_list)

print("Changing spike rate estimation method, to exponential decay, with decay factor 0.8 and at max 20 history bins used for calculation")
ck.set_spike_rate_estimation_method_exponential(.8, 20)

print("Sleeping, to give the spike rate estimator a chance to update the buffers")
time.sleep(bin_width)
print("Getting new spike rate data:")
cd = ck.get_spike_rate_data()
print(cd)

print("Changing spike rate estimation method, to running average, with a window of 3 history bins used for calculation")
ck.set_spike_rate_estimation_method_boxcar(3)

print("Sleeping, to give the spike rat estimator a chance to update the buffers")
time.sleep(bin_width)
print("Getting new spike rate data:")
cd = ck.get_spike_rate_data()
print(cd)


print("Changing spike rate estimation method, to exponential decay, with decay factor 0.4 and at max 10 history bins used for calculation")
ck.set_spike_rate_estimation_method_exponential(.4, 10)
print("Sleeping, to give the spike rat estimator a chance to update the buffers")
time.sleep(bin_width)
print("Getting new spike rate data:")
cd = ck.get_spike_rate_data()
print(cd)
time.sleep(bin_width)

ck.send_close()


print("Some stats: Bin width: {:4.3f}s".format(bin_width))

while ck.get_state() != cc.ccS_Closed:
    time.sleep(0.05)
