#include "gtest/gtest.h"

#include "CereConn.h"

using namespace std::chrono;
using namespace std::literals;

// The fixture for testing class Project1. From google test primer.
class CereConnRecordingTest : public ::testing::Test
{
protected:
    // You can remove any or all of the following functions if its body
    // is empty.

    CereConnRecordingTest()
    {
        spdlog::set_pattern("[%Y-%m-%d %H:%M:%S.%f] [%n] [%^%l%$] [PID:%P] [TID:%t] %v");
        
        ck = new cc::CereConn();
    }

    virtual ~CereConnRecordingTest()
    {
        delete ck;
    }

    // If the constructor and destructor are not enough for setting up
    // and cleaning up each test, you can define the following methods:
    virtual void SetUp() override
    {
        ck->ccOpen();
        auto start = std::chrono::high_resolution_clock::now();
        auto finish = start + std::chrono::milliseconds(5000);

        while ((std::chrono::high_resolution_clock::now() < finish) && ck->getState() != cc::CereConn::State::Idle)
        {
            std::this_thread::sleep_for(std::chrono::milliseconds(10));
        }
        ck->ccRecord();
        
        start = std::chrono::high_resolution_clock::now();
        finish = start + std::chrono::milliseconds(5000);
        while ((std::chrono::high_resolution_clock::now() < finish) && ck->getState() != cc::CereConn::State::Recording)
        {
            std::this_thread::sleep_for(std::chrono::milliseconds(10));
        }
        
    }

    virtual void TearDown() override
    {
        ck->ccClose();
        std::this_thread::sleep_for(std::chrono::milliseconds(1000));
    }

    // Objects declared here can be used by all tests in the test case for Project1.
    
    cc::CereConn *ck;

};


TEST_F(CereConnRecordingTest, StartRecording)
{
    uint32_t ts0, ts1, tdiff;
    cc::SampleGroupList sgl;
    
    std::this_thread::sleep_for(std::chrono::milliseconds(1000));
    ck->getSampleGroupData(sgl, ts0);

    std::this_thread::sleep_for(std::chrono::milliseconds(500));
    ck->getSampleGroupData(sgl, ts1);
    ck->ccIdle();
    tdiff = ts1 - ts0;
    bool received_samples = false;
    std::vector<uint32_t> sgr_chids[8] = {{1}, {2,3}, {4,5,6},{7,8},{9},{}, {}, {}};
    for (int i = 0; i < sgl.sgc.size(); ++i)
    {
        auto  & sgc = sgl.sgc[i];
        auto sampPerCh = sgc.channel_ids.size() ? sgc.buffer.size() / sgc.channel_ids.size() : 0;
        unsigned int expectedSamps = tdiff / cc::constants::SampleGroupPeriods[i];
        if (sampPerCh)
        {
            received_samples = true;
            EXPECT_NEAR(sampPerCh, expectedSamps, expectedSamps / 40) << "Received more or fewer samples per channel than expected";
        }
        EXPECT_EQ(sgr_chids[i], sgc.channel_ids) << "Channel settings should come from test_NSP_settings_01.ccf";
        
    }
    ASSERT_TRUE(received_samples) << "No samples received. Is Central configured to send data?";
    
    
}

TEST_F(CereConnRecordingTest, SendComment)
{
    EXPECT_EQ(ck->sendComment("TestComment", 42), cc::CereConn::ReturnCode::Success);
    std::this_thread::sleep_for(std::chrono::milliseconds(100));
    cc::CommentList cl;
    uint32_t ts;
    EXPECT_EQ(ck->getCommentData(cl, ts), cc::CereConn::ReturnCode::Success);
    ASSERT_GE(cl.size(), 1);
    EXPECT_STREQ(cl[0].comment, "TestComment");
    EXPECT_EQ(cl[0].data, 42);
}

TEST_F(CereConnRecordingTest, SpikeRates)
{
    uint32_t ts0, ts1, tdiff;
    
    cc::sre::SpikeRateEstimator::rate_list_type srl;
    
    std::this_thread::sleep_for(std::chrono::milliseconds(1000));
    ck->getSpikeRateData(srl, ts0);

    std::this_thread::sleep_for(std::chrono::milliseconds(500));
    ck->getSpikeRateData(srl, ts1);
    ck->ccIdle();
    tdiff = ts1 - ts0;
    
    printf("TDiff: %i\n", tdiff);

    printf("Rates:\n");
    for (auto const& row: srl)
    {
        for (auto const& v:row.second)
        {
            printf("%.3f ", v);
        }
        printf("\n");
    }
    bool received_samples = false;
}

TEST_F(CereConnRecordingTest, SpikeRatesSetLoopTime)
{
    uint32_t ts0, ts1, tdiff;
    ck->fillSpikeRateEstimatorChannelUnitList(1, 1);
    
    auto t_loop = ck->getSpikeRateEstimatorLoopInterval();
    EXPECT_EQ(t_loop, 100ms);
    ck->setSpikeRateEstimatorLoopInterval(13ms);
    t_loop = ck->getSpikeRateEstimatorLoopInterval();
    EXPECT_EQ(t_loop, 13ms);
    
    ck->setSpikeRateEstimatorLoopIntervalMS(42);
    t_loop = ck->getSpikeRateEstimatorLoopInterval();
    EXPECT_EQ(t_loop, 42ms);
    auto t_loop_ms = ck->getSpikeRateEstimatorLoopIntervalMS();
    EXPECT_EQ(t_loop_ms, 42);
    
    
    cc::sre::SpikeRateEstimator::rate_list_type srl;
    
    std::this_thread::sleep_for(std::chrono::milliseconds(1000));
    ck->getSpikeRateData(srl, ts0);

    std::this_thread::sleep_for(std::chrono::milliseconds(500));
    ck->getSpikeRateData(srl, ts1);
    ck->ccIdle();
    tdiff = ts1 - ts0;
    
    printf("TDiff: %i\n", tdiff);

    printf("Rates:\n");
    for (auto const& row: srl)
    {
        for (auto const& v:row.second)
        {
            printf("%.3f ", v);
        }
        printf("\n");
    }
    bool received_samples = false;
}

TEST_F(CereConnRecordingTest, SpikeBandPower)
{
    uint32_t ts0, ts1, tdiff;
    
    cc::sbpe::SpikeBandPowerEstimator::sbp_list_type sbpV;
    ck->fillSpikeBandPowerEstimatorChannelList(16);
    
    std::this_thread::sleep_for(std::chrono::milliseconds(1000));
    ck->getSpikeBandPowerData(sbpV, ts0);

    std::this_thread::sleep_for(std::chrono::milliseconds(500));
    ck->getSpikeBandPowerData(sbpV, ts1);
    ck->ccIdle();
    tdiff = ts1 - ts0;
    
    printf("TDiff: %i\n", tdiff);

    printf("SBP:\n");
    for (auto const& row: sbpV)
    {
        for (auto const& v:row.second)
        {
            printf("%.3f ", v);
        }
        printf("\n");
    }
    bool received_samples = false;
}

TEST_F(CereConnRecordingTest, SerialDigIn)
{
    uint32_t ts0, ts1, tdiff;
    ck->spikeRateEstimatorStop();
    ck->spikeBandPowerEstimatorStop();
    cc::DigInEvList dieV;
    
    std::this_thread::sleep_for(std::chrono::milliseconds(100));
    ck->getDigInEvData(dieV, ts0);

    std::this_thread::sleep_for(std::chrono::milliseconds(2000));
    ck->getDigInEvData(dieV, ts1);
    ck->ccIdle();
    tdiff = ts1 - ts0;
    
    printf("TDiff: %i\n", tdiff);

    printf("DigInEv:\n");
    for (auto const& row: dieV)
    {
            printf("%i: %i source: %i\n", row.timestamp, row.data, static_cast<int>(row.source));
    }
    bool received_samples = false;
}
