#include "gtest/gtest.h"

#include "CereConn.h"

// The fixture for testing class Project1. From google test primer.
class CereConnRecordingPerfTest : public ::testing::Test
{
protected:
    virtual void SetUp() override
    {
        ck.ccOpen();
        auto start = std::chrono::high_resolution_clock::now();
        auto finish = start + std::chrono::milliseconds(5000);

        while ((std::chrono::high_resolution_clock::now() < finish) && ck.getState() != cc::CereConn::State::Idle)
        {
            std::this_thread::sleep_for(std::chrono::milliseconds(10));
        }
        ck.ccRecord();

        start = std::chrono::high_resolution_clock::now();
        finish = start + std::chrono::milliseconds(5000);
        while ((std::chrono::high_resolution_clock::now() < finish) && ck.getState() != cc::CereConn::State::Recording)
        {
            std::this_thread::sleep_for(std::chrono::milliseconds(10));
        }

    }

    virtual void TearDown() override
    {
        ck.ccClose();
        std::this_thread::sleep_for(std::chrono::milliseconds(1000));
    }

    cc::CereConn ck;

};


TEST_F(CereConnRecordingPerfTest, Fast)
{
    const size_t n_iter = 1000;
    uint32_t ts0, ts1, tdiff;
    std::vector<size_t> samples_received;
    std::vector<uint32_t> at_times;

    samples_received.reserve(n_iter);
    at_times.reserve(n_iter);


    cc::SampleGroupList sgl;
    auto start = std::chrono::high_resolution_clock::now();

    ck.getSampleGroupData(sgl, ts0);
    ts0 = sgl.sgc[5].last_timestamp;
    for (size_t i = 0; i < n_iter; ++i)
    {
        std::this_thread::sleep_for(std::chrono::milliseconds(2));

        ck.getSampleGroupData(sgl, ts1);
        auto  & sgc = sgl.sgc[5];
        at_times.push_back(sgc.last_timestamp);
        if (sgc.channel_ids.size())
        {
            samples_received.push_back( sgc.buffer.size() / sgc.channel_ids.size() );
        }
    }
    auto finish = std::chrono::high_resolution_clock::now();

    ck.ccIdle();

    size_t total_samples_received = 0, ttime = 0;

    for (auto const & n : samples_received)
    {
        total_samples_received += n;
    }
    ttime = at_times.back() - ts0;

    auto dur_ms = std::chrono::duration_cast<std::chrono::milliseconds>(finish - start).count();
    printf("Received %zu samples in %zu ticks, that is %lld ms or %.1f samples/ms\n", total_samples_received, ttime, dur_ms, (double)total_samples_received/ (double)dur_ms);
}

TEST_F(CereConnRecordingPerfTest, Fast5)
{
    const size_t n_iter = 1000;
    uint32_t ts0, ts1, tdiff;
    std::vector<size_t> samples_received;
    std::vector<uint32_t> at_times;

    samples_received.reserve(n_iter);
    at_times.reserve(n_iter);


    cc::SampleGroupList sgl;
    auto start = std::chrono::high_resolution_clock::now();
    ck.setSampleGroup(6, false);

    ck.getSampleGroupData(sgl, ts0);
    ts0 = sgl.sgc[4].last_timestamp;
    for (size_t i = 0; i < n_iter; ++i)
    {
        std::this_thread::sleep_for(std::chrono::milliseconds(2));

        ck.getSampleGroupData(sgl, ts1);
        auto  & sgc = sgl.sgc[4];
        at_times.push_back(sgc.last_timestamp);
        if (sgc.channel_ids.size())
        {
            samples_received.push_back( sgc.buffer.size() / sgc.channel_ids.size() );
        }
    }
    auto finish = std::chrono::high_resolution_clock::now();

    ck.ccIdle();

    size_t total_samples_received = 0, ttime = 0;

    for (auto const & n : samples_received)
    {
        total_samples_received += n;
    }
    ttime = at_times.back() - ts0;

    auto dur_ms = std::chrono::duration_cast<std::chrono::milliseconds>(finish - start).count();
    printf("Received %zu samples in %zu ticks, that is %lld ms or %.1f samples/ms\n", total_samples_received, ttime, dur_ms, (double)total_samples_received/ (double)dur_ms);
}

TEST_F(CereConnRecordingPerfTest, FastIgnoreGroup6InMiddle)
{
    const size_t n_iter = 10;
    uint32_t ts0, ts1, tdiff;
    std::vector<size_t> samples_received;
    std::vector<uint32_t> at_times;

    samples_received.reserve(n_iter);
    at_times.reserve(n_iter);

    ck.setSampleGroup(6, false);

    cc::SampleGroupList sgl;
    auto start = std::chrono::high_resolution_clock::now();
    ck.getSampleGroupData(sgl, ts0);
    ts0 = sgl.sgc[4].last_timestamp;

    for (size_t i = 0; i < n_iter; ++i)
    {
        std::this_thread::sleep_for(std::chrono::milliseconds(2));

        ck.getSampleGroupData(sgl, ts1);
        auto  & sgc = sgl.sgc[5];
        at_times.push_back(sgc.last_timestamp);
        if (sgc.channel_ids.size())
        {
            samples_received.push_back( sgc.buffer.size() / sgc.channel_ids.size() );
        }

        if (i == n_iter / 2)
        {
            ck.setSampleGroup(6, true);
        }
    }
    auto finish = std::chrono::high_resolution_clock::now();

    ck.ccIdle();

    size_t total_samples_received = 0, ttime = 0;

    for (auto const & n : samples_received)
    {
        total_samples_received += n;
    }
    ttime = at_times.back() - ts0;

    auto dur_ms = std::chrono::duration_cast<std::chrono::milliseconds>(finish - start).count();
    printf("Received %zu samples in %zu ticks, that is %lld ms or %.1f samples/ms\n", total_samples_received, ttime, dur_ms, (double)total_samples_received/ (double)dur_ms);

    ASSERT_LT(total_samples_received, ttime);
    ASSERT_GT(total_samples_received, 0);

}


TEST_F(CereConnRecordingPerfTest, FastWithSBP)
{
    const size_t n_iter = 100;
    uint32_t ts0, ts1, tdiff;
    std::vector<size_t> samples_received;
    std::vector<uint32_t> at_times;

    samples_received.reserve(n_iter);
    at_times.reserve(n_iter);
    
    cc::sbpe::SpikeBandPowerEstimator::sbp_list_type sbpV;
    

    cc::SampleGroupList sgl;


    ck.getSpikeBandPowerData(sbpV, ts0);
    auto start = std::chrono::high_resolution_clock::now();

    for (size_t i = 0; i < n_iter; ++i)
    {
        std::this_thread::sleep_for(std::chrono::milliseconds(50));
        ck.getSpikeBandPowerData(sbpV, ts1);

        samples_received.push_back(sbpV.size());
        if (sbpV.size())
        {
            at_times.push_back(sbpV[sbpV.size() - 1].first);
        }
    }
    auto finish = std::chrono::high_resolution_clock::now();

    ck.ccIdle();

    size_t total_samples_received = 0, ttime = 0;

    for (auto const & n : samples_received)
    {
        total_samples_received += n;
    }
    ttime = at_times.back() - ts0;

    auto dur_ms = std::chrono::duration_cast<std::chrono::milliseconds>(finish - start).count();
    printf("Received %zu spike band power in %zu ticks, that is %lld ms or %.1f samples/s\n", total_samples_received, ttime, dur_ms, (double)total_samples_received/ (double)dur_ms * 1000);
}

