#include <cstdio>
#include "gtest/gtest.h"
#include "rb/SampleGroupRingBuffer.h"
#include "rb/SampleGroupList.h"

namespace cc
{
    
template<size_t NChannels, size_t Size>
using TestSampleGroupRingBuffer = SampleGroupRingBufferT<int16_t, uint32_t, NChannels, Size>;

template<size_t NChannels, size_t Size>
using TestSampleGroupRBC = SampleGroupRingBufferCollectionT<int16_t, uint32_t, NChannels, Size, 3>;

template<size_t NChannels, size_t Size>
using TestSampleGroupChunk = cc::SampleGroupChunkT<int16_t, uint32_t, NChannels, Size>;

// The fixture for testing class Project1. From google test primer.
class ContRBTest : public ::testing::Test
{
protected:
    // Objects declared here can be used by all tests in the test case for Project1.
    static constexpr size_t TRBSize = 5;
    static constexpr size_t TRBNCh1 = 6;
    static constexpr size_t TRBNCh2 = 144;
    
    TestSampleGroupRingBuffer<TRBNCh1, TRBSize> ccrb;
    TestSampleGroupRingBuffer<TRBNCh2, TRBSize> ccrb_cb;

};
constexpr size_t ContRBTest::TRBSize;
constexpr size_t ContRBTest::TRBNCh1;
constexpr size_t ContRBTest::TRBNCh2;


TEST_F(ContRBTest, RingBufferEmpty)
{
    ASSERT_TRUE(ccrb.wasEmpty());
    ASSERT_EQ(ccrb.getChannelCount(), 0);
    ASSERT_EQ(ccrb.ch_list_cbegin(), ccrb.ch_list_cend());
}

TEST_F(ContRBTest, AddOneSample)
{
    TestSampleGroupRingBuffer<TRBNCh1, TRBSize>::BufType oneSample({1, 2, 3, 4, 5, 0});
    ccrb.push(oneSample);
    ASSERT_FALSE(ccrb.wasEmpty());
}

TEST_F(ContRBTest, LastSampleTime)
{
    ASSERT_EQ(ccrb.getLastSampleTime(), 0);
}

TEST_F(ContRBTest, AddSampleFromPkg)
{    
    cbPKT_GROUP pk1 = {42, 0, 1, 3, {23, 42, 13, 7, 9, -1}};
    cbPKT_GROUP pk2 = {121, 0, 1, 3, {9, 0, 12, 8, 19, -1}};
    ChannelNumList_t ccl1 = {1, 2, 6, 7, 8};
    
    ccrb_cb.reset(ccl1);
    ccrb_cb.addSamplesFromPacket(&pk1);
    ASSERT_EQ(ccrb_cb.getLastSampleTime(), 42);
    ASSERT_EQ(ccrb_cb.getChannelCount(), 5);
    for (int i = 0; i < ccrb_cb.getChannelCount(); ++i)
    {
        ASSERT_EQ(*(ccrb_cb.ch_list_cbegin() + i), ccl1[i]);
        ASSERT_NE(ccrb_cb.ch_list_cbegin() + i, ccrb_cb.ch_list_cend());
    }
    ASSERT_EQ(ccrb_cb.ch_list_cbegin() + 5, ccrb_cb.ch_list_cend());
    
    ccrb_cb.addSamplesFromPacket(&pk2);
    ASSERT_EQ(ccrb_cb.getLastSampleTime(), 121);
    ASSERT_EQ(ccrb_cb.getChannelCount(), 5);
    
    TestSampleGroupChunk<TRBNCh2, TRBSize> tsgc;
    ASSERT_TRUE(ccrb_cb.consume(tsgc));
    EXPECT_TRUE(tsgc.channel_ids_changed);

    ASSERT_EQ(tsgc.last_timestamp, 121);
    
    ASSERT_EQ(tsgc.buffer.size(), 2 * 5);

    ASSERT_EQ(tsgc.buffer[0], 23);
    ASSERT_EQ(tsgc.buffer[1], 42);
    ASSERT_EQ(tsgc.buffer[2], 13);
    ASSERT_EQ(tsgc.buffer[3], 7);
    ASSERT_EQ(tsgc.buffer[4], 9);
    ASSERT_EQ(tsgc.buffer[5], 9);
    ASSERT_EQ(tsgc.buffer[6], 0);
    ASSERT_EQ(tsgc.buffer[7], 12);
    ASSERT_EQ(tsgc.buffer[8], 8);
    ASSERT_EQ(tsgc.buffer[9], 19);
    
    ASSERT_EQ(tsgc.channel_ids.size(), 5);
    ASSERT_EQ(tsgc.channel_ids[0], 1);
    ASSERT_EQ(tsgc.channel_ids[1], 2);
    ASSERT_EQ(tsgc.channel_ids[2], 6);
    ASSERT_EQ(tsgc.channel_ids[3], 7);
    ASSERT_EQ(tsgc.channel_ids[4], 8);
    
}

TEST_F(ContRBTest, ResetDataWhenChannelChanges)
{
    cbPKT_GROUP pk1 = {42, 0, 1, 3, {23, 42, 13, 7, 9, -1}};
    cbPKT_GROUP pk2 = {121, 0, 1, 3, {9, 0, 12, 8, 19, -1}};
    ChannelNumList_t ccl1 = {1, 2, 6, 7, 8};
    ChannelNumList_t ccl2 = {1, 2, 6, 7};
    
    ccrb_cb.reset(ccl1);
    ccrb_cb.addSamplesFromPacket(&pk1);
    ASSERT_EQ(ccrb_cb.getLastSampleTime(), 42);
    ASSERT_EQ(ccrb_cb.getChannelCount(), 5);
    ccrb_cb.reset(ccl2);
    ccrb_cb.addSamplesFromPacket(&pk2);
    ASSERT_EQ(ccrb_cb.getLastSampleTime(), 121);
    ASSERT_EQ(ccrb_cb.getChannelCount(), 4);
    
    
    TestSampleGroupChunk<TRBNCh2, TRBSize> tsgc;
    ASSERT_TRUE(ccrb_cb.consume(tsgc));
    EXPECT_TRUE(tsgc.channel_ids_changed);
    
    ASSERT_EQ(tsgc.buffer.size(), 4);

    ASSERT_EQ(tsgc.buffer[0], 9);
    ASSERT_EQ(tsgc.buffer[1], 0);
    ASSERT_EQ(tsgc.buffer[2], 12);
    ASSERT_EQ(tsgc.buffer[3], 8);
    ASSERT_EQ(tsgc.last_timestamp, 121);
    ASSERT_EQ(tsgc.channel_ids.size(), 4);
    ASSERT_EQ(tsgc.channel_ids[0], 1);
    ASSERT_EQ(tsgc.channel_ids[1], 2);
    ASSERT_EQ(tsgc.channel_ids[2], 6);
    ASSERT_EQ(tsgc.channel_ids[3], 7);
    
}

TEST_F(ContRBTest, BufferEmpty)
{
    TestSampleGroupChunk<TRBNCh2, TRBSize> tsgc;
    
    cbPKT_GROUP pk1 = {42, 0, 1, 3, {23, 42, 13, 7, 9, -1}};
    ChannelNumList_t ccl1 = {1, 2, 6, 7, 8};
    
    ccrb_cb.reset(ccl1);
    ASSERT_EQ(ccrb_cb.getChannelCount(), 5);
    
    ccrb_cb.addSamplesFromPacket(&pk1);
    ASSERT_EQ(ccrb_cb.getLastSampleTime(), 42);
    ASSERT_EQ(ccrb_cb.getChannelCount(), 5);
    
    ASSERT_TRUE(ccrb_cb.consume(tsgc));
    ASSERT_EQ(tsgc.buffer.size(), 5);
    ASSERT_TRUE(tsgc.channel_ids_changed);
    
    ASSERT_FALSE(ccrb_cb.consume(tsgc));
    ASSERT_EQ(tsgc.buffer.size(), 0);
    ASSERT_FALSE(tsgc.channel_ids_changed);

    ASSERT_EQ(tsgc.channel_ids.size(), 5);
    ASSERT_EQ(tsgc.channel_ids[0], 1);
    ASSERT_EQ(tsgc.channel_ids[1], 2);
    ASSERT_EQ(tsgc.channel_ids[2], 6);
    ASSERT_EQ(tsgc.channel_ids[3], 7);
    ASSERT_EQ(tsgc.channel_ids[4], 8);

}

TEST_F(ContRBTest, EnoughData)
{
    cbPKT_GROUP pk1 = {42, 0, 1, 2, {23, 42, 13, 7}};
    ChannelNumList_t ccl1 = {1, 2, 6, 7, 8};
    
    ccrb_cb.reset(ccl1);
    ASSERT_EQ(ccrb_cb.getChannelCount(), 5);
    ASSERT_FALSE(ccrb_cb.addSamplesFromPacket(&pk1));
}

TEST_F(ContRBTest, AddSampleFromRawData)
{
    TestSampleGroupChunk<TRBNCh2, TRBSize> tsgc;
    
    cbPKT_GROUP pk1 = {42, 0, 1, 2, {23, 42, 13, 7}};
    ChannelNumList_t ccl1 = {1, 2, 6, 7};
    
    ccrb_cb.reset(ccl1);
    ASSERT_EQ(ccrb_cb.getChannelCount(), 4);
    ASSERT_TRUE(ccrb_cb.addSamplesFromPacket(pk1.time, pk1.data, pk1.dlen * 2));
    ASSERT_FALSE(ccrb_cb.addSamplesFromPacket(pk1.time, pk1.data, pk1.dlen * 2)) << "Packets with same timestamp should be rejected";
    ASSERT_TRUE(ccrb_cb.addSamplesFromPacket(pk1.time + 1, pk1.data, pk1.dlen * 2)) << "Packet with increased timestamp should be acccepted";
    ASSERT_FALSE(ccrb_cb.addSamplesFromPacket(pk1.time, pk1.data, pk1.dlen * 2)) << "Packets with decreased timestamp should be rejected";
    ASSERT_TRUE(ccrb_cb.addSamplesFromPacket(pk1.time + 2, pk1.data, pk1.dlen * 2)) << "Packet with increased timestamp should be acccepted";
    ASSERT_FALSE(ccrb_cb.addSamplesFromPacket(pk1.time + 3, pk1.data, pk1.dlen * 2 - 1)) << "Packets with not enough data should be rejected";
    
    ASSERT_TRUE(ccrb_cb.consume(tsgc)) << "Consumption of buffer should succeed";
    ASSERT_EQ(tsgc.buffer.size(), 3 * 4) << "There should be 12 samples in buffer (3 times, 4 channels)";
    ASSERT_TRUE(tsgc.channel_ids_changed) << "Channel IDs have changed since last 'consume'!";
    
    ASSERT_TRUE(ccrb_cb.addSamplesFromPacket(pk1.time + 3, pk1.data, pk1.dlen * 2)) << "Packet with increased timestamp should be acccepted";
    ASSERT_TRUE(ccrb_cb.addSamplesFromPacket(pk1.time + 4, pk1.data, pk1.dlen * 2)) << "Packet with increased timestamp should be acccepted";
    
    ASSERT_TRUE(ccrb_cb.consume(tsgc)) << "Consumption of buffer should succeed";
    ASSERT_EQ(tsgc.buffer.size(), 2 * 4) << "There should be 8 samples in buffer (2 times, 4 channels)";
    ASSERT_FALSE(tsgc.channel_ids_changed) << "Channel IDs should not have changed since last 'consume'!";
    
}

TEST_F(ContRBTest, SetCAR)
{
    TestSampleGroupChunk<TRBNCh2, TRBSize> tsgc;
    
    ChannelNumList_t ccl1 = {1, 2, 6, 7, 10};
    ChannelNumList_t ccl2 = {2, 6, 8, 10, 11};
    ChannelNumList_t carl = {2, 7, 10};
    ChannelNumList_t carlEmpty = {};
    ChannelNumList_t carlInvalid = {200, 300, 400, 500};
    
    cbPKT_GROUP pk1 = {42, 0, 1, 3, {10, -20, 40, -80, 160, 0}};
    
    ccrb_cb.reset(ccl1);
    ASSERT_FALSE(ccrb_cb.mp_car) << "member pointer to CAR should not be set";
    ASSERT_TRUE(ccrb_cb.setCARChannels(carl));
    ASSERT_TRUE(ccrb_cb.mp_car) << "member pointer to CAR should be set now";
    ASSERT_TRUE(ccrb_cb.addSamplesFromPacket(&pk1)) << "Accept packet";
    // average is = 60 / 3 = 20
    ASSERT_TRUE(ccrb_cb.consume(tsgc)) << "Consumption of buffer should succeed";
    ASSERT_EQ(tsgc.buffer.size(), 5) << "There should be 5 samples in buffer (1 time, 5 channels)";
    EXPECT_EQ(tsgc.buffer[0], -10);
    EXPECT_EQ(tsgc.buffer[1], -40);
    EXPECT_EQ(tsgc.buffer[2], 20);
    EXPECT_EQ(tsgc.buffer[3], -100);
    EXPECT_EQ(tsgc.buffer[4], 140);
    
    // changing sample groups
    ccrb_cb.reset(ccl2);
    ++ pk1.time;
    ASSERT_TRUE(ccrb_cb.mp_car) << "member pointer to CAR should be set now";
    ASSERT_TRUE(ccrb_cb.addSamplesFromPacket(&pk1)) << "Accept packet";
    // avg is now -70 / 2 = -35
    ASSERT_TRUE(ccrb_cb.consume(tsgc)) << "Consumption of buffer should succeed";
    ASSERT_EQ(tsgc.buffer.size(), 5) << "There should be 5 samples in buffer (1 time, 5 channels)";
    EXPECT_EQ(tsgc.buffer[0], 45);
    EXPECT_EQ(tsgc.buffer[1], 15);
    EXPECT_EQ(tsgc.buffer[2], 75);
    EXPECT_EQ(tsgc.buffer[3], -45);
    EXPECT_EQ(tsgc.buffer[4], 195);
    
    // manual complete reset
    ccrb_cb.reset();
    ccrb_cb.reset(ccl2);
    ++ pk1.time;
    ASSERT_FALSE(ccrb_cb.mp_car) << "member pointer to CAR should not be set now";
    ASSERT_TRUE(ccrb_cb.addSamplesFromPacket(&pk1)) << "Accept packet";
    ASSERT_TRUE(ccrb_cb.consume(tsgc)) << "Consumption of buffer should succeed";
    ASSERT_EQ(tsgc.buffer.size(), 5) << "There should be 5 samples in buffer (1 time, 5 channels)";
    EXPECT_EQ(tsgc.buffer[0], 10);
    EXPECT_EQ(tsgc.buffer[1], -20);
    EXPECT_EQ(tsgc.buffer[2], 40);
    EXPECT_EQ(tsgc.buffer[3], -80);
    EXPECT_EQ(tsgc.buffer[4], 160);
    
    
    ccrb_cb.setCARChannels(carl);
    ASSERT_TRUE(ccrb_cb.mp_car) << "member pointer to CAR should be set now";
    ++ pk1.time;
    ASSERT_TRUE(ccrb_cb.addSamplesFromPacket(&pk1)) << "Accept packet";
    ++ pk1.time;
    ccrb_cb.disableCAR();
    ASSERT_FALSE(ccrb_cb.mp_car) << "member pointer to CAR should not be set now";
    ASSERT_TRUE(ccrb_cb.addSamplesFromPacket(&pk1)) << "Accept packet";
    ASSERT_TRUE(ccrb_cb.consume(tsgc)) << "Consumption of buffer should succeed";
    
    ASSERT_EQ(tsgc.buffer.size(), 10) << "There should be 5 samples in buffer (1 time, 5 channels)";
    EXPECT_EQ(tsgc.buffer[0], 45);
    EXPECT_EQ(tsgc.buffer[1], 15);
    EXPECT_EQ(tsgc.buffer[2], 75);
    EXPECT_EQ(tsgc.buffer[3], -45);
    EXPECT_EQ(tsgc.buffer[4], 195);
    EXPECT_EQ(tsgc.buffer[5], 10);
    EXPECT_EQ(tsgc.buffer[6], -20);
    EXPECT_EQ(tsgc.buffer[7], 40);
    EXPECT_EQ(tsgc.buffer[8], -80);
    EXPECT_EQ(tsgc.buffer[9], 160);
    
    
    ASSERT_FALSE(ccrb_cb.setCARChannels(carlEmpty));
    ++ pk1.time;
    ASSERT_FALSE(ccrb_cb.mp_car) << "member pointer to CAR should not be set now";
    ASSERT_TRUE(ccrb_cb.addSamplesFromPacket(&pk1)) << "Accept packet";
    ASSERT_TRUE(ccrb_cb.consume(tsgc)) << "Consumption of buffer should succeed";
    ASSERT_EQ(tsgc.buffer.size(), 5) << "There should be 5 samples in buffer (1 time, 5 channels)";
    EXPECT_EQ(tsgc.buffer[0], 10);
    EXPECT_EQ(tsgc.buffer[1], -20);
    EXPECT_EQ(tsgc.buffer[2], 40);
    EXPECT_EQ(tsgc.buffer[3], -80);
    EXPECT_EQ(tsgc.buffer[4], 160);
    
    ASSERT_FALSE(ccrb_cb.setCARChannels(carlInvalid));
    ++ pk1.time;
    ASSERT_FALSE(ccrb_cb.mp_car) << "member pointer to CAR should not be set now";
    ASSERT_TRUE(ccrb_cb.addSamplesFromPacket(&pk1)) << "Accept packet";
    ASSERT_TRUE(ccrb_cb.consume(tsgc)) << "Consumption of buffer should succeed";
    ASSERT_EQ(tsgc.buffer.size(), 5) << "There should be 5 samples in buffer (1 time, 5 channels)";
    EXPECT_EQ(tsgc.buffer[0], 10);
    EXPECT_EQ(tsgc.buffer[1], -20);
    EXPECT_EQ(tsgc.buffer[2], 40);
    EXPECT_EQ(tsgc.buffer[3], -80);
    EXPECT_EQ(tsgc.buffer[4], 160);
    
    
}

class ContRBC : public ::testing::Test
{
protected:
    virtual void SetUp() override
    {
        sgrbc.setSamplePeriodForGroup(1, 30);
        sgrbc.setSamplePeriodForGroup(2, 15);
        sgrbc.setSamplePeriodForGroup(3, 1);
    }
    TestSampleGroupRBC<6, 5> sgrbc;
};

TEST_F(ContRBC, Init)
{
    ChannelNumList_t ccl1 = {1, 2, 6, 7, 10};
    ChannelNumList_t ccl2 = {2, 6, 8, 10, 11};
    ChannelNumList_t ccl3 = {22, 26, 28, 210, 21};
    
    sgrbc.reset(1, ccl1);
    sgrbc.reset(2, ccl2);
    sgrbc.reset(3, ccl3);
}



} /* cc */
