#include "gtest/gtest.h"
#include "sbpe/SpikeBandPowerEstimator.h"
namespace cc
{
namespace sbpe
{

    using FiltRMSSamplerD = FiltRMSSampler<double, double, double>;

class SpikeBandPowerEstimatorBasic : public ::testing::Test
{
protected:
    virtual void SetUp() override
    {
        spdlog::set_level(spdlog::level::debug);
        spdlog::set_pattern("[%Y-%m-%d %H:%M:%S.%f] [%n] [%^%l%$] [PID:%P] [TID:%t] %v");

        mp_logger = spdlog::get("testSpikeBandPowerEstimator");
        if (!mp_logger) {
            mp_logger = spdlog::stdout_color_mt<spdlog::async_factory>("testSpikeBandPowerEstimator");
        }

    }
    shared_ptr<spdlog::logger> mp_logger;
};

TEST_F(SpikeBandPowerEstimatorBasic, Create)
{
    {
        SpikeBandPowerEstimator sbpe00;
    }
}

TEST_F(SpikeBandPowerEstimatorBasic, AddSamples)
{
    cbPKT_GROUP pk2 = {121, 0, 6, 3, {9, 0, 12, 8, 19, -1}};
    const ChannelNumList_t ccl1 = {1, 2, 6, 7, 8};
    SpikeBandPowerEstimator sbpe00{ccl1};
    
    sbpe00.reset(ccl1);
    sbpe00.m_keep_going = true;
    ASSERT_TRUE(sbpe00.addSamplesFromPacket(pk2.time, pk2.data, pk2.dlen * 2));
    ++ pk2.time;
    ASSERT_TRUE(sbpe00.addSamplesFromPacket(6, pk2.time, pk2.data, pk2.dlen * 2));
    ++ pk2.time;
    ASSERT_TRUE(sbpe00.addSamplesFromPacket(&pk2));
}


TEST_F(SpikeBandPowerEstimatorBasic, AcceptedSampleGroup)
{
    const ChannelNumList_t ccl1 = {1, 2};
    const ChannelNumList_t ccl2 = {71, 82};
    
    SpikeBandPowerEstimator sbpe {ccl1};
    sbpe.setAcceptedSampleGroup(6);
    sbpe.reset(ccl1);
    ASSERT_EQ(sbpe.getAcceptedSampleGroup(), 6);
    sbpe.m_keep_going = true;
        
    cbPKT_GROUP pk1 = {121, 0, 6, 3, {9, 0, 12, 8, 19, -1}};
    cbPKT_GROUP pk2 = {122, 0, 5, 3, {9, 0, 12, 8, 19, -1}};
    
    ASSERT_TRUE(sbpe.addSamplesFromPacket(&pk1));
    ASSERT_FALSE(sbpe.addSamplesFromPacket(&pk2));
    
    sbpe.setAcceptedSampleGroup(5);
    sbpe.reset(ccl2);
    ASSERT_EQ(sbpe.getAcceptedSampleGroup(), 5);
    
    pk1.time += 2;
    pk2.time += 2;
    ASSERT_FALSE(sbpe.addSamplesFromPacket(&pk1));
    ASSERT_TRUE(sbpe.addSamplesFromPacket(&pk2));
    
    
}

TEST_F(SpikeBandPowerEstimatorBasic, SetChannels)
{
    SpikeBandPowerEstimator sbpe;
    SpikeBandPowerEstimator::ch_list_type chlist{1, 2, 4, 5};
    sbpe.setChannelList(chlist);
}

TEST_F(SpikeBandPowerEstimatorBasic, GetChannels)
{
    SpikeBandPowerEstimator::ch_list_type chlist{3, 1, 2, 4};
    SpikeBandPowerEstimator sbpe{chlist, 10ms};
    
    
    // EXPECT_EQ(sre.mp_unit_hist_container->size(), 4);
    
    auto ch_map_pair = sbpe.getChannelMapping();
    
    auto ch_list = ch_map_pair.first;
    auto ch_map = ch_map_pair.second;
    
    EXPECT_EQ(ch_list.size(), 4);
    EXPECT_EQ(ch_map.size(), 4);
    
    EXPECT_EQ(ch_list[0], 3);
    EXPECT_EQ(ch_list[1], 1);
    EXPECT_EQ(ch_list[2], 2);
    EXPECT_EQ(ch_list[3], 4);
    
    EXPECT_EQ(ch_map.at(3), 0);
    EXPECT_EQ(ch_map.at(1), 1);
    EXPECT_EQ(ch_map.at(2), 2);
    EXPECT_EQ(ch_map.at(4), 3);
    EXPECT_THROW(ch_map.at(0), std::exception);
    EXPECT_THROW(ch_map.at(5), std::exception);
    
}



TEST_F(SpikeBandPowerEstimatorBasic, AddContSamples)
{
    cbPKT_GROUP pk01 = {0, 0, 1, 2, {1,   0, -1,  1}};
    cbPKT_GROUP pk02 = {0, 0, 1, 2, {0,   0,  1,  2}};
    cbPKT_GROUP pk03 = {0, 0, 1, 2, {1,   0, -1,  3}};
    cbPKT_GROUP pk04 = {0, 0, 1, 2, {0,   0,  1,  4}};
    uint32_t ts = 100;
    const ChannelNumList_t ccl1 = {1, 3, 5, 7};
    
    auto mysbpRB = std::make_shared<SpikeBandPowerEstimator::sbp_ring_buffer_type>();

    // SpikeBandPowerEstimator::ch_list_type chlist{0, 2, 4, 5};
    // SpikeBandPowerEstimator::ch_list_type chlist{0, 2, 4, 6};
    SpikeBandPowerEstimator sbpe{ccl1, 10ms};
    sbpe.m_keep_going = true;
    sbpe.setIIRFilterCoeffs({1}, {1});
    sbpe.setIIRFilterNSamples(4);
    
    sbpe.attachSpikeBandPowerRB(mysbpRB);
    
    sbpe.reset(ccl1);
    EXPECT_TRUE(sbpe.addSamplesFromPacket(++ts, pk01.data, pk01.dlen * 2));
    sbpe.applyFilters();
    EXPECT_TRUE(sbpe.addSamplesFromPacket(++ts, pk02.data, pk02.dlen * 2));
    EXPECT_TRUE(sbpe.addSamplesFromPacket(++ts, pk03.data, pk03.dlen * 2));
    EXPECT_TRUE(sbpe.addSamplesFromPacket(++ts, pk04.data, pk04.dlen * 2));
    
    sbpe.applyFilters();
    
    SpikeBandPowerEstimator::sbp_ring_buffer_type::VecT sbp;

    EXPECT_EQ(sbp.size(), 0) << "Buffer should first be empty";
    mysbpRB->consume(sbp);
    ASSERT_EQ(sbp.size(), 1) << "Buffer should have one entry after consuming the data";
    
    auto cit = sbp[0].second.cbegin();
    EXPECT_DOUBLE_EQ(*cit, sqrt(.5));
    EXPECT_DOUBLE_EQ(*(++cit), 0);
    EXPECT_DOUBLE_EQ(*(++cit), 1);
    EXPECT_DOUBLE_EQ(*(++cit), sqrt(15.0/2.0));
    
    EXPECT_TRUE(sbpe.addSamplesFromPacket(++ts, pk01.data, pk01.dlen * 2));
    EXPECT_TRUE(sbpe.addSamplesFromPacket(++ts, pk02.data, pk02.dlen * 2));
    EXPECT_TRUE(sbpe.addSamplesFromPacket(++ts, pk03.data, pk03.dlen * 2));
    EXPECT_TRUE(sbpe.addSamplesFromPacket(++ts, pk04.data, pk04.dlen * 2));
    EXPECT_TRUE(sbpe.addSamplesFromPacket(++ts, pk01.data, pk01.dlen * 2));
    EXPECT_TRUE(sbpe.addSamplesFromPacket(++ts, pk02.data, pk02.dlen * 2));
    EXPECT_TRUE(sbpe.addSamplesFromPacket(++ts, pk03.data, pk03.dlen * 2));
    EXPECT_TRUE(sbpe.addSamplesFromPacket(++ts, pk04.data, pk04.dlen * 2));
    EXPECT_TRUE(sbpe.addSamplesFromPacket(++ts, pk01.data, pk01.dlen * 2));
    sbpe.applyFilters();
    
    mysbpRB->consume(sbp);
    EXPECT_EQ(sbp.size(), 2) << "Buffer should have two entries after consuming the data";
    
    for (size_t i = 0; i < sbp.size(); ++ i)
    {
        cit = sbp[i].second.cbegin();
        EXPECT_DOUBLE_EQ(*cit, sqrt(.5));
        EXPECT_DOUBLE_EQ(*(++cit), 0);
        EXPECT_DOUBLE_EQ(*(++cit), 1);
        EXPECT_DOUBLE_EQ(*(++cit), sqrt(15.0/2.0));
    }
}

TEST_F(SpikeBandPowerEstimatorBasic, AddContSamplesWithChannelChange1)
{
    uint32_t ts = 100;
    
    const cbPKT_GROUP pk01 = {0, 0, 1, 2, {1,   0, -1,  1}};
    const cbPKT_GROUP pk02 = {0, 0, 1, 2, {0,   0,  1,  2}};
    const cbPKT_GROUP pk03 = {0, 0, 1, 2, {1,   0, -1,  3}};
    const cbPKT_GROUP pk11 = {0, 0, 1, 3, {0,   0,  1,  4, 1, -1}};
    const ChannelNumList_t ccl01 = {1, 3, 5, 7};
    const ChannelNumList_t ccl02 = {2, 1, 3, 5, 7};
    
    auto mysbpRB = std::make_shared<SpikeBandPowerEstimator::sbp_ring_buffer_type>();

    // SpikeBandPowerEstimator::ch_list_type chlist{0, 2, 4, 5};
    const ChannelNumList_t chlist{1, 3, 5, 7};
    SpikeBandPowerEstimator sbpe{chlist, 10ms};
    sbpe.m_keep_going = true;
    sbpe.setIIRFilterCoeffs({1}, {1});
    sbpe.setIIRFilterNSamples(2);
    
    sbpe.attachSpikeBandPowerRB(mysbpRB);
    sbpe.reset(ccl01);
    EXPECT_TRUE(sbpe.addSamplesFromPacket(++ts, pk01.data, pk01.dlen * 2));
    EXPECT_TRUE(sbpe.addSamplesFromPacket(++ts, pk02.data, pk02.dlen * 2));
    EXPECT_TRUE(sbpe.addSamplesFromPacket(++ts, pk03.data, pk03.dlen * 2));
    sbpe.reset(ccl02);
    EXPECT_TRUE(sbpe.addSamplesFromPacket(++ts, pk11.data, pk11.dlen * 2));
    EXPECT_TRUE(sbpe.addSamplesFromPacket(++ts, pk11.data, pk11.dlen * 2));
    sbpe.applyFilters();
    
    SpikeBandPowerEstimator::sbp_ring_buffer_type::VecT sbp;

    EXPECT_EQ(sbp.size(), 0) << "Buffer should first be empty";
    mysbpRB->consume(sbp);
    ASSERT_EQ(sbp.size(), 1) << "Buffer should have one entry after consuming the data";
    
    auto cit = sbp[0].second.cbegin();
    EXPECT_DOUBLE_EQ(*cit, 0);
    EXPECT_DOUBLE_EQ(*(++cit), 1);
    EXPECT_DOUBLE_EQ(*(++cit), 4);
    EXPECT_DOUBLE_EQ(*(++cit), 1.0);
    
}

TEST_F(SpikeBandPowerEstimatorBasic, AddContSamplesWithChannelChange2)
{
    uint32_t ts = 100;
    const cbPKT_GROUP pk01 = {0, 0, 1, 2, {1, 0, -1,  1}};
    const cbPKT_GROUP pk02 = {0, 0, 1, 2, {0, 0,  1,  2}};
    const cbPKT_GROUP pk03 = {0, 0, 1, 2, {1, 0, -1,  3}};
    const cbPKT_GROUP pk04 = {0, 0, 1, 2, {0, 0,  1,  4}};
    const ChannelNumList_t ccl01 = {1, 3, 5, 7};
    
    auto mysbpRB = std::make_shared<SpikeBandPowerEstimator::sbp_ring_buffer_type>();

    const ChannelNumList_t chlist01{1, 5, 7, 8};
    const ChannelNumList_t chlist02{1, 3, 5, 7};
    SpikeBandPowerEstimator sbpe{chlist01, 10ms};
    sbpe.m_keep_going = true;
    sbpe.setIIRFilterCoeffs({1}, {1});
    sbpe.setIIRFilterNSamples(2);
    
    sbpe.attachSpikeBandPowerRB(mysbpRB);
    sbpe.reset(ccl01);
    sbpe.addSamplesFromPacket(++ts, pk01.data, pk01.dlen*2);
    sbpe.addSamplesFromPacket(++ts, pk02.data, pk02.dlen*2);
    sbpe.addSamplesFromPacket(++ts, pk03.data, pk03.dlen*2);
    sbpe.applyFilters();
    sbpe.setChannelList(chlist02);
    sbpe.addSamplesFromPacket(++ts, pk04.data, pk04.dlen*2);
    sbpe.addSamplesFromPacket(++ts, pk01.data, pk01.dlen*2);
    sbpe.applyFilters();
    
    SpikeBandPowerEstimator::sbp_ring_buffer_type::VecT sbp;

    EXPECT_EQ(sbp.size(), 0) << "Buffer should first be empty";
    mysbpRB->consume(sbp);
    ASSERT_EQ(sbp.size(), 1) << "Buffer should have one entry after consuming the data";
    
    auto cit = sbp[0].second.cbegin();
    EXPECT_DOUBLE_EQ(*cit, sqrt(0.50));
    EXPECT_DOUBLE_EQ(*(++cit), 0);
    EXPECT_DOUBLE_EQ(*(++cit), 1);
    EXPECT_DOUBLE_EQ(*(++cit), sqrt(17.0/2.0));
    
}


TEST_F(SpikeBandPowerEstimatorBasic, AvgBins)
{
    cbPKT_GROUP pk01 = {0, 0, 1, 2, {1,   0, -1,  1}};
    cbPKT_GROUP pk02 = {0, 0, 1, 2, {0,   0,  1,  2}};
    cbPKT_GROUP pk03 = {0, 0, 1, 2, {1,   0, -1,  3}};
    cbPKT_GROUP pk04 = {0, 0, 1, 2, {0,   0,  1,  4}};
    uint32_t ts = 100;
    const ChannelNumList_t ccl1 = {1, 3, 5, 7};
    
    auto mysbpRB = std::make_shared<SpikeBandPowerEstimator::sbp_ring_buffer_type>();

    // SpikeBandPowerEstimator::ch_list_type chlist{0, 2, 4, 5};
    // SpikeBandPowerEstimator::ch_list_type chlist{0, 2, 4, 6};
    SpikeBandPowerEstimator sbpe{ccl1, 10ms};
    sbpe.m_keep_going = true;
    sbpe.setIIRFilterCoeffs({1}, {1});
    sbpe.setIIRFilterNSamples(2);
    sbpe.setRMSNBins(5);
    
    sbpe.attachSpikeBandPowerRB(mysbpRB);
    
    sbpe.reset(ccl1);
    EXPECT_TRUE(sbpe.addSamplesFromPacket(++ts, pk01.data, pk01.dlen * 2));
    sbpe.applyFilters();
    EXPECT_TRUE(sbpe.addSamplesFromPacket(++ts, pk02.data, pk02.dlen * 2));
    EXPECT_TRUE(sbpe.addSamplesFromPacket(++ts, pk03.data, pk03.dlen * 2));
    EXPECT_TRUE(sbpe.addSamplesFromPacket(++ts, pk04.data, pk04.dlen * 2));
    
    sbpe.applyFilters();
    
    SpikeBandPowerEstimator::sbp_ring_buffer_type::VecT sbp;

    EXPECT_EQ(sbp.size(), 0) << "Buffer should first be empty";
    mysbpRB->consume(sbp);
    ASSERT_EQ(sbp.size(), 2) << "Buffer should have two entries after consuming the data";
    
    auto cit = sbp[0].second.cbegin();
    EXPECT_DOUBLE_EQ(*cit, sqrt(0.5/5.0));
    EXPECT_DOUBLE_EQ(*(++cit), 0);
    EXPECT_DOUBLE_EQ(*(++cit), sqrt(1.0/5.0));
    EXPECT_DOUBLE_EQ(*(++cit), sqrt(2.5/5.0));

    cit = sbp[1].second.cbegin();
    EXPECT_DOUBLE_EQ(*cit, sqrt(1.0/5.0));
    EXPECT_DOUBLE_EQ(*(++cit), 0);
    EXPECT_DOUBLE_EQ(*(++cit), sqrt(2.0/5.0));
    EXPECT_DOUBLE_EQ(*(++cit), sqrt(15.0/5.0));
    
    EXPECT_TRUE(sbpe.addSamplesFromPacket(++ts, pk01.data, pk01.dlen * 2));
    EXPECT_TRUE(sbpe.addSamplesFromPacket(++ts, pk02.data, pk02.dlen * 2));
    EXPECT_TRUE(sbpe.addSamplesFromPacket(++ts, pk03.data, pk03.dlen * 2));
    EXPECT_TRUE(sbpe.addSamplesFromPacket(++ts, pk04.data, pk04.dlen * 2));
    EXPECT_TRUE(sbpe.addSamplesFromPacket(++ts, pk01.data, pk01.dlen * 2));
    EXPECT_TRUE(sbpe.addSamplesFromPacket(++ts, pk02.data, pk02.dlen * 2));
    EXPECT_TRUE(sbpe.addSamplesFromPacket(++ts, pk03.data, pk03.dlen * 2));
    EXPECT_TRUE(sbpe.addSamplesFromPacket(++ts, pk04.data, pk04.dlen * 2));
    EXPECT_TRUE(sbpe.addSamplesFromPacket(++ts, pk01.data, pk01.dlen * 2));
    sbpe.applyFilters();

    mysbpRB->consume(sbp);
    EXPECT_EQ(sbp.size(), 4) << "Buffer should have four entries after consuming the data";

    cit = sbp[0].second.cbegin();
    EXPECT_DOUBLE_EQ(*cit, sqrt(1.5 / 5.0));
    EXPECT_DOUBLE_EQ(*(++cit), 0);
    EXPECT_DOUBLE_EQ(*(++cit), sqrt(3.0 / 5.0));
    EXPECT_DOUBLE_EQ(*(++cit), sqrt(17.50 / 5.0));

    cit = sbp[1].second.cbegin();
    EXPECT_DOUBLE_EQ(*cit, sqrt(2.0 / 5.0));
    EXPECT_DOUBLE_EQ(*(++cit), 0);
    EXPECT_DOUBLE_EQ(*(++cit), sqrt(4.0 / 5.0));
    EXPECT_DOUBLE_EQ(*(++cit), sqrt(30.0 / 5.0));

    cit = sbp[2].second.cbegin();
    EXPECT_DOUBLE_EQ(*cit, sqrt(2.50 / 5.0));
    EXPECT_DOUBLE_EQ(*(++cit), 0);
    EXPECT_DOUBLE_EQ(*(++cit), 1.0);
    EXPECT_DOUBLE_EQ(*(++cit), sqrt(32.50 / 5.0));

    cit = sbp[3].second.cbegin();
    EXPECT_DOUBLE_EQ(*cit, sqrt(2.50 / 5.0));
    EXPECT_DOUBLE_EQ(*(++cit), 0);
    EXPECT_DOUBLE_EQ(*(++cit), 1.0);
    EXPECT_DOUBLE_EQ(*(++cit), sqrt(42.50 / 5.0));

}



TEST_F(SpikeBandPowerEstimatorBasic, ProcessThread)
{
    const cbPKT_GROUP pk01 = {120, 0, 1, 2, {1,   0, -1,  1}};
    const cbPKT_GROUP pk02 = {121, 0, 1, 2, {0,   0,  1,  2}};
    const cbPKT_GROUP pk03 = {122, 0, 1, 2, {1,   0, -1,  3}};
    const cbPKT_GROUP pk04 = {123, 0, 1, 2, {0,   0,  1,  4}};
    const ChannelNumList_t ccl01 = {1, 3, 5, 7};
    
    auto mysbpRB = std::make_shared<SpikeBandPowerEstimator::sbp_ring_buffer_type>();

    SpikeBandPowerEstimator sbpe{ccl01, 10ms};
    sbpe.setIIRFilterCoeffs({1}, {1});
    sbpe.setIIRFilterNSamples(600);
    
    sbpe.attachSpikeBandPowerRB(mysbpRB);
    
    sbpe.signalStart();
    sbpe.reset(ccl01);
    sbpe.addSamplesFromPacket(pk01.time, pk01.data, pk01.dlen * 2);
    
    
    auto last_tm = SpikeBandPowerEstimator::clock::now();
    Timestamp_t smp_tm = pk01.time + 1;
    SpikeBandPowerEstimator::sbp_ring_buffer_type::VecT sbp;
    std::ostringstream oss;
    mp_logger->set_level(spdlog::level::info);
    
    for (size_t i=0; i < 100500; ++i)
    {
        sbpe.addSamplesFromPacket(++smp_tm, pk02.data, pk02.dlen * 2);
        sbpe.addSamplesFromPacket(++smp_tm, pk03.data, pk03.dlen * 2);
        sbpe.addSamplesFromPacket(++smp_tm, pk04.data, pk04.dlen * 2);
        sbpe.addSamplesFromPacket(++smp_tm, pk01.data, pk01.dlen * 2);
        
        if (last_tm + 50ms < SpikeBandPowerEstimator::clock::now())
        {
            mysbpRB->consume(sbp);
            for (size_t ii = 0; ii < sbp.size(); ++ii)
            {
                auto cit = sbp[ii].second.cbegin();
                EXPECT_NEAR(*cit, sqrt(0.50), 1e-5);
                EXPECT_NEAR(*(++cit), 0, 1e-5);
                EXPECT_NEAR(*(++cit), 1, 1e-5);
                EXPECT_NEAR(*(++cit), sqrt(15.0/2.0), 1e-5);
        
                oss.str("");
                oss.clear();
        
                std::for_each(sbp[ii].second.cbegin(), sbp[ii].second.cend(), [&oss](auto x){oss << x << ", ";} );
                mp_logger->debug("RMS: {}: [{}]", sbp[ii].first, oss.str());
            }
            last_tm = SpikeBandPowerEstimator::clock::now();
        }
        
        std::this_thread::sleep_for(std::chrono::microseconds(33));
        
    }
    std::this_thread::sleep_for(std::chrono::seconds(1));

    mp_logger->info("collecting remaining SBP snippets");

    mysbpRB->consume(sbp);
    for (size_t ii = 0; ii < sbp.size(); ++ii)
    {
        auto cit = sbp[ii].second.cbegin();
        EXPECT_NEAR(*cit, sqrt(0.50), 1e-5);
        EXPECT_NEAR(*(++cit), 0, 1e-5);
        EXPECT_NEAR(*(++cit), 1, 1e-5);
        EXPECT_NEAR(*(++cit), sqrt(15.0/2.0), 1e-5);

        oss.str("");
        oss.clear();

        std::for_each(sbp[ii].second.cbegin(), sbp[ii].second.cend(), [&oss](auto x){oss << x << ", ";} );
        mp_logger->info("RMS: {}: [{}]", sbp[ii].first, oss.str());
    }
}

TEST_F(SpikeBandPowerEstimatorBasic, SelectSampleGroup)
{
    const cbPKT_GROUP pk01 = {120, 0, 1, 2, {1,   0, -1,  1}};
    auto mysbpRB = std::make_shared<SpikeBandPowerEstimator::sbp_ring_buffer_type>();

    const SpikeBandPowerEstimator::ch_list_type chlist01{1, 3, 5, 7};
    SpikeBandPowerEstimator sbpe{chlist01, 10ms};
    sbpe.m_keep_going = true;
    sbpe.setIIRFilterCoeffs({1}, {1});
    sbpe.setIIRFilterNSamples(1);
    sbpe.attachSpikeBandPowerRB(mysbpRB);
    
    EXPECT_EQ(sbpe.getAcceptedSampleGroup(), 6);
    EXPECT_TRUE(sbpe.addSamplesFromPacket(6, 1, pk01.data, pk01.dlen * 2));
    EXPECT_FALSE(sbpe.addSamplesFromPacket(5, 2, pk01.data, pk01.dlen * 2));
    sbpe.setAcceptedSampleGroup(5);
    EXPECT_EQ(sbpe.getAcceptedSampleGroup(), 5);
    EXPECT_TRUE(sbpe.addSamplesFromPacket(5, 3, pk01.data, pk01.dlen * 2));
    EXPECT_FALSE(sbpe.addSamplesFromPacket(6, 4, pk01.data, pk01.dlen * 2));
}

TEST_F(SpikeBandPowerEstimatorBasic, OverFlow)
{
    cbPKT_GROUP pk01 = {1, 0, 6, 2, {1,   0, -1,  1}};
    auto mysbpRB = std::make_shared<SpikeBandPowerEstimator::sbp_ring_buffer_type>();
    const SpikeBandPowerEstimator::ch_list_type chlist01{1, 3, 5, 7};
    SpikeBandPowerEstimator sbpe{chlist01, 1ms};
    sbpe.m_keep_going = true;
    sbpe.setIIRFilterCoeffs({1}, {1});
    sbpe.setIIRFilterNSamples(30);
    sbpe.attachSpikeBandPowerRB(mysbpRB);
    sbpe.reset(chlist01);
    EXPECT_EQ(sbpe.getAcceptedSampleGroup(), 6);

    for (int i = 0; i < 60000; ++i)
    {
        ++pk01.time;
        EXPECT_TRUE(sbpe.addSamplesFromPacket(&pk01));
        sbpe.applyFilters();
    }
    SpikeBandPowerEstimator::sbp_ring_buffer_type::VecT sbp;
    size_t overflow_count;
    mysbpRB->consume(sbp, overflow_count);
    EXPECT_GE(sbp.size(), mysbpRB->size()) << "Buffer should be at capacity";
    EXPECT_GE(overflow_count, 1);

    mp_logger->debug("SBP RB size: {}", sbp.size());
    for (int i = 0; i < 600; ++i)
    {
        ++pk01.time;
        EXPECT_TRUE(sbpe.addSamplesFromPacket(&pk01));
        sbpe.applyFilters();
    }
    mysbpRB->consume(sbp, overflow_count);
    EXPECT_GE(sbp.size(), 1) << "Even after buffer overflow, SBPE should continue";
    EXPECT_EQ(overflow_count, 0);
    mp_logger->debug("SBP RB size: {}", sbp.size());
}

TEST_F(SpikeBandPowerEstimatorBasic, ProcessThreadOverflow)
{
    cbPKT_GROUP pk01 = {1, 0, 6, 2, {1,   0, -1,  1}};
    const ChannelNumList_t ccl01 = {1, 3, 5, 7};

    auto mysbpRB = std::make_shared<SpikeBandPowerEstimator::sbp_ring_buffer_type>();

    SpikeBandPowerEstimator sbpe{ccl01, 1ms};
    sbpe.setIIRFilterCoeffs({1}, {1});
    sbpe.setIIRFilterNSamples(10);

    sbpe.attachSpikeBandPowerRB(mysbpRB);

    sbpe.signalStart();
    sbpe.reset(ccl01);
    sbpe.addSamplesFromPacket(&pk01);

    auto last_tm = SpikeBandPowerEstimator::clock::now();
    SpikeBandPowerEstimator::sbp_ring_buffer_type::VecT sbp;
    std::ostringstream oss;
    mp_logger->set_level(spdlog::level::info);

    for (size_t i=0; i < 1100; ++i)
    {
        for (size_t j=0; j < 10; ++j) {
            ++pk01.time;
            EXPECT_TRUE(sbpe.addSamplesFromPacket(&pk01));
        }
        std::this_thread::sleep_for(std::chrono::microseconds(100));
    }
    std::this_thread::sleep_for(std::chrono::milliseconds(1));
    size_t overflow_count;
    mysbpRB->consume(sbp, overflow_count);
    EXPECT_GE(sbp.size(), mysbpRB->size()) << "Buffer should be at capacity";
    EXPECT_GE(overflow_count, 100) << "sbp ring buffer should have missed 100 items";

    for (size_t i=0; i < 10; ++i)
    {
        for (size_t j=0; j < 10; ++j) {
            ++pk01.time;
            EXPECT_TRUE(sbpe.addSamplesFromPacket(&pk01));
        }
        std::this_thread::sleep_for(std::chrono::microseconds(2000));
    }
    std::this_thread::sleep_for(std::chrono::milliseconds(1));
    mysbpRB->consume(sbp, overflow_count);
    EXPECT_GE(sbp.size(), 10) << "Buffer should be at capacity";
    EXPECT_GE(overflow_count, 0) << "sbp ring buffer should have missed 100 items";

}


TEST_F(SpikeBandPowerEstimatorBasic, ChannelMapping)
{
    cbPKT_GROUP pk01 = {1, 0, 5, 3, {100, 200, 300, 400, 500, 600}};
    cbPKT_GROUP pk02 = {2, 0, 5, 3, {1000, 1010, 1020, 1030, 1040, 1050}};
    auto mysbpRB = std::make_shared<SpikeBandPowerEstimator::sbp_ring_buffer_type>();
    SpikeBandPowerEstimator::sbp_ring_buffer_type::VecT sbp;

    const ChannelNumList_t chlist01{1, 2, 3, 8};
    const ChannelNumList_t chlist02{7, 3, 5, 6};
    const ChannelNumList_t chlist0a{1, 2, 3, 4, 5, 6};
    const ChannelNumList_t chlist0b{3, 4, 5, 6, 8, 7};

    SpikeBandPowerEstimator sbpe{chlist01, 10ms};
    sbpe.setIIRFilterCoeffs({1}, {1});
    sbpe.setIIRFilterNSamples(1);
    sbpe.setAcceptedSampleGroup(5);
    sbpe.reset(chlist0a);
    sbpe.m_keep_going = true;
    sbpe.attachSpikeBandPowerRB(mysbpRB);
    
    EXPECT_TRUE(sbpe.addSamplesFromPacket(&pk01));
    sbpe.applyFilters();
    mysbpRB->consume(sbp);
    ASSERT_EQ(sbp.size(), 1) << "Buffer should have one entry after consuming the data";
    
    auto cit = sbp[0].second.cbegin();
    EXPECT_DOUBLE_EQ(*cit, 100);
    EXPECT_DOUBLE_EQ(*(++cit), 200);
    EXPECT_DOUBLE_EQ(*(++cit), 300);
    EXPECT_DOUBLE_EQ(*(++cit), 0);
    
    sbpe.setChannelList(chlist02);
    ++ pk01.time;
    EXPECT_TRUE(sbpe.addSamplesFromPacket(&pk01));
    sbpe.applyFilters();
    mysbpRB->consume(sbp);
    ASSERT_EQ(sbp.size(), 1) << "Buffer should have one entry after consuming the data";
    cit = sbp[0].second.cbegin();
    EXPECT_DOUBLE_EQ(*cit, 0);
    EXPECT_DOUBLE_EQ(*(++cit), 300);
    EXPECT_DOUBLE_EQ(*(++cit), 500);
    EXPECT_DOUBLE_EQ(*(++cit), 600);
    
    sbpe.reset(chlist0b);
    pk02.time = pk01.time + 1;
    EXPECT_TRUE(sbpe.addSamplesFromPacket(&pk02));
    sbpe.applyFilters();
    mysbpRB->consume(sbp);
    ASSERT_EQ(sbp.size(), 1) << "Buffer should have one entry after consuming the data";
    cit = sbp[0].second.cbegin();
    EXPECT_DOUBLE_EQ(*cit, 1050);
    EXPECT_DOUBLE_EQ(*(++cit), 1000);
    EXPECT_DOUBLE_EQ(*(++cit), 1020);
    EXPECT_DOUBLE_EQ(*(++cit), 1030);
    
    sbpe.setChannelList(chlist01);
    pk02.time = pk02.time + 1;
    EXPECT_TRUE(sbpe.addSamplesFromPacket(&pk02));
    sbpe.applyFilters();
    mysbpRB->consume(sbp);
    ASSERT_EQ(sbp.size(), 1) << "Buffer should have one entry after consuming the data";
    cit = sbp[0].second.cbegin();
    EXPECT_DOUBLE_EQ(*cit, 0);
    EXPECT_DOUBLE_EQ(*(++cit), 0);
    EXPECT_DOUBLE_EQ(*(++cit), 1000);
    EXPECT_DOUBLE_EQ(*(++cit), 1040);
    
}

TEST_F(SpikeBandPowerEstimatorBasic, NBins)
{
    const ChannelNumList_t chlist01{1, 2, 3, 8};
    SpikeBandPowerEstimator sbpe{chlist01, 10ms};
    
    ASSERT_EQ(sbpe.getRMSNBins(), 1);
    sbpe.setRMSNBins(17);
    ASSERT_EQ(sbpe.getRMSNBins(), 17);
    
}

TEST(FiltRMSSamplerBasic, ProcessVector)
{
    FiltRMSSamplerD fs(3);
    fs.setNSamples(4);
    
    std::vector<FiltRMSSamplerD::sample_type> s01{1,0,1};
    std::vector<FiltRMSSamplerD::sample_type> s02{0,0,2};
    std::vector<FiltRMSSamplerD::sample_type> s03{1,0,3};
    std::vector<FiltRMSSamplerD::sample_type> s04{0,0,4};
    
    EXPECT_FALSE(fs(s01.cbegin(), s01.cend()));
    EXPECT_FALSE(fs(s02.cbegin(), s02.cend()));
    EXPECT_FALSE(fs(s03.cbegin(), s03.cend()));
    EXPECT_TRUE(fs(s04.cbegin(), s04.cend()));
    
    auto cit = fs.cbegin();
    EXPECT_DOUBLE_EQ(*cit, sqrt(.5));
    EXPECT_DOUBLE_EQ(*(++cit), 0);
    EXPECT_DOUBLE_EQ(*(++cit), sqrt(15.0/2.0));

}

TEST(FiltRMSSamplerBasic, NBins)
{
    FiltRMSSamplerD fs(2);
    std::vector<FiltRMSSamplerD::sample_type> s01{1,0};
    std::vector<FiltRMSSamplerD::sample_type> s02{-2,0};
    std::vector<FiltRMSSamplerD::sample_type> s03{1,2};
    
    
    ASSERT_EQ(fs.getNAvgBins(), 1);
    fs.setNAvgBins(3);
    ASSERT_EQ(fs.getNAvgBins(), 3);
    fs.setNSamples(4);
    
    EXPECT_FALSE(fs(s01.cbegin(), s01.cend()));
    EXPECT_FALSE(fs(s02.cbegin(), s02.cend()));
    EXPECT_FALSE(fs(s01.cbegin(), s01.cend()));
    EXPECT_TRUE(fs(s02.cbegin(), s02.cend()));
    
    auto cit = fs.cbegin();
    EXPECT_DOUBLE_EQ(*cit, sqrt(2.5/3.0));
    EXPECT_DOUBLE_EQ(*(++cit), 0);
    
    EXPECT_FALSE(fs(s03.cbegin(), s03.cend()));
    EXPECT_FALSE(fs(s02.cbegin(), s02.cend()));
    EXPECT_FALSE(fs(s03.cbegin(), s03.cend()));
    EXPECT_TRUE(fs(s02.cbegin(), s02.cend()));
    
    cit = fs.cbegin();
    EXPECT_DOUBLE_EQ(*cit, sqrt(5.0/3.0));
    EXPECT_DOUBLE_EQ(*(++cit), sqrt(2.0/3.0));

    EXPECT_FALSE(fs(s03.cbegin(), s03.cend()));
    EXPECT_FALSE(fs(s01.cbegin(), s01.cend()));
    EXPECT_FALSE(fs(s03.cbegin(), s03.cend()));
    EXPECT_TRUE(fs(s01.cbegin(), s01.cend()));

    cit = fs.cbegin();
    EXPECT_DOUBLE_EQ(*cit, sqrt(2.0));
    EXPECT_DOUBLE_EQ(*(++cit), sqrt(4.0/3.0));

    EXPECT_FALSE(fs(s03.cbegin(), s03.cend()));
    EXPECT_FALSE(fs(s01.cbegin(), s01.cend()));
    EXPECT_FALSE(fs(s03.cbegin(), s03.cend()));
    EXPECT_TRUE(fs(s01.cbegin(), s01.cend()));

    cit = fs.cbegin();
    EXPECT_DOUBLE_EQ(*cit, sqrt(4.5/3.0));
    EXPECT_DOUBLE_EQ(*(++cit), sqrt(2.0));

}
TEST(FiltRMSSamplerBasic, NBins1Ch)
{
    FiltRMSSamplerD fs(1);
    std::vector<FiltRMSSamplerD::sample_type> s01{1};
    std::vector<FiltRMSSamplerD::sample_type> s02{-2};
    
    
    ASSERT_EQ(fs.getNAvgBins(), 1);
    fs.setNAvgBins(3);
    ASSERT_EQ(fs.getNAvgBins(), 3);
    fs.setNSamples(2);
    
    EXPECT_FALSE(fs(s01.cbegin(), s01.cend()));
    EXPECT_TRUE(fs(s02.cbegin(), s02.cend()));
    
    auto cit = fs.cbegin();
    EXPECT_DOUBLE_EQ(*cit, sqrt(2.5/3.0));
    
    EXPECT_FALSE(fs(s01.cbegin(), s01.cend()));
    EXPECT_TRUE(fs(s02.cbegin(), s02.cend()));
    
    cit = fs.cbegin();
    EXPECT_DOUBLE_EQ(*cit, sqrt(5.0/3.0));

    EXPECT_FALSE(fs(s02.cbegin(), s02.cend()));
    EXPECT_TRUE(fs(s01.cbegin(), s01.cend()));

    cit = fs.cbegin();
    EXPECT_DOUBLE_EQ(*cit, sqrt(7.50/3.0));

    EXPECT_FALSE(fs(s02.cbegin(), s02.cend()));
    EXPECT_TRUE(fs(s01.cbegin(), s01.cend()));

    cit = fs.cbegin();
    EXPECT_DOUBLE_EQ(*cit, sqrt(7.5/3.0));

}

TEST(FiltRMSSamplerBasic, NBinsResize)
{
    FiltRMSSamplerD fs(1);
    std::vector<FiltRMSSamplerD::sample_type> s01{1};
    std::vector<FiltRMSSamplerD::sample_type> s02{-2};
    std::vector<FiltRMSSamplerD::sample_type> s201{1,2};
    std::vector<FiltRMSSamplerD::sample_type> s202{-2,-3};
    
    
    ASSERT_EQ(fs.getNAvgBins(), 1);
    fs.setNAvgBins(3);
    ASSERT_EQ(fs.getNAvgBins(), 3);
    fs.setNSamples(2);
    
    EXPECT_FALSE(fs(s01.cbegin(), s01.cend()));
    EXPECT_TRUE(fs(s02.cbegin(), s02.cend()));
    
    auto cit = fs.cbegin();
    EXPECT_DOUBLE_EQ(*cit, sqrt(2.5/3.0));
    
    fs.resize(2);
    
    EXPECT_FALSE(fs(s201.cbegin(), s201.cend()));
    EXPECT_TRUE(fs(s202.cbegin(), s202.cend()));
    
    cit = fs.cbegin();
    EXPECT_DOUBLE_EQ(*cit, sqrt(2.5/3.0));
    EXPECT_DOUBLE_EQ(*(++cit), sqrt(6.5/3.0));
    

    EXPECT_FALSE(fs(s202.cbegin(), s202.cend()));
    EXPECT_TRUE(fs(s201.cbegin(), s201.cend()));

    cit = fs.cbegin();
    EXPECT_DOUBLE_EQ(*cit, sqrt(5.0/3.0));
    EXPECT_DOUBLE_EQ(*(++cit), sqrt(13.0/3.0));
    
    fs.setNAvgBins(4);
    ASSERT_EQ(fs.getNAvgBins(), 4);
    
    EXPECT_FALSE(fs(s202.cbegin(), s202.cend()));
    EXPECT_TRUE(fs(s201.cbegin(), s201.cend()));
    
    cit = fs.cbegin();
    EXPECT_DOUBLE_EQ(*cit, sqrt(2.5/4.0));
    EXPECT_DOUBLE_EQ(*(++cit), sqrt(6.5/4.0));
    
}



TEST(FiltRMSSamplerBasic, ProcessVector2)
{
    FiltRMSSamplerD fs(3, 4, {1}, {1});
    
    std::vector<FiltRMSSamplerD::sample_type> s01{1,0,1};
    std::vector<FiltRMSSamplerD::sample_type> s02{0,0,2};
    std::vector<FiltRMSSamplerD::sample_type> s03{1,0,3};
    std::vector<FiltRMSSamplerD::sample_type> s04{0,0,4};
    
    EXPECT_FALSE(fs(s01.cbegin(), s01.cend()));
    EXPECT_FALSE(fs(s02.cbegin(), s02.cend()));
    EXPECT_FALSE(fs(s03.cbegin(), s03.cend()));
    EXPECT_TRUE(fs(s04.cbegin(), s04.cend()));
    
    auto cit = fs.cbegin();
    EXPECT_DOUBLE_EQ(*cit, sqrt(.5));
    EXPECT_DOUBLE_EQ(*(++cit), 0);
    EXPECT_DOUBLE_EQ(*(++cit), sqrt(15.0/2.0));

}

} // namespace sbpe
} // namespace cc

