#include "gtest/gtest.h"
#include "sbpe/SpikeBandPowerEstimator.h"

#include <chrono>
#include <functional>
#include <numeric>

namespace cc
{
namespace sbpe
{

TEST(SpikeBandPowerEstimatorPerformance, FastEat)
{
    spdlog::set_level(spdlog::level::info);
    spdlog::set_pattern("[%Y-%m-%d %H:%M:%S.%f] [%n] [%^%l%$] [PID:%P] [TID:%t] %v");
    
    const size_t n_ch = 128, n_samples_len = 600, n_hist = 1000;
    const size_t n_rep = 1000999;
    const size_t n_grp = 5;
    const size_t group = 1;
    
    SpikeBandPowerEstimator sbpe(n_ch, 50ms);
    sbpe.setIIRFilterNSamples(n_samples_len);
    auto sbp_rb = std::make_shared<SpikeBandPowerEstimator::sbp_ring_buffer_type>();
    SpikeBandPowerEstimator::sbp_ring_buffer_type::VecT sbpV;
    sbpV.reserve(sbp_rb->size());
    sbpe.setAcceptedSampleGroup(group);
    sbpe.attachSpikeBandPowerRB(sbp_rb);
    sbpe.m_keep_going = true;
    
    std::vector<cbPKT_GROUP> d_v(n_grp);
    for (size_t i = 0; i < n_grp; ++i)
    {
        d_v[i].dlen = n_ch / 2 + (n_ch % 2 != 0);
        d_v[i].type = group;
    }
    for (size_t i = 0; i < n_ch; ++i)
    {
        d_v[0].data[i] = i % 2;
        d_v[1].data[i] = i % ((i % 4) + 1);
        d_v[2].data[i] = -i % ((i % 7) + 1);
        d_v[3].data[i] = i % ((i % 5) + 1);
        d_v[4].data[i] = i % ((i % 3) + 1);
    }
    
    ChannelNumList_t ch_list(n_ch);
    std::iota (std::begin(ch_list), std::end(ch_list), 1);
    
    sbpe.reset(1, ch_list);
    auto start_sim = std::chrono::high_resolution_clock::now();
    
    
    auto start_add = std::chrono::high_resolution_clock::now();
    auto dur_add = std::chrono::duration_cast<std::chrono::nanoseconds>(start_add - start_add).count();

    auto start_proc = std::chrono::high_resolution_clock::now();
    auto dur_proc = std::chrono::duration_cast<std::chrono::nanoseconds>(start_proc - start_proc).count();

    auto start_cons = std::chrono::high_resolution_clock::now();
    auto dur_cons = std::chrono::duration_cast<std::chrono::nanoseconds>(start_cons - start_cons).count();
    
    size_t sbp_count = 0;
    size_t sbp_apply = 0;

    auto next_loop_time = start_sim;
    for (size_t i = 0; i < n_rep; ++i)
    {
        start_add = std::chrono::high_resolution_clock::now();
        
        d_v[(i + 1) % n_grp].data[0] = i % 300;
        d_v[(i + 1) % n_grp].time = i + 1;
        
        ASSERT_TRUE(sbpe.addSamplesFromPacket(&d_v[(i + 1) % n_grp]));
        
        dur_add += std::chrono::duration_cast<std::chrono::nanoseconds>(std::chrono::high_resolution_clock::now() - start_add).count();
        
        if ((i)%(n_samples_len) == 0)
        {
            start_proc = std::chrono::high_resolution_clock::now();
            sbpe.applyFilters();
            ++ sbp_apply;
            dur_proc += std::chrono::duration_cast<std::chrono::nanoseconds>(std::chrono::high_resolution_clock::now()  - start_proc).count();
            
        }
        
        if ((i+1)%(sbp_rb->size()) == 0)
        {
            start_cons = std::chrono::high_resolution_clock::now();
            sbp_rb->consume(sbpV);
            sbp_count += sbpV.size();
            dur_cons += std::chrono::duration_cast<std::chrono::nanoseconds>(std::chrono::high_resolution_clock::now() - start_cons).count();
        }
    }
    sbpe.signalStop();
    
    std::this_thread::sleep_for(1s);
    auto finish_sim = std::chrono::high_resolution_clock::now();
    
    auto dur_ms_sim = std::chrono::duration_cast<std::chrono::milliseconds>(finish_sim - start_sim).count();

    sbp_rb->consume(sbpV);
    printf("In the loop, we received %lu spike band power values\n", sbp_count);
    
    
    printf("SBP:\n");
    for (auto const& row: sbpV)
    {
        for (auto const& v: row.second)
        {
            printf("%.3f ", v);
        }
        printf("\n");
    }
    
    sbp_count += sbpV.size();
    
    printf("It took %lldms to run %zu loops, or %fms/loop.\n", dur_ms_sim, n_rep, (double)dur_ms_sim / n_rep);
    printf("it took %.1fms to add samples, %.1fms to process sbp (per cycle: %.1fms), and %.1fms to consume the rates.\n", dur_add / (double) n_rep, dur_proc / (double)n_rep, ((double) dur_proc / (double)n_rep) / sbp_apply, dur_cons/ (double) n_rep);
    printf("In total, we received %lu spike band power values\n", sbp_count);
}


}
}

