#include "gtest/gtest.h"
#include "sre/SpikeRateEstimationMethods.h"
#include "sre/UnitHist.h"
namespace cc
{
namespace sre
{

using UH = UnitHist<unsigned int>;
using Container = UH::container_type;

class SREstTest : public ::testing::Test
{
protected:
    std::vector<UH> uhv{{0}, Container{0}, Container{1}, Container{0, 0}, Container{0, 1},
    Container{1, 0}, Container{1, 1}, Container{0, 0, 0}, Container{0, 0, 1}, Container{0, 1, 0},
    Container{1, 0, 0}, Container{1, 1, 1}, Container{0, 0, 0, 0}, Container{0, 0, 0, 1}, Container{0, 0, 1, 0},
    Container{0, 1, 0, 0}, Container{1, 0, 0, 0}, Container{1, 1, 1, 1}, Container{0, 0, 0, 5}, Container{0, 0, 5, 0},
    Container{0, 5, 0, 0}, Container{5, 0, 0, 0}, Container{1, 5, 1, 5}};
};

TEST_F(SREstTest, Id01)
{
    std::vector<double> expected_results{0, 0, 1, 0, 1,
        0, 1, 0, 1, 0,
        0, 1, 0, 1, 0,
        0, 0, 1, 5, 0,
        0, 0, 5};

    auto res_it = expected_results.cbegin();
    for (auto const& u: uhv)
    {
        auto r = r_estimate_id(u.crbegin(), u.crend());
        EXPECT_DOUBLE_EQ(*res_it, r);
        auto r2 = r_estimate_id_w_cont(u);
        EXPECT_DOUBLE_EQ(r, r2);
        ++res_it;
    }
}


TEST_F(SREstTest, ExpDecay01)
{
    const size_t max_length = 1;
    const double decay_factor = .5;
    
    std::vector<double> expected_results{0, 0, 1, 0, 1,
        0, 1, 0, 1, 0,
        0, 1, 0, 1, 0,
        0, 0, 1, 5, 0,
        0, 0, 5};

    auto res_it = expected_results.cbegin();
    for (auto const& u: uhv)
    {
        auto r = r_estimate_with_exp_decay(u.crbegin(), u.crend(), decay_factor, max_length);
        EXPECT_DOUBLE_EQ(*res_it, r);
        auto r2 = r_estimate_with_exp_decay_w_cont(u, decay_factor, max_length);
        EXPECT_DOUBLE_EQ(r, r2);
        ++res_it;
    }
}

TEST_F(SREstTest, ExpDecay02)
{
    const size_t max_length = 2;
    const double decay_factor = .5;
    
    std::vector<double> expected_results{0, 0, 1, 0, 2/3.0,
        1.0/3.0, 1, 0, 2.0/3.0, 1.0/3.0,
        0, 1, 0, 2.0/3.0, 1.0/3.0,
        0, 0, 1, 10.0/3.0, 5.0/3.0,
        0, 0, 11.0/3.0};

    auto res_it = expected_results.cbegin();
    for (auto const& u: uhv)
    {
        auto r = r_estimate_with_exp_decay(u.crbegin(), u.crend(), decay_factor, max_length);
        EXPECT_DOUBLE_EQ(*res_it, r);
        auto r2 = r_estimate_with_exp_decay_w_cont(u, decay_factor, max_length);
        EXPECT_DOUBLE_EQ(r, r2);
        ++res_it;
    }
}

TEST_F(SREstTest, ExpDecay03)
{
    const size_t max_length = 3;
    const double decay_factor = .5;
    
    std::vector<double> expected_results{0, 0, 1, 0, 2/3.0,
        1.0/3.0, 1, 0, (1.0) / (1.0 + 0.5 + 0.25), (0.5) / (1.0 + 0.5 + 0.25),
        (.25) / (1.0 + 0.5 + 0.25), 1, 0, (1.0) / (1.0 + 0.5 + 0.25), (.5) / (1.0 + 0.5 + 0.25),
        (.25) / (1.0 + 0.5 + 0.25), 0, 1, (5.0) / (1.0 + 0.5 + 0.25), (2.5) / (1.0 + 0.5 + 0.25),
        (1.25) / (1.0 + 0.5 + 0.25), 0, (5 + 0.5 + 5.0*.25) / (1.0 + 0.5 + 0.25)};

    auto res_it = expected_results.cbegin();
    for (auto const& u: uhv)
    {
        auto r = r_estimate_with_exp_decay(u.crbegin(), u.crend(), decay_factor, max_length);
        EXPECT_DOUBLE_EQ(*res_it, r);
        auto r2 = r_estimate_with_exp_decay_w_cont(u, decay_factor, max_length);
        EXPECT_DOUBLE_EQ(r, r2);
        ++res_it;
    }
}

TEST_F(SREstTest, ExpDecay04)
{
    const size_t max_length = 4;
    const double decay_factor = .5;
    
    std::vector<double> expected_results{0, 0, 1, 0, 2/3.0,
        1.0/3.0, 1, 0, (1.0) / (1.0 + 0.5 + 0.25), (0.5) / (1.0 + 0.5 + 0.25),
        (.25) / (1.0 + 0.5 + 0.25), 1, 0, (1.0) / (1.0 + 0.5 + 0.25 + 0.125), (.5) / (1.0 + 0.5 + 0.25 + 0.125),
        (.25) / (1.0 + 0.5 + 0.25 + 0.125), (.125) / (1.0 + 0.5 + 0.25 + 0.125), 1, (5.0) / (1.0 + 0.5 + 0.25 + 0.125), (2.5) / (1.0 + 0.5 + 0.25 + 0.125),
        (1.25) / (1.0 + 0.5 + 0.25 + 0.125), (1.25*.5) / (1.0 + 0.5 + 0.25 + 0.125), (5 + 0.5 + 5.0*.25 + 0.125) / (1.0 + 0.5 + 0.25 + 0.125)};

    auto res_it = expected_results.cbegin();
    for (auto const& u: uhv)
    {
        auto r = r_estimate_with_exp_decay(u.crbegin(), u.crend(), decay_factor, max_length);
        EXPECT_DOUBLE_EQ(*res_it, r);
        auto r2 = r_estimate_with_exp_decay_w_cont(u, decay_factor, max_length);
        EXPECT_DOUBLE_EQ(r, r2);
        // printf("%li\t%f\t%f\n", res_it - expected_results.cbegin() + 1, *res_it, r);
        ++res_it;
    }
}

TEST_F(SREstTest, BoxCar01)
{
    const size_t box_length = 1;
    
    std::vector<double> expected_results{
        0, 0, 1, 0, 1,
        0, 1, 0, 1, 0,
        0, 1, 0, 1, 0,
        0, 0, 1, 5, 0,
        0, 0, 5
    };

    auto res_it = expected_results.cbegin();
    for (auto const& u: uhv)
    {
        auto r = r_estimate_with_boxcar(u.crbegin(), u.crend(), box_length);
        EXPECT_DOUBLE_EQ(*res_it, r);
        auto r2 = r_estimate_with_boxcar_w_cont(u, box_length);
        EXPECT_DOUBLE_EQ(r, r2);
        ++res_it;
    }
}

TEST_F(SREstTest, BoxCar02)
{
    const size_t box_length = 2;
    
    std::vector<double> expected_results{
        0,      0,      1,      0,     .5,
        0.5,    1,      0,     .5,     .5,
        0,      1,      0,     .5,     .5,
        0,      0,      1,    2.5,    2.5,
        0,      0,    3.0
    };

    auto res_it = expected_results.cbegin();
    for (auto const& u: uhv)
    {
        auto r = r_estimate_with_boxcar(u.crbegin(), u.crend(), box_length);
        EXPECT_DOUBLE_EQ(*res_it, r);
        auto r2 = r_estimate_with_boxcar_w_cont(u, box_length);
        EXPECT_DOUBLE_EQ(r, r2);
        ++res_it;
    }
}

TEST_F(SREstTest, BoxCar03)
{
    const size_t box_length = 3;
    
    std::vector<double> expected_results{
        0,      0,      1,      0,     .5,
        0.5,    1,      0,1.0/3.0,1.0/3.0,
  1.0/3.0,      1,      0,1.0/3.0,1.0/3.0,
  1.0/3.0,      0,      1,5.0/3.0,5.0/3.0,
  5.0/3.0,      0,11.0/3.0
    };

    auto res_it = expected_results.cbegin();
    for (auto const& u: uhv)
    {
        auto r = r_estimate_with_boxcar(u.crbegin(), u.crend(), box_length);
        EXPECT_DOUBLE_EQ(*res_it, r);
        auto r2 = r_estimate_with_boxcar_w_cont(u, box_length);
        EXPECT_DOUBLE_EQ(r, r2);
        ++res_it;
    }
}

    // {0},          {0},          {1},          {0, 0},       {0, 1},
    // {1, 0},       {1, 1},       {0, 0, 0},    {0, 0, 1},    {0, 1, 0},
    // {1, 0, 0},    {1, 1, 1},    {0, 0, 0, 0}, {0, 0, 0, 1}, {0, 0, 1, 0},
    // {0, 1, 0, 0}, {1, 0, 0, 0}, {1, 1, 1, 1}, {0, 0, 0, 5}, {0, 0, 5, 0},
    // {0, 5, 0, 0}, {5, 0, 0, 0}, {1, 5, 1, 5}

TEST_F(SREstTest, BoxCar04)
{
    const size_t box_length = 4;
    
    std::vector<double> expected_results{
        0,      0,      1,      0,     .5,
        0.5,    1,      0,1.0/3.0,1.0/3.0,
  1.0/3.0,      1,      0,1.0/4.0,1.0/4.0,
  1.0/4.0,  1.0/4,      1,5.0/4.0,5.0/4.0,
  5.0/4.0,  5.0/4,12.0/4.0
    };

    auto res_it = expected_results.cbegin();
    for (auto const& u: uhv)
    {
        auto r = r_estimate_with_boxcar(u.crbegin(), u.crend(), box_length);
        EXPECT_DOUBLE_EQ(*res_it, r);
        auto r2 = r_estimate_with_boxcar_w_cont(u, box_length);
        EXPECT_DOUBLE_EQ(r, r2);
        ++res_it;
    }
}


} // namespace sre
} // namespace cc