#include "gtest/gtest.h"
#include "sre/SpikeRateEstimator.h"

#include <functional>
#include <numeric>

namespace cc
{
namespace sre
{
    class SpikeRateEstimatorBasic : public ::testing::Test
    {
    protected:
        virtual void SetUp() override
        {
            spdlog::set_level(spdlog::level::debug);
            spdlog::set_pattern("[%Y-%m-%d %H:%M:%S.%f] [%n] [%^%l%$] [PID:%P] [TID:%t] %v");
        }
    };

TEST_F(SpikeRateEstimatorBasic, Create)
{
    {
        SpikeRateEstimator sre00(10, 6);
    }
    {
        SpikeRateEstimator sre01;
    }
}

TEST_F(SpikeRateEstimatorBasic, NormFactor)
{
    {
        SpikeRateEstimator sre(1, 1, 2, 10ms);
        EXPECT_DOUBLE_EQ(sre.m_rate_norm_factor, 100.0);
    }
    {
        SpikeRateEstimator sre(1, 1, 2, 100ms);
        EXPECT_DOUBLE_EQ(sre.m_rate_norm_factor, 10.0);
    }
    {
        SpikeRateEstimator sre(1, 1, 2, 1s);
        EXPECT_DOUBLE_EQ(sre.m_rate_norm_factor, 1.0);
    }

    {
        SpikeRateEstimator sre(1, 1, 2, 2s);
        EXPECT_DOUBLE_EQ(sre.m_rate_norm_factor, 0.5);
    }

}

TEST_F(SpikeRateEstimatorBasic, ChannelUnitArithmetic)
{
    SpikeRateEstimator sre(10, 6);
    EXPECT_EQ(sre.n_ch_u(), 60);
    SpikeRateEstimator::size_type i;
    EXPECT_TRUE(sre.ch_u_offset(1, 0, i));
    EXPECT_EQ(i, 0);
    EXPECT_TRUE(sre.ch_u_offset(1, 1, i));
    EXPECT_EQ(i, 1);
    EXPECT_TRUE(sre.ch_u_offset(1, 2, i));
    EXPECT_EQ(i, 2);
    EXPECT_TRUE(sre.ch_u_offset(1, 3, i));
    EXPECT_EQ(i, 3);
    EXPECT_TRUE(sre.ch_u_offset(1, 4, i));
    EXPECT_EQ(i, 4);
    EXPECT_TRUE(sre.ch_u_offset(1, 5, i));
    EXPECT_EQ(i, 5);
    EXPECT_FALSE(sre.ch_u_offset(1, 6, i));
    EXPECT_FALSE(sre.ch_u_offset(1, 7, i));
    EXPECT_FALSE(sre.ch_u_offset(1, 8, i));
    EXPECT_TRUE(sre.ch_u_offset(2, 0, i));
    EXPECT_EQ(i, 6);
    EXPECT_TRUE(sre.ch_u_offset(2, 1, i));
    EXPECT_EQ(i, 7);
    EXPECT_TRUE(sre.ch_u_offset(2, 5, i));
    EXPECT_EQ(i, 11);
    EXPECT_FALSE(sre.ch_u_offset(2, 6, i));
    EXPECT_FALSE(sre.ch_u_offset(2, 7, i));
    EXPECT_TRUE(sre.ch_u_offset(10, 0, i));
    EXPECT_EQ(i, 54);
    EXPECT_TRUE(sre.ch_u_offset(10, 5, i));
    EXPECT_EQ(i, 59);
    EXPECT_FALSE(sre.ch_u_offset(10, 6, i));
    EXPECT_FALSE(sre.ch_u_offset(11, 0, i));
    EXPECT_FALSE(sre.ch_u_offset(12, 7, i));
}

// check size of unit_hists in member container
TEST_F(SpikeRateEstimatorBasic, UnitHistSize)
{
    SpikeRateEstimator sre(10, 6);
    EXPECT_EQ(sre.mp_unit_hist_container->size(), 60);
    for (auto const& uh: *(sre.mp_unit_hist_container))
    {
        EXPECT_EQ(uh.size(), sre.m_history_length);
    }

    SpikeRateEstimator sre00(2, 3, 10);
    EXPECT_EQ(sre00.mp_unit_hist_container->size(), 6);
    for (auto const& uh: *sre00.mp_unit_hist_container)
    {
        EXPECT_EQ(uh.size(), 10);
    }
}

TEST_F(SpikeRateEstimatorBasic, ValidateSpikes)
{
    SpikeRateEstimator sre(5, 7);
    {
        spike spk_00(0, 1, 0);
        EXPECT_TRUE(sre.validate_spk(spk_00));
    }
    {
        spike spk_00(0, 1, 6);
        EXPECT_TRUE(sre.validate_spk(spk_00));
    }
    {
        spike spk_00(0, 1, 7);
        EXPECT_FALSE(sre.validate_spk(spk_00));
    }
    {
        spike spk_00(0, 5, 0);
        EXPECT_TRUE(sre.validate_spk(spk_00));
    }
    {
        spike spk_00(0, 6, 0);
        EXPECT_FALSE(sre.validate_spk(spk_00));
    }
    {
        spike spk_00(0, 6, 7);
        EXPECT_FALSE(sre.validate_spk(spk_00));
    }
}

TEST_F(SpikeRateEstimatorBasic, AddSpikesToQueue)
{
    SpikeRateEstimator sre(5, 7);
    EXPECT_EQ(sre.mp_sq->sizeGuess(), 0);
    {
        spike spk_00(0, 1, 0);
        EXPECT_TRUE(sre.addSpike(spk_00));
        EXPECT_EQ(sre.mp_sq->sizeGuess(), 1);
    }
    {
        spike spk_00(0, 1, 6);
        EXPECT_TRUE(sre.addSpike(spk_00));
        EXPECT_EQ(sre.mp_sq->sizeGuess(), 2);
    }
    {
        spike spk_00(0, 1, 7);
        EXPECT_FALSE(sre.addSpike(spk_00));
        EXPECT_EQ(sre.mp_sq->sizeGuess(), 2);
    }
    {
        spike spk_00(0, 5, 0);
        EXPECT_TRUE(sre.addSpike(spk_00));
        EXPECT_EQ(sre.mp_sq->sizeGuess(), 3);
    }
    {
        spike spk_00(0, 6, 0);
        EXPECT_FALSE(sre.addSpike(spk_00));
        EXPECT_EQ(sre.mp_sq->sizeGuess(), 3);
    }
    {
        spike spk_00(0, 6, 7);
        EXPECT_FALSE(sre.addSpike(spk_00));
        EXPECT_EQ(sre.mp_sq->sizeGuess(), 3);
    }
    {
        EXPECT_TRUE(sre.addSpike(1, 2));
        EXPECT_EQ(sre.mp_sq->sizeGuess(), 4);
    }
    {
        EXPECT_FALSE(sre.addSpike(10, 2));
        EXPECT_EQ(sre.mp_sq->sizeGuess(), 4);
    }
}

// checks that a list of spikes are correctly sorted into unit_hists
TEST_F(SpikeRateEstimatorBasic, ConsumeQueue)
{
    SpikeRateEstimator sre(2, 3, 10);
    
    auto sr_rb = std::make_shared<SpikeRateEstimator::rate_ring_buffer_type>();
    SpikeRateEstimator::rate_ring_buffer_type::VecT rateV;
    rateV.reserve(sr_rb->size());
    
    sre.bindCallback([](const SpikeRateEstimator::unit_hist_type& uh){ return r_estimate_with_boxcar_w_cont(uh, 4);});
    sre.attachSpikeRateRB(sr_rb);
    
    EXPECT_EQ(sre.mp_sq->sizeGuess(), 0);
    std::vector<std::vector<spike>> spk_vv {
        {
            {0, 2, 0}, {1, 2, 1}, {1, 1, 0}, {3, 2, 0}, {3, 5, 0}
        },
        {
            {4, 1, 0}, {5, 1, 1}, {5, 2, 2}, {7, 1, 1}, {8, 1, 2}, {10, 2, 0}
        },
        {
            {11, 1, 0}, {12, 1, 2}, {13, 2, 0}
        },
        {
            {14, 1, 1}, {16, 2, 1}, {13, 2, 2}, {13, 1, 2}
        },
        {
            {14, 1, 2}
        },
        {
            {14, 2, 0}, {16, 2, 1}, {13, 2, 1}, {13, 1, 1}
        },
        {
        },
        {
            {4, 1, 0}, {5, 1, 1}, {5, 2, 2}, {7, 1, 2}, {8, 1, 2}, {10, 2, 0}
        }
    };
    const std::vector<std::vector<SpikeRateEstimator::size_type>> spk_counts {
        {1, 0, 0, 2, 1, 0},
        {2, 2, 1, 3, 1, 1},
        {3, 2, 2, 4, 1, 1},
        {3, 3, 3, 4, 2, 2},
        {3, 3, 4, 4, 2, 2},
        {3, 4, 4, 5, 4, 2},
        {3, 4, 4, 5, 4, 2},
        {4, 5, 6, 6, 4, 3}
    };
    
    const std::vector< SpikeRateEstimator::size_type> counts{4, 6, 3, 4, 1, 4, 0, 6};
    
    auto cit = counts.cbegin();
    auto s_it = spk_counts.cbegin();
    
    for (auto const& spk_v: spk_vv)
    {
        for (auto const& spk: spk_v)
        {
            sre.addSpike(spk);
        }
        auto n = sre.consumeQueue();
        EXPECT_EQ(*cit, n);
        ++ cit;
        auto sc_it = (*s_it).cbegin();
        for (auto const& uh: *(sre.mp_unit_hist_container))
        {
            auto uhc = std::accumulate(uh.cbegin(), uh.cend(), 0LU);
            EXPECT_EQ(uhc, *sc_it);
            ++ sc_it;
        }
        ++ s_it;
        
        sre.applyCallback();
    }
    sr_rb->consume(rateV);

}

TEST_F(SpikeRateEstimatorBasic, ApplyCallbackWithoutConfig)
{
    {
        SpikeRateEstimator sre(2, 3, 10);
        sre.applyCallback();
    }
    {
        SpikeRateEstimator sre(2, 3, 10);
        auto sr_rb = std::make_shared<SpikeRateEstimator::rate_ring_buffer_type>();
        SpikeRateEstimator::rate_ring_buffer_type::VecT rateV;
        rateV.reserve(sr_rb->size());
    
        sre.attachSpikeRateRB(sr_rb);
        
        sre.applyCallback();
    }
    {
        SpikeRateEstimator sre(2, 3, 10);
        sre.bindCallback([](const SpikeRateEstimator::unit_hist_type& uh){ return r_estimate_with_boxcar_w_cont(uh, 4);});
        
        sre.applyCallback();
    }
    
    {
        SpikeRateEstimator sre(2, 3, 10);
        auto my_log = spdlog::stdout_color_mt<spdlog::async_factory>("SRE");
        my_log->set_pattern(">>>>>>>>> %H:%M:%S %z %n %L %v <<<<<<<<<");
        sre.applyCallback();
        sre.set_logger(my_log);
        sre.applyCallback();
        
    }

}

TEST_F(SpikeRateEstimatorBasic, RunProcessThread)
{
    SpikeRateEstimator sre(1, 1, 25, 10ms);
    spike spk{0,1,0};
    auto sr_rb = std::make_shared<SpikeRateEstimator::rate_ring_buffer_type>();
    SpikeRateEstimator::rate_ring_buffer_type::VecT rateV;
    rateV.reserve(sr_rb->size());
    sre.attachSpikeRateRB(sr_rb);
    sre.setSpikeRateNormalization(false);

    // sre.bindCallback([](const SpikeRateEstimator::unit_hist_type& uh){ return r_estimate_with_boxcar_w_cont(uh, 10);});
    sre.bindCallback([](const SpikeRateEstimator::unit_hist_type& uh){ return r_estimate_with_exp_decay_w_cont(uh, .891, 10);});
    
    sre.signalStart();
    
    auto start = SpikeRateEstimator::clock::now() + 55ms;
    std::this_thread::sleep_until(start);
    start = SpikeRateEstimator::clock::now();
    auto next_time{SpikeRateEstimator::clock::now()};
    
    double spk_count = 0;
    for (int i = 0; i < 50; ++i)
    {
        sre.addSpike(spk);
        ++ spk_count;
        next_time += 10ms;
        std::this_thread::sleep_until(next_time);
    }
    auto dur = SpikeRateEstimator::clock::now() - start;

    start = SpikeRateEstimator::clock::now() + 150ms;
        std::this_thread::sleep_until(start);


    sre.signalStop();
    
    sr_rb->consume(rateV);
    
    auto rate_sum = std::accumulate(rateV.begin(), rateV.end(), 0.0, [](double a, const SpikeRateEstimator::rate_ring_buffer_type::BufType& v){return a + v.second[0];});
    
    EXPECT_DOUBLE_EQ(rate_sum, spk_count) << "Adding not-normalized rates should equal spike count";
}

TEST_F(SpikeRateEstimatorBasic, ValidateSpikeRate)
{
    {
        SpikeRateEstimator sre(1, 1, 25, 10ms);
        spike spk{0,1,0};
        auto sr_rb = std::make_shared<SpikeRateEstimator::rate_ring_buffer_type>();
        SpikeRateEstimator::rate_ring_buffer_type::VecT rateV;
        rateV.reserve(sr_rb->size());
        sre.attachSpikeRateRB(sr_rb);
        sre.setSpikeRateNormalization(false);

        sre.bindCallback([](const SpikeRateEstimator::unit_hist_type& uh){ return r_estimate_with_boxcar_w_cont(uh, 5);});
    
        sre.addSpike(spk);
        sre.processSpikes();
        sre.addSpike(spk);
        sre.addSpike(spk);
        sre.processSpikes();
        sre.processSpikes();
    
        sr_rb->consume(rateV);
    
        SpikeRateEstimator::rate_ring_buffer_type::BufType y{0, {.2, .6, .6}};
    
        EXPECT_EQ(rateV.size(), y.second.size()) << "Vectors v and y are of unequal length";

        for (size_t i = 0; i < rateV.size(); ++i)
        {
            EXPECT_DOUBLE_EQ(rateV[i].second[0], y.second[i]) << "Vectors v and y differ at index " << i;
        }
    }
    {
        SpikeRateEstimator sre(1, 1, 25, 10ms);
        spike spk{0,1,0};
        auto sr_rb = std::make_shared<SpikeRateEstimator::rate_ring_buffer_type>();
        SpikeRateEstimator::rate_ring_buffer_type::VecT rateV;
        rateV.reserve(sr_rb->size());
        sre.attachSpikeRateRB(sr_rb);
        sre.setSpikeRateNormalization(true);

        sre.bindCallback([](const SpikeRateEstimator::unit_hist_type& uh){ return r_estimate_with_boxcar_w_cont(uh, 5);});
    
        sre.addSpike(spk);
        sre.processSpikes();
        sre.addSpike(spk);
        sre.addSpike(spk);
        sre.processSpikes();
        sre.processSpikes();
    
        sr_rb->consume(rateV);
    
        SpikeRateEstimator::rate_ring_buffer_type::BufType y{0, {20, 60, 60}};
    
        EXPECT_EQ(rateV.size(), y.second.size()) << "Vectors v and y are of unequal length";

        for (size_t i = 0; i < rateV.size(); ++i)
        {
            EXPECT_DOUBLE_EQ(rateV[i].second[0], y.second[i]) << "Vectors v and y differ at index " << i;
        }
    }
}


TEST_F(SpikeRateEstimatorBasic, SetChannelUnits)
{
    SpikeRateEstimator sre{10, 3, 25, 10ms};
    SpikeRateEstimator::ch_un_list_type chunlist{{1, 1}, {2, 1}, {3, 0}, {4, 2}};
    sre.setChannelUnitList(chunlist);
    EXPECT_EQ(sre.mp_unit_hist_container->size(), 4);
    
    EXPECT_TRUE(sre.addSpike(1, 1));
    EXPECT_TRUE(sre.addSpike(2, 1));
    EXPECT_TRUE(sre.addSpike(3, 0));
    EXPECT_TRUE(sre.addSpike(4, 2));
    EXPECT_FALSE(sre.addSpike(1, 0));
    EXPECT_FALSE(sre.addSpike(2, 0));
    EXPECT_FALSE(sre.addSpike(3, 1));
    EXPECT_FALSE(sre.addSpike(3, 2));
    EXPECT_FALSE(sre.addSpike(9, 2));
    EXPECT_FALSE(sre.addSpike(10, 0));
    
}

TEST_F(SpikeRateEstimatorBasic, GetChannelUnits)
{
    SpikeRateEstimator::ch_un_list_type chunlist{{1, 1}, {2, 1}, {3, 0}, {4, 2}};
    SpikeRateEstimator sre{chunlist, 25, 10ms};
    
    EXPECT_EQ(sre.mp_unit_hist_container->size(), 4);
    
    auto ch_u_map_pair = sre.getChannelUnitMapping();
    
    auto ch_u_list = ch_u_map_pair.first;
    auto ch_u_map = ch_u_map_pair.second;
    
    EXPECT_EQ(ch_u_list.size(), 4);
    EXPECT_EQ(ch_u_map.size(), 4);
    
    EXPECT_EQ(ch_u_list[0].first, 1);
    EXPECT_EQ(ch_u_list[0].second, 1);
    EXPECT_EQ(ch_u_list[1].first, 2);
    EXPECT_EQ(ch_u_list[1].second, 1);
    EXPECT_EQ(ch_u_list[2].first, 3);
    EXPECT_EQ(ch_u_list[2].second, 0);
    EXPECT_EQ(ch_u_list[3].first, 4);
    EXPECT_EQ(ch_u_list[3].second, 2);
    
    EXPECT_EQ((ch_u_map.at(std::make_pair(1,1))), 0);
    EXPECT_EQ((ch_u_map.at(std::make_pair(2,1))), 1);
    EXPECT_EQ((ch_u_map.at(std::make_pair(3,0))), 2);
    EXPECT_EQ((ch_u_map.at(std::make_pair(4,2))), 3);
    EXPECT_THROW((ch_u_map.at(std::make_pair(2,2))), std::exception);
    
}

TEST_F(SpikeRateEstimatorBasic, ResizeQueue)
{
    SpikeRateEstimator sre{1, 1, 25, 10ms};
    auto sr_rb = std::make_shared<SpikeRateEstimator::rate_ring_buffer_type>();
    SpikeRateEstimator::rate_ring_buffer_type::VecT rateV;
    rateV.reserve(sr_rb->size());
    sre.attachSpikeRateRB(sr_rb);
    sre.bindCallback([](const SpikeRateEstimator::unit_hist_type& uh){ return r_estimate_with_boxcar_w_cont(uh, 5);});
    
    
    EXPECT_EQ(sre.mp_sq->capacity(), 20);
    
    SpikeRateEstimator::ch_un_list_type chunlist{{0, 1}, {1, 1}};
    sre.setChannelUnitList(chunlist);
    EXPECT_EQ(sre.mp_sq->capacity(), 40);
    EXPECT_TRUE(sre.ringBufferInvalid());
    sre.processSpikes();
    EXPECT_FALSE(sre.ringBufferInvalid());
    
    
    sre.setSpikeRateLoopInterval(100ms);
    
    EXPECT_EQ(sre.mp_sq->capacity(), 400);
    EXPECT_TRUE(sre.ringBufferInvalid());
    sre.processSpikes();
    EXPECT_FALSE(sre.ringBufferInvalid());

    sre.setSpikeRateLoopInterval(1s);
    EXPECT_TRUE(sre.ringBufferInvalid());
    sre.processSpikes();
    EXPECT_FALSE(sre.ringBufferInvalid());
    
    EXPECT_EQ(sre.mp_sq->capacity(), 4000);
    
}


TEST_F(SpikeRateEstimatorBasic, TimestampCallbackNotSet)
{
    {
        SpikeRateEstimator sre(1, 1, 25, 10ms);
        spike spk{0,1,0};
        auto sr_rb = std::make_shared<SpikeRateEstimator::rate_ring_buffer_type>();
        SpikeRateEstimator::rate_ring_buffer_type::VecT rateV;
        rateV.reserve(sr_rb->size());
        sre.attachSpikeRateRB(sr_rb);
        sre.setSpikeRateNormalization(false);

        sre.bindCallback([](const SpikeRateEstimator::unit_hist_type& uh){ return r_estimate_with_boxcar_w_cont(uh, 2);});
    
        sre.addSpike(spk);
        sre.processSpikes();
        sre.addSpike(spk);
        sre.addSpike(spk);
        sre.processSpikes();
        sre.processSpikes();
    
        sr_rb->consume(rateV);
    
        std::vector<Timestamp_t> tss{0, 0, 0};
        std::vector<SpikeRateEstimator::rate_type> y{0.5, 1.5, 1};
        
    
        for (size_t i = 0; i < rateV.size(); ++i)
        {
            EXPECT_EQ(rateV[i].first, tss[i]);
            EXPECT_EQ(rateV[i].second[0], y[i]);
        }
    }
}
TEST_F(SpikeRateEstimatorBasic, TimestampCallback)
{
    {
        SpikeRateEstimator sre(1, 1, 25, 10ms);
        spike spk{0,1,0};
        auto sr_rb = std::make_shared<SpikeRateEstimator::rate_ring_buffer_type>();
        SpikeRateEstimator::rate_ring_buffer_type::VecT rateV;
        rateV.reserve(sr_rb->size());
        sre.attachSpikeRateRB(sr_rb);
        sre.setSpikeRateNormalization(false);

        sre.bindCallback([](const SpikeRateEstimator::unit_hist_type& uh){ return r_estimate_with_boxcar_w_cont(uh, 2);});
        sre.bindTimestampCallback([](){static Timestamp_t ts = 0; ++ts; return ts;});
    
        sre.addSpike(spk);
        sre.processSpikes();
        sre.addSpike(spk);
        sre.addSpike(spk);
        sre.processSpikes();
        sre.processSpikes();
    
        sr_rb->consume(rateV);
    
        std::vector<Timestamp_t> tss{1, 2, 3};
        std::vector<SpikeRateEstimator::rate_type> y{0.5, 1.5, 1};
    
        for (size_t i = 0; i < rateV.size(); ++i)
        {
            EXPECT_EQ(rateV[i].first, tss[i]);
            EXPECT_EQ(rateV[i].second[0], y[i]);
        }
    }
}


// The fixture for testing class Project1. From google test primer.
class SpikeRateEstimatorBase : public ::testing::Test
{
protected:
    // You can remove any or all of the following functions if its body
    // is empty.

    SpikeRateEstimatorBase()
    {
        
    }

    virtual ~SpikeRateEstimatorBase()
    {
        // You can do clean-up work that doesn't throw exceptions here.
    }

    // If the constructor and destructor are not enough for setting up
    // and cleaning up each test, you can define the following methods:
    virtual void SetUp() override
    {
        spdlog::set_level(spdlog::level::debug);
        spdlog::set_pattern("[%Y-%m-%d %H:%M:%S.%f] [%n] [%^%l%$] [PID:%P] [TID:%t] %v");
    }

    virtual void TearDown() override
    {
    }
    
    SpikeRateEstimator sre;

    std::vector<spike> spk_v{{2, 2, 0}, {12, 6, 0}, {14, 6, 0}, {18, 6, 0}, {19, 5, 0}, {35, 6, 0}, {36, 4, 0}, {43, 6, 0}, {45, 4, 0}, {51, 6, 0}, {54, 6, 0}, {62, 6, 0}, {64, 6, 0}, {71, 6, 0}, {75, 4, 0}, {76, 5, 0}, {78, 5, 0}, {81, 6, 0}, {83, 6, 0}, {88, 4, 0}, {93, 5, 0}, {98, 6, 0}, {104, 5, 0}, {104, 6, 0}, {113, 2, 0}, {119, 6, 0}, {120, 4, 0}, {125, 5, 0}, {127, 6, 0}, {128, 4, 0}, {134, 6, 0}, {136, 4, 0}, {142, 5, 0}, {143, 6, 0}, {144, 6, 0}, {145, 5, 0}, {146, 4, 0}, {148, 6, 0}, {150, 6, 0}, {151, 2, 0}, {151, 3, 0}, {157, 6, 0}, {158, 5, 0}, {159, 4, 0}, {159, 5, 0}, {160, 6, 0}, {163, 5, 0}, {166, 2, 0}, {167, 5, 0}, {174, 2, 0}, {176, 6, 0}, {181, 5, 0}, {188, 6, 0}, {189, 5, 0}, {190, 5, 0}, {195, 4, 0}, {202, 5, 0}, {204, 5, 0}, {205, 6, 0}, {206, 6, 0}, {209, 6, 0}, {221, 4, 0}, {222, 5, 0}, {224, 4, 0}, {225, 3, 0}, {229, 6, 0}, {236, 6, 0}, {239, 6, 0}, {240, 6, 0}, {247, 6, 0}, {257, 1, 0}, {272, 5, 0}, {274, 2, 0}, {279, 5, 0}, {279, 6, 0}, {280, 5, 0}, {285, 5, 0}, {285, 6, 0}, {294, 4, 0}, {297, 5, 0}, {297, 6, 0}, {306, 6, 0}, {313, 5, 0}, {315, 5, 0}, {326, 5, 0}, {329, 6, 0}, {332, 5, 0}, {332, 6, 0}, {339, 5, 0}, {340, 2, 0}, {344, 5, 0}, {352, 5, 0}, {353, 5, 0}, {362, 6, 0}, {366, 6, 0}, {367, 6, 0}, {379, 6, 0}, {382, 6, 0}, {383, 4, 0}, {385, 6, 0}, {386, 5, 0}, {389, 5, 0}, {393, 2, 0}, {393, 5, 0}, {405, 6, 0}, {413, 5, 0}, {417, 6, 0}, {420, 6, 0}, {424, 3, 0}, {426, 2, 0}, {427, 4, 0}, {431, 6, 0}, {435, 6, 0}, {436, 4, 0}, {438, 6, 0}, {440, 6, 0}, {445, 6, 0}, {448, 4, 0}, {451, 5, 0}, {454, 3, 0}, {454, 5, 0}, {458, 3, 0}, {461, 6, 0}, {462, 2, 0}, {464, 6, 0}, {467, 5, 0}, {470, 5, 0}, {481, 6, 0}, {484, 6, 0}, {490, 6, 0}, {496, 6, 0}, {499, 5, 0}, {501, 5, 0}, {510, 5, 0}, {515, 5, 0}, {520, 2, 0}, {524, 6, 0}, {531, 6, 0}, {543, 6, 0}, {547, 6, 0}, {552, 6, 0}, {557, 6, 0}, {560, 6, 0}, {565, 4, 0}, {579, 3, 0}, {582, 6, 0}, {592, 2, 0}, {595, 6, 0}, {600, 6, 0}, {606, 5, 0}, {608, 3, 0}, {609, 4, 0}, {609, 6, 0}, {612, 6, 0}, {614, 5, 0}, {621, 5, 0}, {621, 6, 0}, {630, 3, 0}, {642, 2, 0}, {643, 6, 0}, {648, 3, 0}, {648, 6, 0}, {649, 6, 0}, {652, 4, 0}, {654, 5, 0}, {655, 5, 0}, {659, 6, 0}, {663, 4, 0}, {670, 5, 0}, {671, 1, 0}, {677, 6, 0}, {679, 6, 0}, {689, 6, 0}, {691, 6, 0}, {696, 5, 0}, {697, 5, 0}, {699, 6, 0}, {701, 5, 0}, {702, 5, 0}, {711, 6, 0}, {712, 4, 0}, {714, 5, 0}, {716, 5, 0}, {720, 5, 0}, {721, 6, 0}, {722, 6, 0}, {726, 3, 0}, {732, 6, 0}, {733, 4, 0}, {735, 5, 0}, {735, 6, 0}, {737, 6, 0}, {743, 6, 0}, {744, 6, 0}, {746, 6, 0}, {747, 3, 0}, {747, 6, 0}, {748, 5, 0}, {751, 5, 0}, {761, 6, 0}, {762, 5, 0}, {764, 4, 0}, {765, 6, 0}, {774, 6, 0}, {776, 5, 0}, {778, 3, 0}, {779, 5, 0}, {780, 6, 0}, {785, 6, 0}, {799, 6, 0}, {810, 5, 0}, {815, 6, 0}, {819, 5, 0}, {821, 5, 0}, {822, 6, 0}, {823, 5, 0}, {834, 5, 0}, {839, 4, 0}, {840, 6, 0}, {846, 5, 0}, {847, 5, 0}, {847, 6, 0}, {852, 6, 0}, {858, 5, 0}, {868, 5, 0}, {869, 4, 0}, {874, 5, 0}, {877, 4, 0}, {878, 6, 0}, {885, 5, 0}, {889, 2, 0}, {892, 4, 0}, {905, 2, 0}, {915, 6, 0}, {918, 6, 0}, {921, 6, 0}, {926, 6, 0}, {927, 6, 0}, {930, 2, 0}, {932, 5, 0}, {936, 4, 0}, {936, 6, 0}, {946, 4, 0}, {951, 6, 0}, {958, 5, 0}, {962, 6, 0}, {975, 6, 0}, {978, 5, 0}, {982, 5, 0}, {983, 4, 0}, {988, 6, 0}, {989, 6, 0}, {992, 5, 0}, {993, 6, 0}, {996, 3, 0}, {998, 5, 0}, {998, 6, 0}, {999, 4, 0}, {1002, 5, 0}, {1021, 3, 0}, {1024, 3, 0}, {1029, 5, 0}, {1029, 6, 0}, {1041, 6, 0}, {1045, 3, 0}, {1045, 6, 0}, {1047, 6, 0}, {1048, 2, 0}, {1051, 4, 0}, {1060, 3, 0}, {1064, 4, 0}, {1065, 6, 0}, {1071, 4, 0}, {1072, 6, 0}, {1073, 3, 0}, {1080, 5, 0}, {1081, 5, 0}, {1082, 6, 0}, {1085, 2, 0}, {1090, 3, 0}, {1093, 6, 0}, {1102, 6, 0}, {1107, 5, 0}, {1124, 5, 0}, {1132, 5, 0}, {1134, 3, 0}, {1141, 5, 0}, {1144, 5, 0}, {1146, 4, 0}, {1147, 6, 0}, {1149, 3, 0}, {1151, 4, 0}, {1153, 6, 0}, {1157, 6, 0}, {1160, 5, 0}, {1168, 4, 0}, {1173, 6, 0}, {1180, 5, 0}, {1184, 5, 0}, {1184, 6, 0}, {1191, 6, 0}, {1193, 4, 0}, {1199, 6, 0}, {1204, 4, 0}, {1207, 6, 0}, {1210, 6, 0}, {1216, 6, 0}, {1219, 6, 0}, {1221, 4, 0}, {1225, 6, 0}, {1226, 3, 0}, {1227, 3, 0}, {1229, 5, 0}, {1231, 6, 0}, {1243, 6, 0}, {1247, 6, 0}, {1249, 5, 0}, {1249, 6, 0}, {1251, 3, 0}, {1252, 6, 0}, {1254, 2, 0}, {1258, 5, 0}, {1262, 1, 0}, {1268, 5, 0}, {1269, 5, 0}, {1274, 5, 0}, {1276, 3, 0}, {1276, 6, 0}, {1287, 4, 0}, {1288, 3, 0}, {1288, 5, 0}, {1295, 6, 0}, {1296, 6, 0}, {1301, 5, 0}, {1305, 5, 0}, {1308, 5, 0}, {1313, 6, 0}, {1319, 5, 0}, {1322, 4, 0}, {1323, 6, 0}, {1324, 5, 0}, {1326, 6, 0}, {1327, 3, 0}, {1327, 6, 0}, {1329, 6, 0}, {1330, 6, 0}, {1332, 6, 0}, {1336, 5, 0}, {1337, 6, 0}, {1339, 3, 0}, {1343, 6, 0}, {1344, 2, 0}, {1346, 5, 0}, {1346, 6, 0}, {1349, 4, 0}, {1353, 4, 0}, {1353, 5, 0}, {1354, 5, 0}, {1364, 6, 0}, {1366, 5, 0}, {1370, 4, 0}, {1371, 6, 0}, {1372, 6, 0}, {1373, 5, 0}, {1381, 5, 0}, {1392, 6, 0}, {1394, 5, 0}, {1397, 5, 0}, {1399, 4, 0}, {1399, 5, 0}, {1400, 6, 0}, {1401, 6, 0}, {1403, 4, 0}, {1405, 6, 0}, {1412, 6, 0}, {1414, 2, 0}, {1419, 5, 0}, {1424, 6, 0}, {1431, 2, 0}, {1431, 6, 0}, {1434, 6, 0}, {1436, 3, 0}, {1438, 3, 0}, {1439, 6, 0}, {1442, 2, 0}, {1445, 2, 0}, {1454, 5, 0}, {1455, 6, 0}, {1456, 5, 0}, {1459, 6, 0}, {1460, 4, 0}, {1460, 5, 0}, {1460, 6, 0}, {1462, 5, 0}, {1465, 2, 0}, {1469, 5, 0}, {1470, 6, 0}, {1472, 5, 0}, {1478, 5, 0}, {1480, 1, 0}, {1494, 6, 0}, {1496, 4, 0}, {1498, 4, 0}, {1498, 5, 0}, {1501, 6, 0}, {1503, 5, 0}, {1506, 5, 0}, {1511, 2, 0}, {1516, 5, 0}, {1517, 3, 0}, {1525, 6, 0}, {1526, 3, 0}, {1527, 3, 0}, {1528, 3, 0}, {1535, 4, 0}, {1535, 5, 0}, {1536, 4, 0}, {1536, 6, 0}, {1540, 5, 0}, {1541, 4, 0}, {1542, 5, 0}, {1543, 3, 0}, {1545, 6, 0}, {1546, 3, 0}, {1546, 6, 0}, {1553, 6, 0}, {1554, 6, 0}, {1555, 4, 0}, {1563, 6, 0}, {1567, 3, 0}, {1567, 5, 0}, {1571, 6, 0}, {1572, 4, 0}, {1575, 6, 0}, {1583, 6, 0}, {1584, 6, 0}, {1588, 6, 0}, {1595, 5, 0}, {1597, 1, 0}, {1598, 4, 0}, {1604, 4, 0}, {1609, 5, 0}, {1612, 6, 0}, {1613, 4, 0}, {1620, 5, 0}, {1626, 5, 0}, {1628, 5, 0}, {1637, 4, 0}, {1637, 6, 0}, {1643, 6, 0}, {1648, 6, 0}, {1649, 4, 0}, {1650, 5, 0}, {1651, 6, 0}, {1652, 5, 0}, {1653, 4, 0}, {1659, 5, 0}, {1659, 6, 0}, {1663, 6, 0}, {1670, 6, 0}, {1677, 5, 0}, {1677, 6, 0}, {1679, 5, 0}, {1684, 6, 0}, {1685, 3, 0}, {1691, 6, 0}, {1694, 6, 0}, {1696, 4, 0}, {1697, 6, 0}, {1699, 6, 0}, {1707, 6, 0}, {1709, 5, 0}, {1716, 6, 0}, {1720, 5, 0}, {1728, 6, 0}, {1729, 6, 0}, {1730, 5, 0}, {1736, 4, 0}, {1737, 6, 0}, {1741, 6, 0}, {1748, 6, 0}, {1754, 5, 0}, {1761, 5, 0}, {1769, 5, 0}, {1776, 3, 0}, {1779, 6, 0}, {1780, 6, 0}, {1781, 6, 0}, {1783, 2, 0}, {1785, 6, 0}, {1787, 5, 0}, {1787, 6, 0}, {1797, 6, 0}, {1798, 6, 0}, {1805, 6, 0}, {1807, 5, 0}, {1810, 6, 0}, {1812, 4, 0}, {1823, 6, 0}, {1824, 4, 0}, {1827, 4, 0}, {1829, 5, 0}, {1839, 6, 0}, {1841, 4, 0}, {1844, 4, 0}, {1851, 3, 0}, {1855, 5, 0}, {1862, 6, 0}, {1865, 5, 0}, {1869, 6, 0}, {1873, 6, 0}, {1881, 5, 0}, {1888, 6, 0}, {1898, 1, 0}, {1912, 6, 0}, {1914, 6, 0}, {1916, 4, 0}, {1923, 5, 0}, {1925, 5, 0}, {1926, 6, 0}, {1934, 4, 0}, {1937, 6, 0}, {1938, 6, 0}, {1940, 5, 0}, {1940, 6, 0}, {1941, 3, 0}, {1944, 5, 0}, {1947, 5, 0}, {1948, 4, 0}, {1950, 6, 0}, {1953, 6, 0}, {1962, 5, 0}, {1963, 5, 0}, {1977, 5, 0}, {1983, 4, 0}, {1986, 5, 0}, {1988, 6, 0}, {1989, 5, 0}, {1996, 4, 0}};
    
};


TEST_F(SpikeRateEstimatorBase, Create)
{
    sre.addSpike(spike(10, 5, 1));
    spike spk(12, 42, 2);
    sre.addSpike(spk);
    
    auto p_s1 = sre.mp_sq->frontPtr();
    EXPECT_EQ(p_s1->timestamp, 10);
    sre.mp_sq->popFront();
    auto p_s2 = sre.mp_sq->frontPtr();
    EXPECT_EQ(p_s2->timestamp, 12);
    sre.mp_sq->popFront();
    EXPECT_TRUE(sre.mp_sq->isEmpty());
}
/*
TEST_F(SpikeRateEstimatorBase, Callback)
{
    using namespace std::placeholders;
    // sre.rate_callback = std::bind(r_estimate_with_boxcar_w_cont<SpikeRateEstimator::unit_hist_type,double> , _1, 5);
    sre.rate_callback = std::bind(r_estimate_with_exp_decay_w_cont<SpikeRateEstimator::unit_hist_type,double> , _1, .96, 100);
    
    sre.addSpike(spike(10, 5, 1));
    sre.addSpike(spike(10, 5, 1));
    sre.addSpike(spike(10, 5, 1));
    EXPECT_EQ(sre.consumeQueue(), 3);
    sre.applyCallback();
    EXPECT_EQ(sre.consumeQueue(), 0);
    sre.applyCallback();
    
    sre.addSpike(spike(10, 5, 1));
    sre.addSpike(spike(10, 5, 1));
    EXPECT_EQ(sre.consumeQueue(), 2);
    sre.applyCallback();
    
    EXPECT_EQ(sre.consumeQueue(), 0);
    sre.applyCallback();
    
    EXPECT_EQ(sre.consumeQueue(), 0);
    sre.applyCallback();
    
    sre.addSpike(spike(10, 5, 1));
    sre.addSpike(spike(10, 5, 1));
    sre.addSpike(spike(10, 5, 1));
    sre.addSpike(spike(10, 5, 1));
    EXPECT_EQ(sre.consumeQueue(), 4);
    sre.applyCallback();
    
    EXPECT_EQ(sre.consumeQueue(), 0);
    sre.applyCallback();
    sre.addSpike(spike(10, 5, 1));
    sre.addSpike(spike(10, 5, 1));
    EXPECT_EQ(sre.consumeQueue(), 2);
    sre.applyCallback();
    EXPECT_EQ(sre.consumeQueue(), 0);
    sre.applyCallback();
    EXPECT_EQ(sre.consumeQueue(), 0);
    sre.applyCallback();
    sre.addSpike(spike(10, 5, 1));
    sre.addSpike(spike(10, 5, 1));
    sre.addSpike(spike(10, 5, 1));
    sre.addSpike(spike(10, 5, 1));
    sre.addSpike(spike(10, 5, 1));
    sre.addSpike(spike(10, 5, 1));
    sre.addSpike(spike(10, 5, 1));
    sre.addSpike(spike(10, 5, 1));
    EXPECT_EQ(sre.consumeQueue(), 8);
    sre.applyCallback();
    EXPECT_EQ(sre.consumeQueue(), 0);
    sre.applyCallback();
    EXPECT_EQ(sre.consumeQueue(), 0);
    sre.applyCallback();
    EXPECT_EQ(sre.consumeQueue(), 0);
    sre.applyCallback();
}
*/

} // namespace sre
} // namespace cc

