#include "gtest/gtest.h"
#include "sre/SpikeRateEstimator.h"

#include <chrono>
#include <functional>
#include <numeric>

namespace cc
{
namespace sre
{

TEST(SpikeRateEstimatorPerformance, FastEat)
{
    spdlog::set_level(spdlog::level::info);
    spdlog::set_pattern("[%Y-%m-%d %H:%M:%S.%f] [%n] [%^%l%$] [PID:%P] [TID:%t] %v");
    
    const size_t n_ch = 128, n_u = 6, n_win_len = 50, n_hist = 1000;
    const size_t n_rep = 1000;
    
    SpikeRateEstimator sre(n_ch, n_u, n_hist);
    
    auto sr_rb = std::make_shared<SpikeRateEstimator::rate_ring_buffer_type>();
    SpikeRateEstimator::rate_ring_buffer_type::VecT rateV;
    rateV.reserve(sr_rb->size());
    
    sre.bindCallback([n_win_len](const SpikeRateEstimator::unit_hist_type& uh){ return r_estimate_with_exp_decay_w_cont(uh, .85, n_win_len);});
    // sre.bindCallback([n_win_len](const SpikeRateEstimator::unit_hist_type& uh){ return r_estimate_with_boxcar_w_cont(uh, n_win_len);});
    sre.attachSpikeRateRB(sr_rb);
    sre.signalStart();
    std::vector<spike> spk_v;
    for (size_t i = 0; i < n_ch; ++i)
    {
        for (size_t j = 0; j < n_u ; ++j)
        {
            spk_v.emplace_back(0, i, j);
        }
    }
    
    auto start_sim = std::chrono::high_resolution_clock::now();
    
    
    auto start_add = std::chrono::high_resolution_clock::now();
    auto dur_add = std::chrono::duration_cast<std::chrono::nanoseconds>(start_add - start_add).count();

    auto start_cons = std::chrono::high_resolution_clock::now();
    auto dur_cons = std::chrono::duration_cast<std::chrono::nanoseconds>(start_cons - start_cons).count();
    
    
    auto next_loop_time = start_sim;
    for (size_t i = 0; i < n_rep; ++i)
    {
        if (i%2 == 0)
        {
            start_add = std::chrono::high_resolution_clock::now();
            for (auto const& spk: spk_v)
            {
                sre.addSpike(spk);
            }
            dur_add += std::chrono::duration_cast<std::chrono::nanoseconds>(std::chrono::high_resolution_clock::now() - start_add).count();
            
        }
        next_loop_time += 10ms;
        std::this_thread::sleep_until(next_loop_time);
        
        if ((i+1)%(sr_rb->size()) == 0)
        {
            start_cons = std::chrono::high_resolution_clock::now();
            sr_rb->consume(rateV);
            dur_cons += std::chrono::duration_cast<std::chrono::nanoseconds>(std::chrono::high_resolution_clock::now() - start_cons).count();
        }
    }
    // std::this_thread::sleep_for(1s);
    auto finish_sim = std::chrono::high_resolution_clock::now();
    
    auto dur_ms_sim = std::chrono::duration_cast<std::chrono::milliseconds>(finish_sim - start_sim).count();
    printf("Rates:\n");
    for (auto const& row: rateV)
    {
        for (auto const& v: row.second)
        {
            printf("%.3f ", v);
        }
        printf("\n");
    }
    printf("It took %lldms to run %zu loops, or %fms/loop.\n", dur_ms_sim, n_rep, (double)dur_ms_sim / n_rep);
    printf("it took %lldms to add spikes, and %lldms to consume the rates.\n", dur_add / 1'000'000, dur_cons/ 1'000'000);
}


}
}

