#include <algorithm>
#include <cstdio>
#include <string>
#include <vector>

#include "gtest/gtest.h"
#include "sre/UnitHist.h"

namespace cc
{
namespace sre
{

using UH = UnitHist<unsigned int>;
class UHTest : public ::testing::Test
{
protected:
    UHTest()
    : uh(5)
    {
        
    }
    
    virtual ~UHTest()
    {
    }
    
    virtual void SetUp() override
    {
        
    }

    virtual void TearDown() override
    {
    }
    
    
    UH uh;
};

TEST_F(UHTest, Create)
{
    // Should have size 5
    ASSERT_EQ(uh.size(), 5);
    
    UH::container_type ct{1, 2, 3, 4, 5};
    UH uh2(ct);
    EXPECT_TRUE(std::equal(uh2.cbegin(), uh2.cend(), ct.cbegin())) << "Forward iterating UH yields sequence as used to fill it";
    ASSERT_EQ(uh2.size(), ct.size());
    ASSERT_TRUE(uh2.has_spikes());
    
    UH uh3(UH::container_type({9, 7, 5, 3}));
    ASSERT_EQ(uh3.size(), 4);
    ASSERT_EQ(*(uh3.cbegin()), 9);
    ASSERT_EQ(*(uh3.crbegin()), 3);
    ASSERT_TRUE(uh3.has_spikes());
    
    UH uh4(ct);
    ASSERT_EQ(uh4.size(), 5);
    ASSERT_EQ(*(uh4.cbegin()), 1);
    ASSERT_EQ(*(uh4.crbegin()), 5);
    ASSERT_TRUE(uh4.has_spikes());
    
    UH uh5(uh4);
    EXPECT_TRUE(std::equal(uh5.cbegin(), uh5.cend(), uh4.cbegin())) << "Forward iterating UH yields sequence as used to fill it";
    ASSERT_EQ(uh5.size(), uh4.size());
    ASSERT_TRUE(uh5.has_spikes());
    
    uh5.step();
    
    UH uh6(std::move(uh5));
    ASSERT_EQ(uh6.size(), 5);
    ASSERT_EQ(*(uh6.cbegin()), 2);
    ASSERT_EQ(*(uh6.crbegin()), 0);
    ASSERT_TRUE(uh6.has_spikes());
    
    UH::container_type ct2{0, 0, 0};
    UH uh7(ct2);
    EXPECT_TRUE(std::equal(uh7.cbegin(), uh7.cend(), ct2.cbegin())) << "Forward iterating UH yields sequence as used to fill it";
    ASSERT_EQ(uh7.size(), ct2.size());
    ASSERT_FALSE(uh7.has_spikes());
    
    UH uh8(UH::container_type({0, 0, 0, 0}));
    ASSERT_EQ(uh8.size(), 4);
    ASSERT_EQ(*(uh8.cbegin()), 0);
    ASSERT_EQ(*(uh8.crbegin()), 0);
    ASSERT_FALSE(uh8.has_spikes());
    
    UH uh9(UH::container_type({0, 0, 0, 1}));
    UH uh10(uh9);
    ASSERT_TRUE(uh9.has_spikes());
    ASSERT_TRUE(uh10.has_spikes());
    uh9.push(0);
    uh10.push(0);
    ASSERT_TRUE(uh9.has_spikes());
    ASSERT_TRUE(uh10.has_spikes());
    uh9.push(0);
    uh10.push(0);
    ASSERT_TRUE(uh9.has_spikes());
    ASSERT_TRUE(uh10.has_spikes());
    
    
}

TEST_F(UHTest, ZeroSize)
{
    UH uhe{0};
    ASSERT_EQ(uhe.size(), 0);
    {
        auto it1 = uhe.cbegin();
        auto it2 = uhe.cend();
        ASSERT_EQ(it1, it2);
        ++it1;
        ASSERT_EQ(it1, it2);
    }
    {
        auto it1 = uhe.crbegin();
        auto it2 = uhe.crend();
        ASSERT_EQ(it1, it2);
        ++it1;
        ASSERT_EQ(it1, it2);
    }
    uhe.push(42);
    ASSERT_EQ(uhe.size(), 0);
}

TEST_F(UHTest, AddItem)
{
    ASSERT_EQ(uh.size(), 5);
    uh.push(1);
    
    ASSERT_EQ(uh.size(), 5) << "adding element should not change size";
        
    ASSERT_EQ(*(-- uh.cend()), 1) << "Iterator to beginning should return last added element";
    ASSERT_EQ(*(uh.cbegin()), 0);
}

TEST_F(UHTest, IteratorTests)
{
    uh.push(1);
    uh.push(5);
    uh.push(3);
    
    auto it1 = uh.begin();
    auto it2 = uh.cbegin();
    auto it2_end = uh.cend();
    EXPECT_NE(it2, it2_end);
    // ASSERT_EQ(*it2, *it2_end);
    
    EXPECT_EQ(it1, it2) << "Iterators should compare the same for same base object and position";
    ++it1, ++it2;
    EXPECT_EQ(it1, it2) << "Iterators should compare the same for same base object and position";
    ++it2;
    EXPECT_NE(it1, it2) << "Iterators should compare different for same base object but different position";
    
    UH uh2(5);
    uh2.push(1);
    uh2.push(5);
    uh2.push(3);
    auto it3 = uh.begin();
    auto it4 = uh2.begin();
    EXPECT_NE(it3, it4) << "Iterators should compare different for different base objects with same contents and same position";
}

TEST_F(UHTest, UnitHistEquality)
{
    uh.push(1);
    uh.push(5);
    UH uh_long(6);
    uh_long.push(1);
    uh_long.push(5);
    ASSERT_NE(uh, uh_long) << "UnitHist of different sizes should not be equal";
    
    UH uh_short(4);
    uh_short.push(1);
    uh_short.push(5);
    ASSERT_NE(uh, uh_short) << "UnitHist of different sizes should not be equal";
    
    UH uh2(5);
    uh2.push(1);
    uh2.push(5);
    ASSERT_EQ(uh, uh2) << "UnitHist of same size and content should be equal";
    
    UH uh3(5);
    uh3.push(0);
    uh3.push(1);
    uh3.push(5);
    
    ASSERT_EQ(uh, uh3) << "UnitHist of same size and content should be equal";
}

TEST_F(UHTest, Copy)
{
    uh.push(1);
    uh.push(5);
    UH uh_copy(uh);
    ASSERT_EQ(uh, uh_copy) << "Copy of UnitHist should be equal to original";
    
    UH uh_move(std::move(uh_copy));
    ASSERT_EQ(uh_move.size(), 5) << "After move, size is that of original";
    ASSERT_EQ(uh, uh_move) << "After move, UnitHist equals that of original";
    
}

TEST_F(UHTest, IteratorCopyAssign)
{
    uh.push(42);
    UH::const_iterator it(uh.begin());
    UH::const_iterator it2(uh.end());
    ASSERT_NE(it, it2) << "Iterators different before copy assignment";
    ASSERT_NO_THROW(it = it2) << "Copy assignment should work for same base object";
    ASSERT_EQ(it, it2) << "Iterators equal after copy assignment";
    
    UH uh_copy(uh);
    
    UH::const_iterator it3(uh_copy.end());
    ASSERT_NE(it2, it3) << "Iterators different with different base objects";
    ASSERT_THROW(it = it3, std::invalid_argument) << "Copy assignment invalid for different base objects";
}

TEST_F(UHTest, FillEquality)
{
    std::vector<unsigned int> v1;
    for (long int i = 0; i < 10; ++i)
    {
        v1.push_back(i + 1);
        // make new unit hist and fill it
        UH uh2(5);
        
        for (auto it = v1.cbegin(); it != v1.cend(); ++it)
        {
            uh2.push(*it);
        }
        
        if (v1.size() <= uh2.size())
        {
            EXPECT_TRUE(std::equal(uh2.cend() - (i + 1), uh2.cend(), v1.cbegin())) << "Forward iterating UH yields sequence as used to fill it";
        }
        else
        {
            ASSERT_TRUE(std::equal(uh2.cbegin(), uh2.cend(), v1.cbegin() + (i + 1 - uh2.size()))) << "Forward iterating UH yields sequence as used to fill it";
        }
    }
}
 
TEST_F(UHTest, ReverseIteratorsBasics)
{
    {
        UH uh2(5);
        std::vector<unsigned int> v1{1};
        for (auto it = v1.cbegin(); it != v1.cend(); ++it)
        {
            uh2.push(*it);
        }
        
        EXPECT_EQ(*uh2.crbegin(), 1);
        EXPECT_EQ(*(uh2.crend() - 1), 0);
    }

    {
        UH uh2(5);
        std::vector<unsigned int> v1{1, 2, 3};
        for (auto it = v1.cbegin(); it != v1.cend(); ++it)
        {
            uh2.push(*it);
        }
        
        EXPECT_EQ(*uh2.crbegin(), 3);
        EXPECT_EQ(*(uh2.crend() - 1), 0);
    }
    
    {
        UH uh2(5);
        
        std::vector<unsigned int> v1{1, 2, 3, 4};
        for (auto it = v1.cbegin(); it != v1.cend(); ++it)
        {
            uh2.push(*it);
        }
        
        EXPECT_EQ(*uh2.crbegin(), 4);
        EXPECT_EQ(*(uh2.crend() - 1), 0);
    }
    
    {
        UH uh2(5);
        
        std::vector<unsigned int> v1{1, 2, 3, 4, 5};
        for (auto it = v1.cbegin(); it != v1.cend(); ++it)
        {
            uh2.push(*it);
        }
        
        EXPECT_EQ(*uh2.crbegin(), 5);
        EXPECT_EQ(*(uh2.crend() - 1), 1);
        
    }
    
    {
        UH uh2(5);
        
        std::vector<unsigned int> v1{1, 2, 3, 4, 5, 6};
        for (auto it = v1.cbegin(); it != v1.cend(); ++it)
        {
            uh2.push(*it);
        }
        
        EXPECT_EQ(*uh2.crbegin(), 6);
        EXPECT_EQ(*(uh2.crend() - 1), 2);
    }
    
    {
        UH uh2(5);
        
        std::vector<unsigned int> v1{1, 2, 3, 4, 5, 6, 7};
        for (auto it = v1.cbegin(); it != v1.cend(); ++it)
        {
            uh2.push(*it);
        }
        
        EXPECT_EQ(*uh2.crbegin(), 7);
        EXPECT_EQ(*(uh2.crend() - 1), 3);
    }
    
    {
        UH uh2(5);
        
        std::vector<unsigned int> v1{1, 2, 3, 4, 5, 6, 7, 8};
        for (auto it = v1.cbegin(); it != v1.cend(); ++it)
        {
            uh2.push(*it);
        }
        
        EXPECT_EQ(*uh2.crbegin(), 8);
        EXPECT_EQ(*(uh2.crend() - 1), 4);
    }
    
}

TEST_F(UHTest, ReverseIteratorsFill)
{
    std::vector<unsigned int> v1{1, 2, 3, 4, 5};
    
    for (auto it = v1.cbegin(); it != v1.cend(); ++it)
    {
        uh.push(*it);
    }
    
    ASSERT_TRUE(std::equal(uh.cbegin(), uh.cend(), v1.cend() - uh.size())) << "Forward iterating UH yields reverse sequence as used to fill it";
    ASSERT_TRUE(std::equal(uh.crbegin(), uh.crend(), v1.crbegin())) << "Reverse iterating UH yields same sequence as used to fill it";

}

TEST_F(UHTest, ReverseIteratorsFillMiddle)
{
    std::vector<unsigned int> v1{1, 2, 3, 4, 5, 6, 7, 8};
    
    for (auto it = v1.cbegin(); it != v1.cend(); ++it)
    {
        uh.push(*it);
    }
    
    ASSERT_TRUE(std::equal(uh.cbegin(), uh.cend(),  v1.cend() - uh.size())) << "Forward iterating UH yields reverse sequence as used to fill it";
    ASSERT_TRUE(std::equal(uh.crbegin(), uh.crend(), v1.crbegin())) << "Reverse iterating UH yields same sequence as used to fill it";
    
}

TEST_F(UHTest, ReverseIterators)
{

    UH uh3(5);
    std::vector<unsigned int> v3{1, 2, 3};
    
    for (auto it = v3.cbegin(); it != v3.cend(); ++it)
    {
        uh3.push(*it);
    }
    
    ASSERT_TRUE(std::equal(v3.cbegin(), v3.cend(), ++(++uh3.cbegin()) )) << "Reverse iterating UH yields same sequence as used to fill it; for fewer items than size";

    ASSERT_TRUE(std::equal(v3.crbegin(), v3.crend(), uh3.crbegin())) << "Forward iterating UH yields reverse sequence as used to fill it; for fewer items than size";
}

TEST_F(UHTest, IteratorArithmeticBase)
{
    {
        UH uh2(3);
        auto it = uh2.cbegin();
        it += 0;
    }
    
    {
        UH uh2(3);
        uh2.push(1);
        auto it = uh2.cbegin();
        ASSERT_EQ(*it, 0);
        it += 0;
        ASSERT_EQ(*it, 0);
    }

    {
        UH uh2(3);
        uh2.push(1);
        auto it = uh2.cbegin();
        ASSERT_EQ(*it, 0);
        it += 1;
        ASSERT_EQ(*it, 0);
    }

    {
        UH uh2(3);
        uh2.push(1);
        auto it = uh2.cbegin();
        ASSERT_EQ(*it, 0);
        it += 2;
        ASSERT_EQ(*it, 1);
    }
    {
        UH uh2(3);
        uh2.push(1);
        auto it = uh2.cbegin();
        ASSERT_EQ(*it, 0);
        it += 3;
        ASSERT_EQ(*it, 0);
    }
    {
        UH uh2(3);
        uh2.push(1);
        auto it = uh2.cbegin();
        ASSERT_EQ(*it, 0);
        it += 4;
        ASSERT_EQ(*it, 0);
    }
    {
        UH uh2(3);
        uh2.push(1);
        auto it = uh2.cbegin();
        ASSERT_EQ(*it, 0);
        it += 5;
        ASSERT_EQ(*it, 1);
    }
    {
        UH uh2(3);
        uh2.push(1);
        auto it = uh2.cbegin();
        ASSERT_EQ(*it, 0);
        it += 6;
        ASSERT_EQ(*it, 0);
    }
    {
        UH uh2(3);
        uh2.push(1);
        uh2.push(2);
        auto it = uh2.cbegin();
        ASSERT_EQ(*it, 0);
        it += 0;
        ASSERT_EQ(*it, 0);
    }
    {
        UH uh2(3);
        uh2.push(1);
        uh2.push(2);
        auto it = uh2.cbegin();
        ASSERT_EQ(*it, 0);
        it += 1;
        ASSERT_EQ(*it, 1);
    }

    {
        UH uh2(3);
        uh2.push(1);
        uh2.push(2);
        auto it = uh2.cbegin();
        ASSERT_EQ(*it, 0);
        it += 2;
        ASSERT_EQ(*it, 2);
    }

    {
        UH uh2(3);
        uh2.push(1);
        uh2.push(2);
        auto it = uh2.cbegin();
        ASSERT_EQ(*it, 0);
        it += 3;
        ASSERT_EQ(*it, 0);
    }

    {
        UH uh2(3);
        uh2.push(1);
        uh2.push(2);
        auto it = uh2.cbegin();
        ASSERT_EQ(*it, 0);
        it += 4;
        ASSERT_EQ(*it, 1);
    }

    {
        UH uh2(3);
        uh2.push(1);
        uh2.push(2);
        auto it = uh2.cbegin();
        ASSERT_EQ(*it, 0);
        it += 5;
        ASSERT_EQ(*it, 2);
    }
    {
        UH uh2(3);
        uh2.push(1);
        uh2.push(2);
        auto it = uh2.cbegin();
        ASSERT_EQ(*it, 0);
        it += 6;
        ASSERT_EQ(*it, 0);
    }
    

    {
        UH uh2(3);
        uh2.push(1);
        uh2.push(2);
        uh2.push(3);
        auto it = uh2.cbegin();
        ASSERT_EQ(*it, 1);
        it += 0;
        ASSERT_EQ(*it, 1);
    }
    {
        UH uh2(3);
        uh2.push(1);
        uh2.push(2);
        uh2.push(3);
        auto it = uh2.cbegin();
        ASSERT_EQ(*it, 1);
        it += 1;
        ASSERT_EQ(*it, 2);
    }
    {
        UH uh2(3);
        uh2.push(1);
        uh2.push(2);
        uh2.push(3);
        auto it = uh2.cbegin();
        ASSERT_EQ(*it, 1);
        it += 2;
        ASSERT_EQ(*it, 3);
    }
    {
        UH uh2(3);
        uh2.push(1);
        uh2.push(2);
        uh2.push(3);
        auto it = uh2.cbegin();
        ASSERT_EQ(*it, 1);
        it += 3;
        ASSERT_EQ(*it, 1);
    }
    {
        UH uh2(3);
        uh2.push(1);
        uh2.push(2);
        uh2.push(3);
        auto it = uh2.cbegin();
        ASSERT_EQ(*it, 1);
        it += 4;
        ASSERT_EQ(*it, 2);
    }
    {
        UH uh2(3);
        uh2.push(1);
        uh2.push(2);
        uh2.push(3);
        auto it = uh2.cbegin();
        ASSERT_EQ(*it, 1);
        it += 5;
        ASSERT_EQ(*it, 3);
    }
    
    {
        UH uh2(3);
        uh2.push(1);
        uh2.push(2);
        uh2.push(3);
        uh2.push(4);
        auto it = uh2.cbegin();
        ASSERT_EQ(*it, 2);
        it += 0;
        ASSERT_EQ(*it, 2);
    }
    {
        UH uh2(3);
        uh2.push(1);
        uh2.push(2);
        uh2.push(3);
        uh2.push(4);
        auto it = uh2.cbegin();
        ASSERT_EQ(*it, 2);
        it += 1;
        ASSERT_EQ(*it, 3);
    }
    {
        UH uh2(3);
        uh2.push(1);
        uh2.push(2);
        uh2.push(3);
        uh2.push(4);
        auto it = uh2.cbegin();
        ASSERT_EQ(*it, 2);
        it += 2;
        ASSERT_EQ(*it, 4);
    }
    {
        UH uh2(3);
        uh2.push(1);
        uh2.push(2);
        uh2.push(3);
        uh2.push(4);
        auto it = uh2.cbegin();
        ASSERT_EQ(*it, 2);
        it += 3;
        ASSERT_EQ(*it, 2);
    }
    {
        UH uh2(3);
        uh2.push(1);
        uh2.push(2);
        uh2.push(3);
        uh2.push(4);
        auto it = uh2.cbegin();
        ASSERT_EQ(*it, 2);
        it += 4;
        ASSERT_EQ(*it, 3);
    }
    {
        UH uh2(3);
        uh2.push(1);
        uh2.push(2);
        uh2.push(3);
        uh2.push(4);
        auto it = uh2.cbegin();
        ASSERT_EQ(*it, 2);
        it += 5;
        ASSERT_EQ(*it, 4);
    }
    
    {
        UH uh2(3);
        uh2.push(1);
        uh2.push(2);
        uh2.push(3);
        uh2.push(4);
        auto it = uh2.cbegin();
        ASSERT_EQ(*it, 2);
        it += 6;
        ASSERT_EQ(*it, 2);
    }
    
    {
        UH uh2(3);
        uh2.push(1);
        uh2.push(2);
        uh2.push(3);
        uh2.push(4);
        {
            auto it1 = uh2.cbegin();
            auto it2 = uh2.cbegin();
            ASSERT_EQ(*it1, 2);
            ASSERT_EQ(*it2, 2);
            it1 += 1;
            ++ it2;
            ASSERT_EQ(it1, it2);
            ASSERT_EQ(*it1, *it2);
        }
        
        {
            auto it1 = uh2.cbegin();
            auto it2 = uh2.cbegin();
            ASSERT_EQ(*it1, 2);
            ASSERT_EQ(*it2, 2);
            it1 += 2;
            ++ (++ it2);
            ASSERT_EQ(it1, it2);
            ASSERT_EQ(*it1, *it2);
        }
        
        {
            auto it1 = uh2.cbegin();
            auto it2 = uh2.cbegin();
            ASSERT_EQ(*it1, 2);
            ASSERT_EQ(*it2, 2);
            it1 += 3;
            ++ (++ (++ it2));
            ASSERT_EQ(it1, it2);
            ASSERT_EQ(*it1, *it2);
        }
        
        {
            auto it1 = uh2.cbegin();
            auto it2 = uh2.cbegin();
            ASSERT_EQ(*it1, 2);
            ASSERT_EQ(*it2, 2);
            it1 += 4;
            ++ (++ (++ (++ it2)));
            ASSERT_EQ(it1, it2);
            ASSERT_EQ(*it1, *it2);
        }
    }
    
}
    
TEST_F(UHTest, IteratorArithmetic)
{
    std::vector<unsigned int> v1{1, 2, 3, 4, 5};
    for (auto it = v1.cbegin(); it != v1.cend(); ++it)
    {
        uh.push(*it);
    }
    {
        // testing operator+(difference_type)
        auto it = uh.cbegin();
        ASSERT_EQ(*it, 1);
        auto it1 = it + 2;
        ASSERT_EQ(*it1, 3);
        auto it2 = it + 0;
        ASSERT_EQ(*it2, 1);
        auto it3 = it + (-2);
        ASSERT_EQ(*it3, 4);
    }
    {
        // testing operator+(difference_type, it)
        auto it = uh.cbegin();
        ASSERT_EQ(*it, 1);
        auto it1 = 1 + it;
        ASSERT_EQ(*it1, 2);
        auto it2 = 0 + it1;
        ASSERT_EQ(*it2, 2);
        auto it3 = -10 + it1;
        ASSERT_EQ(*it3, 2);
    }
    
    {
        // testing operator-=(difference_type)
        auto it = uh.cbegin();
        ASSERT_EQ(*it, 1);
        it -= 2;
        ASSERT_EQ(*it, 4);
        it -= 10;
        ASSERT_EQ(*it, 4);
        it -= 0;
        ASSERT_EQ(*it, 4);
        it -= 11;
        ASSERT_EQ(*it, 3);
        it -= -1;
        ASSERT_EQ(*it, 4);
    }
    {
        // testing operator-(difference_type)
        auto it = uh.cbegin();
        ASSERT_EQ(*it, 1);
        auto it1 = it - 2;
        ASSERT_EQ(*it1, 4);
        auto it2 = it - 0;
        ASSERT_EQ(*it2, 1);
        auto it3 = it - (-2);
        ASSERT_EQ(*it3, 3);
    }

    {
        // testing operator-(it1, it2)
        auto it = uh.cbegin();
        ASSERT_EQ(*it, 1);
        auto it1 = it + 2;
        ASSERT_EQ(it1 - it, 2);
        ASSERT_EQ(it - it1, -2);
        ASSERT_EQ(it - it, 0);
        
        auto it_e = uh.cend();
        ASSERT_EQ(it_e - it, 5);
        ASSERT_EQ(it - it_e, -5);
        ASSERT_EQ(it_e - it_e, 0);
        it += 5;
        
        ASSERT_EQ(it, it_e);
        
        UH uh1(7);
        auto it2 = uh1.cbegin();
        ASSERT_THROW(it2 - it, std::invalid_argument);
    }

}

TEST_F(UHTest, HasSpikes)
{
    EXPECT_EQ(uh.has_spikes(), false);
    uh.push(5);
    EXPECT_EQ(uh.has_spikes(), true);
    uh.push(0);
    uh.push(0);
    uh.push(0);
    uh.push(0);
    EXPECT_EQ(uh.has_spikes(), true);
    uh.push(0);
    EXPECT_EQ(uh.has_spikes(), false);
}

TEST_F(UHTest, AddToCurrent)
{
    // state of uh: 
    //      [0] 0 0 0 0 0
    // note: [] indicate posiiton of m_c_it, or next insertion; it is not valid data
    EXPECT_EQ(uh.has_spikes(), false);
    uh.step();                              // 0 [0] 0 0 0 0
    EXPECT_EQ(uh.has_spikes(), false);
    uh.add_to_current(1);                   // 0 [1] 0 0 0 0
    EXPECT_EQ(uh.has_spikes(), false);
    EXPECT_EQ(*uh.crbegin(), 0);
    EXPECT_EQ(*uh.cbegin(), 0);
    uh.step();                              // 0 1 [0] 0 0 0
    EXPECT_EQ(uh.has_spikes(), true);
    EXPECT_EQ(*uh.crbegin(), 1);
    EXPECT_EQ(*uh.cbegin(), 0);
    uh.add_to_current(1);                   // 0 1 [1] 0 0 0
    uh.push(6);                             // 0 1 6 [0] 0 0
    EXPECT_EQ(*uh.crbegin(), 6);
    EXPECT_EQ(*uh.cbegin(), 0);
    uh.push(6);                             // 0 1 6 6 [0] 0
    EXPECT_EQ(*uh.crbegin(), 6);
    EXPECT_EQ(*uh.cbegin(), 0);
    uh.push(7);                             // 0 1 6 6 7 [0]
    EXPECT_EQ(*uh.crbegin(), 7);
    EXPECT_EQ(*uh.cbegin(), 0);
    uh.step();                              // [0] 1 6 6 7 0
    EXPECT_EQ(uh.has_spikes(), true);
    uh.add_to_current(1);                   // [1] 1 6 6 7 0
    EXPECT_EQ(*uh.crbegin(), 0);
    EXPECT_EQ(*uh.cbegin(), 1);
    uh.push(0);                             // 0 [0] 6 6 7 0
    EXPECT_EQ(*uh.crbegin(), 0);
    EXPECT_EQ(*uh.cbegin(), 6);    
    uh.add_to_current(10);                  //  0 [10] 6 6 7 0
    EXPECT_EQ(*uh.crbegin(), 0);
    EXPECT_EQ(*uh.cbegin(), 6);
    uh.add_to_current(-10);                 // 0 [0] 6 6 7 0
    EXPECT_EQ(*uh.crbegin(), 0);
    EXPECT_EQ(*uh.cbegin(), 6);    
    uh.step();                              // 0 0 [0] 6 7 0
    EXPECT_EQ(*uh.crbegin(), 0);
    EXPECT_EQ(*uh.cbegin(), 6);
    uh.add_to_current(10);                  // 0 0 [10] 6 7 0
    EXPECT_EQ(*uh.crbegin(), 0);
    EXPECT_EQ(*uh.cbegin(), 6);
    uh.add_to_current(-20);                 // 0 0 [4294967286] 6 7 0
    EXPECT_EQ(*uh.crbegin(), 0);
    EXPECT_EQ(*uh.cbegin(), 6);
    uh.push(0);                             // 0 0 0 [0] 7 0
    EXPECT_TRUE(uh.has_spikes());
    uh.push(0);                             // 0 0 0 0 [0] 0
    EXPECT_FALSE(uh.has_spikes());
}

TEST_F(UHTest, Increment)
{
    ++uh;
    uh.step();
    ASSERT_EQ(*uh.crbegin(), 1);
    ++++uh;
    uh.step();
    ASSERT_EQ(*uh.crbegin(), 2);

}

TEST_F(UHTest, StreamOP)
{
    uh << 3;
    EXPECT_EQ(*uh.crbegin(), 3);
    EXPECT_EQ(*uh.cbegin(), 0);
    uh << 4 << 5 << 6 << 7 << 1 << 2;
    EXPECT_EQ(*uh.crbegin(), 2);
    EXPECT_EQ(*uh.cbegin(), 5);
}


} // namespace sre
} // namespace cc