'''
description: script to test data acquired via kiap_bci
author: Ioannis Vlachos
date:   22.11.18

Copyright (c) 2018  Ioannis Vlachos
All rights reserved.'''


import copy
import importlib
import pickle
import sys

import matplotlib.pyplot as plt
import numpy as np
import scipy.signal as signal
from scipy import io

import aux
import modules.classifier as clf
from aux import log
from helpers import data_management as dm

importlib.reload(aux)
importlib.reload(clf)
importlib.reload(dm)

get_features = int(sys.argv[1])
compute_model = int(sys.argv[2])

params = aux.load_config()
params.classifier.online = False        # here it is offline use

# features_fname = '/kiap/data/tom/train_trials.pkl'
features_fname = params.file_handling.datafile_path + '/train_trials.pkl'    # name of features array
model_fname1   = params.file_handling.datafile_path + '/model1.pkl'          # save sklearn LDA model
model_fname2   = params.file_handling.datafile_path + '/model2.pkl'          # save explicit LDA model

print('Select files for training:')
data_tot, tt, triggers = dm.get_raw(n_triggers=params.classifier.n_classes)


session_ids = list(range(0, data_tot.shape[0]))         # list with sessions
clf = clf.Classifier(params)
clf.triggers_shape = triggers.shape
clf.triggers = triggers
if get_features:
    # log.info('Get feature array from new spike-rate data')
    # clf.get_trials(data_tot[session_ids, :], triggers, features_fname, train_trials_comparison=[], save_features=1)
    try:
        clf.get_trials_kiap(data_tot[session_ids, :], triggers, features_fname=features_fname)                      # 1. GET FEATURES
    except Exception as e:
        log.error(e)
        sys.exit(1)
else:
    log.info('Old feature array will be used')

if compute_model:
    log.info('Computing classifier model...')
    try:
        _, _, _, _ = clf.train_LDA(clf.train_trials, baseline=0, features_fname=features_fname,                     # 2. TRAIN MODEL
                    model_fname1=model_fname1, model_fname2=model_fname2, solver='svd', save_model=1)
    except:
        raise
        sys.exit(1)
else:
    log.info('Stored classifier model will be used.')




print('Select files for testing:')
data_tot, tt, triggers = dm.get_raw(n_triggers=params.classifier.n_classes)
session_ids = list(range(0, data_tot.shape[0]))         # list with sessions
clf.triggers_shape = triggers.shape
clf.triggers = triggers

session_id = 0
for sess in range(session_id, session_id + 1):
    print(sess)
    cur_data = data_tot[session_id:session_id + 1]

    test_trials = clf.get_test_trials_kiap(cur_data)                                                                # 3. TEST MODEL
    
    if test_trials.shape[1] != clf.clf1.coef_.shape[1]:
        log.warning(f'# of features, testrials: {test_trials.shape[1]}, model coef: {clf.clf1.coef_.shape[1]}')
        log.critical('Mismatch in # of features between stored model and test_trials')
        sys.exit(1)

    clf.test_LDA_kiap(test_trials, session_id=sess, compare_results=0)


# log.info(f'Model 1 saved in {model_fname1}')
# log.info(f'Model 2 saved in {model_fname1}')

#  PLOT RESULTS
col = ['b', 'r', 'g']
# col = ['#1f77b4', '#ff7f0e', '#555555']
plt.figure(1)
plt.clf()

fct = params.daq.spike_rates.loop_interval / 1000.
tolerance = 1 # sec

xx1 = np.arange(cur_data[0, 0].shape[0]) * fct
xx2 = np.arange(test_trials.shape[0]) * fct

t_offset = params.recording.timing.t_response/2. * 0


ax1 = plt.subplot(211)
for cl_id in range(triggers.shape[1]):
    markerline, stemlines, baseline = plt.stem(triggers[0, cl_id][0] * fct + t_offset, triggers[0, cl_id][0] * 0 + 200, '-', basefmt=' ')
    plt.setp(stemlines, alpha=0.8, color=col[cl_id], lw=1)
    plt.setp(markerline, marker=None)

plt.gca().set_prop_cycle(None)
plt.plot(xx1, cur_data[0, 0][:, :2])

ax2 = plt.subplot(212, sharex = ax1)
for ii in range(triggers.shape[1]):
    idx = clf.rel_prob[:,ii] > 0
    plt.plot(xx2[idx], clf.rel_prob[idx, ii] * 25, '.', color=col[ii], lw=2)


# COMPUTE TRUE POSITIVES ETC.
# ----------------------------
tp = np.zeros((params.classifier.n_classes), dtype=int)
fp = np.zeros((params.classifier.n_classes), dtype=int)
fn = np.zeros((params.classifier.n_classes), dtype=int)
tn = np.zeros((params.classifier.n_classes), dtype=int)
tpr = np.zeros((params.classifier.n_classes))
fpr = np.zeros((params.classifier.n_classes))
fnr = np.zeros((params.classifier.n_classes))
tnr = np.zeros((params.classifier.n_classes))



for cl_id in range(params.classifier.n_classes):
    peaks, _ = signal.find_peaks(clf.rel_prob[:, cl_id], height=0.8, width=3, distance=40)
    if peaks.size==0:
        print(f'No peaks found for class {cl_id}')
        continue
    markerline, stemlines, baseline = plt.stem(xx2[peaks], xx2[peaks] * 0 + 200, '--')
    plt.setp(markerline, alpha=0.8, color=col[cl_id], lw=1)


    for idx in peaks:
        if np.any(np.abs(xx2[idx] - (triggers[0, cl_id] * fct + t_offset)) < tolerance):
            tp[cl_id] += 1
        else:
            fp[cl_id] += 1

    fn[cl_id] = triggers[0, cl_id].size - tp[cl_id]


    
    tpr[cl_id] = tp[cl_id] / triggers[0, cl_id].size
    fpr[cl_id] = fp[cl_id] / (fp[cl_id] + sum([triggers[0, ii].size for ii in range(params.classifier.n_classes)]) - triggers[0, cl_id].size)
    fnr[cl_id] = fn[cl_id] / (fn[cl_id] + tp[cl_id])
    tnr[cl_id] = 1 - fpr[cl_id]

# if params.classifier.n_classes == 2:
#     tpr[1] = tp[1] / (triggers[0, 1].size + triggers[0, 2].size)


plt.show()

# np.set_printoptions(precision=2)
np.set_printoptions(precision=2, formatter={'int': lambda x: '%2d' % x, 'float': lambda x: '%.2f' % x}, )
print('tp', tp, tpr)
print('fp', fp, fpr)
print('fn', fn, fnr)
print('tn', tn, tnr)
np.set_printoptions(precision=8)

np.testing.assert_array_almost_equal(tpr, 1 - fnr)
np.testing.assert_array_almost_equal(fpr, 1 - tnr)
print('Classification rates are valid')
