'''
description: implementation of lfp decoder; based on Milekovic et al. 2018
author: Ioannis Vlachos
date:   03.06.18

Copyright (c) 2018 Ioannis Vlachos.
All rights reserved.


steps:

- anti-alias filter
- down-sample to 15 khz?
- psd, exclude channels with high 60Hz noise
- car for remaining channels


- low-pass filter, 30th order windowed linear phase FIR

'''
import importlib
import pickle

import matplotlib.pyplot as plt
import munch
import numpy as np
from scipy import signal

import aux
import methods

importlib.reload(methods)



def read_config():
    '''read parameters from yaml file using munch package
    munch allows attribute style access'''

    with open(config_fname) as stream:
        params = munch.fromYAML(stream)
    return params


def buffer_to_array(data_trial):
    '''copy single trial continuous data to a numpy array'''

    arr = np.zeros((len(data_trial), data_trial[0][1].size), dtype=np.float32)
    for ii, val in enumerate(data_trial):
        arr[ii] = val[1] / 4.
    return arr


def car(arr):
    return np.mean(arr, axis=0)


def get_bad_channels(arr, freqs, freq_rng, thr):
    '''get ids of channels that have power in range freq_rng above thr'''

    fidx = (freqs > freq_rng[0]) & (freqs < freq_rng[1])
    res = arr[:, fidx]
    ch_ids = res > thr
    rows, col = np.where(ch_ids == True)
    return np.unique(rows)


def print_bad_channels(ch_ids):
    '''print on screen the list with bad channels'''

    print('trial_nr: {}'.format(trial_nr))
    if any(ch_ids):
        print(ch_ids)
        plot_orig_downsmpl()
        input('Press enter')


def plot_orig_downsmpl():
    '''plot original,  downsampled version and its spectrum'''

    xx = np.arange(arr0.shape[1])
    xx2 = np.arange(0, arr0.shape[1], params.sampling.smpl_fct)
    plt.figure(1)
    plt.clf()
    plt.subplot(211)
    plt.plot(xx, arr0[0], '-.', label='original')
    plt.plot(xx2, arr[0], '-.', label='downsampled')
    plt.legend()

    plt.subplot(212)
    plt.plot(freqs, psd[0])
    plt.xlim([0, 200])
    plt.ylim([0, 500])
    plt.pause(.1)



# def get_bands(arr, f, n_bins):
#     band_tot = []
#     for ff in np.arange(0, f[-1], n_bins):
#         # print(ff)
#         idx = (f >= ff) & (f < ff + n_bins)
#         # print(idx)
#         band_tot.append(np.sum(arr[:, idx, :], axis=1) / n_bins)
#     band_tot = np.array(band_tot)
#     band_tot = np.moveaxis(band_tot, 0, 1)
#     return band_tot


def get_bands(arr, f, n_bins, params):
    '''arr: stft: ch x fbins x tbins
    returns sum of different bands accross time'''

    # band_tot = np.zeros((26, 26, 96, 20))
    bb = int(np.ceil(f.max() / n_bins))
    band_tot = np.zeros((bb, bb, params.general.n_channels, arr.shape[2]))

    for ii, f1 in enumerate(np.arange(0, f.max(), n_bins)):
        for jj, f2 in enumerate(np.arange(0, f.max(), n_bins)):
            if f2 > f1:
                idx = (f >= f1) & (f < f2)
                try:
                    # print(f[ii], f[jj], f1, f2)
                    band_tot[ii, jj, :, :] = np.sum(arr[:, idx, :], axis=1) / (f2 - f1)
                    # print(f1,f2)
                except Exception as e:
                    print(e)
                    import pdb
                    pdb.set_trace()
    return band_tot


def compute_SNR(arr):
    # compute mu and sigma
    # averaged accross trials (blocks) in each session

    mu_cue = np.mean(arr[:2, :, :, :, :], axis=0)
    std_cue = np.std(arr[:2, :, :, :, :], axis=0)
    mu_bl = np.mean(arr[2:4, :, :, :, :], axis=0)
    std_bl = np.std(arr[2:4, :, :, :, :], axis=0)

    snr_el = np.abs(mu_cue - mu_bl) / (std_cue + std_bl)    # bstart x bend x ch  x tbins

    snr0 = []
    for ii in range(96):
        snr0.append(np.max(snr_el[:, :, ii, :], axis=2))
    snr = np.sum(np.array(snr0), axis=0)                     # bstart x bend

    # plt.figure(1)
    # plt.clf()
    # plt.pcolormesh(np.log10(snr), shading='gouraud')
    # plt.draw()
    # plt.show()

    return snr


def select_features(arr, freq_comp):
    '''arr: spectrogram_tot: dims: ch x fbins x tbins x trials
    returns snr and features

    features : n_features x 2; n_feat =  n_channels x n_bands (LFC, IFC, HFC)
    e.g. features = [channel_nr, freq-comp]; freq_comp = 0,1,2 (LFC, IFC, HFC)'''
    

    snr = np.zeros((96, 20, 3))                         # ch x tbins x n_bands
    features = np.zeros((50, 2), dtype=int)             # first 50 IFC and HFC features

    for ii, comp in enumerate(freq_comp):
        mu_act = np.mean(arr[:, comp, :, :2], axis=2)
        std_act = np.std(arr[:, comp, :, :2], axis=2)
        mu_bl = np.mean(arr[:, comp, :, 2:4], axis=2)
        std_bl = np.std(arr[:, comp, :, 2:4], axis=2)

        snr1 = np.abs(mu_act - mu_bl) / (std_act + std_bl)
        snr[:, :, ii] = snr1

    n_feat = params.general.n_channels * params.features.n_bands
    idx0 = np.argpartition(snr.flatten(), -n_feat)[-n_feat:]
    idx0 = np.vstack(np.unravel_index(idx0, snr.shape)).T     # [channel_nr, tbin, freq_comp]
    idx1 = idx0[:, [0, 2]].astype(int)
    _, ww = np.unique(idx1, axis=0, return_index=True)
    ww.sort()
    idx1 = idx1[ww]

    features[:5, :] = idx1[idx1[:, 1] == 0, :][:5]
    features[5:10, :] = idx1[idx1[:, 1] == 1, :][:5]
    features[10:15, :] = idx1[idx1[:, 1] == 2, :][:5]

    idx = idx1
    idx = np.delete(idx, np.where(idx[:, 1] == 0)[0][:5], axis=0)
    idx = np.delete(idx, np.where(idx[:, 1] == 1)[0][:5], axis=0)
    idx = np.delete(idx, np.where(idx[:, 1] == 2)[0][:5], axis=0)

    features[15:50] = idx[:35]
    return snr, features

# def select_features(snr):
#     '''arr: spectrogram_tot: dims: bstart x bend x ch x tbins

#     features : n_features x 2; n_feat =  n_channels x n_bands (LFC, IFC, HFC)
#     e.g. features = [channel_nr, freq-comp]; freq_comp = 0,1,2 (LFC, IFC, HFC)'''
#     features = np.zeros((50, 2), dtype=int)

#     # snr = np.arange(150)
#     # np.random.shuffle(snr)
#     # snr = snr.reshape(50, 3)


#     num = snr.size
#     idx0 = np.argpartition(snr.flatten(), range(num))[-num:]
#     idx1 = np.vstack(np.unravel_index(idx0, snr.shape)).T     # [channel_nr, tbin, freq_comp]
#     idx2 = idx1[:, [0, 1]].astype(int)
#     _, ww = np.unique(idx2, axis=0, return_index=True)
#     ww.sort()
#     # idx3 = idx2[ww]
#     idx3 = np.flipud(idx2[ww])

#     # check if elemens are monotonically decreasing
#     aa = np.array([snr[nn[0], nn[1]] for nn in idx3])
#     assert(np.all(aa[1:] <= aa[:-1]))                   
#     print(aa)

#     plt.plot(aa,'o')

#     features[:5, :] = idx3[idx3[:, 1] == 0, :][:5]
#     features[5:10, :] = idx3[idx3[:, 1] == 1, :][:5]
#     features[10:15, :] = idx3[idx3[:, 1] == 2, :][:5]

#     idx = idx3
#     idx = np.delete(idx, np.where(idx[:, 1] == 0)[0][:5], axis=0)
#     idx = np.delete(idx, np.where(idx[:, 1] == 1)[0][:5], axis=0)
#     idx = np.delete(idx, np.where(idx[:, 1] == 2)[0][:5], axis=0)

#     features[15:50] = idx[:35]


#     aa = np.array([snr[nn[0], nn[1]] for nn in features])
#     # assert(np.all(aa[1:] <= aa[:-1]))                   
#     print(aa)

#     plt.plot(aa, 'o')
#     plt.show()

#     return res



def test():


    import pdb
    pdb.set
    aa = np.random.randint(0,30,(10,3))

    n_feat = 30
    idx0 = np.argpartition(aa.flatten(), -n_feat)[-n_feat:]
    idx0 = np.vstack(np.unravel_index(idx0, snr.shape)).T     # [channel_nr, tbin, freq_comp]
    idx1 = idx0[:, [0, 2]].astype(int)
    _, ww = np.unique(idx1, axis=0, return_index=True)
    ww.sort()
    idx1 = idx1[ww]

    return

def plot_results():
    # PLOT RESULTS
    plt.figure(1)
    plt.clf()
    plt.plot(arr2[0])
    plt.plot(arr2_lp[0])
    plt.legend(['original', 'filtered'])
    plt.xlabel('sample points')
    plt.title('Single channel-trial')
    methods.filter_response(taps, params)                     # plot response of filter

    plt.figure(2)
    plt.clf()
    plt.pcolormesh(t3, f3, np.log10(np.abs(spectrogram[0])))         # plot stft
    plt.xlabel('time')
    plt.ylabel('Freqs (Hz)')
    # plt.ylim(0, params.lp.fc)

    plt.figure(3)
    plt.clf()
    ch = np.arange(norm_mat.shape[0])
    plt.pcolormesh(ch, f3, np.log10(np.abs(norm_mat.T)))
    plt.title('Normalization matrix')
    plt.xlabel('time')
    plt.ylabel('Freqs (Hz)')


if __name__ == '__main__':

    # MAIN PART
    config_fname = 'config.yaml'
    fname = '/kiap/data/2018-10-02/data.bin'

    ncols = 96

    data_total = np.fromfile(fname, dtype=np.float32)
    data_total = data_total.reshape(data_total.size // ncols, ncols)

    # data_total = pickle.load(open('buffer.pkl', 'rb'))
    params = aux.load_config()
    # plt.ioff()

    # band_tot_sess = np.empty((params.general.n_channels, 125, params.trials.n_trials))
    band_sess = []


    # ----------------------------
    # COMPUTE NORMALIZATION MATRIX
    # ----------------------------

    # for trial_nr in range(len(data_total['neural']['raw']))[:params.trials.n_norm]:
    for trial_nr in range(1):

        arr0 = data_total[-30000:, :].T
        # arr0 = buffer_to_array(data_trial)
        arr = signal.resample(arr0, arr0.shape[1] // params.sampling.smpl_fct, axis=1)          # downsample
        fs = params.sampling.fs // params.sampling.smpl_fct                                     # update fs
        freqs, psd = signal.welch(arr, fs=fs, nfft=10000, nperseg=10000)                        # compute psd
        ch_ids = get_bad_channels(psd, freqs, [90, 110], 80)                                    # get bad channels, e.g. strong powerline noise
        # print_bad_channels(ch_ids)

        av = np.mean(arr, axis=0)                                                               # car: remove common average
        arr2 = arr - av

        # SPECTRAL AMPLITUDE NORMALIZATION
        arr2_lp, taps = methods.low_pass_filter(arr2, params)
        f3, t3, arr3 = signal.stft(arr2_lp, fs=fs, window='hamming', nperseg=256, noverlap=221)  # arr3: n_chan x n_freq_bins x time_bins
        norm_mat = np.mean(np.abs(arr3), axis=2)                                                 # 96 x 129 (n_chan x fbins)


    n_trials = 4
    # spectrogram_tot = np.zeros((params.general.n_channels, 129, 20, params.trials.n_tot))
    spectrogram_tot = np.zeros((params.general.n_channels, 129, 20, n_trials))


    # --------------------------
    # DETERMINE FREQUENCY BANDS
    # --------------------------

    # for trial_ii, trial_nr in enumerate(range(len(data_total['neural']['raw']))[params.trials.n_norm:params.trials.n_norm + params.trials.n_tot]):
    for trial_ii, trial_nr in enumerate(range(n_trials)):
        # print(trial_ii, trial_nr)
        # data_trial = data_total['neural']['raw'][trial_nr][1][2]
        # arr0 = buffer_to_array(data_trial)
        # arr0 = data_total[-30000:, :].T
        arr0 = methods.get_trial_data(trial_ii, data_total)
        arr = signal.resample(arr0, arr0.shape[1] // params.sampling.smpl_fct, axis=1)          # downsample
        fs = params.sampling.fs // params.sampling.smpl_fct                                     # update fs
        freqs, psd = signal.welch(arr, fs=fs, nfft=10000, nperseg=10000)                        # compute psd
        # ch_ids = get_bad_channels(psd, freqs, [90, 110], 80)                                  # we may not need it here
        # print_bad_channels(ch_ids)

        av = np.mean(arr, axis=0)                                                               # car: remove common average
        arr2 = arr - av

        arr2_lp, taps = methods.low_pass_filter(arr2, params)

        # DETERMINE FREQUENCY BANDS (INTERMEDIATE (IFC) AND HIGH-FREQUENCY COMPONENTS (HFC))
        f3, t3, spectrogram = signal.stft(arr2_lp, fs=fs, window='hamming', nperseg=256, noverlap=203)

        # NORMALIZE
        norm_mat2 = norm_mat.reshape(norm_mat.shape + (1,)).repeat(spectrogram.shape[2], axis=2)    # repeat norm_mat per tbin
        methods.validate_norm_mat2(norm_mat, norm_mat2)                                             # check whether norm_mat2 is correct
        spectrogram = np.abs(spectrogram) / norm_mat2                                               # ch x fbins x tbins
        spectrogram_tot[:, :, :, trial_ii] = spectrogram                                            # ch x fbins x tbins x trials

        band_trial = get_bands(spectrogram, f3, params.bands.n_bins)                                # dims: bstart x bend x ch x  tbins
        # band_sess[:, :, trial_nr] = band_trial
        band_sess.append(band_trial)                                                                # dims: trials x bstart x bend x ch  x tbins

        # plot_results()

    band_sess = np.array(band_sess)                                                                 # dims: trials x bstart x bend x ch  x tbins
    # band_sess = np.moveaxis(band_sess, 0, 3)


    # snr = compute_SNR(band_sess)                                                                    # compute SNR per session
    # todo: compute SNR for all sessions


    # TODO: THIS IS CURRENTLY COMPUTED ONCE, BUT IT CAN BE COMPUTED ADAPTIVELY
    # SELECT FEATURES BASED ON IDENTIFIED FREQUENCY COMPONENTS
    snr, features = select_features(spectrogram_tot, [5, 8, 19])
    # snr, features = select_features(band_trial, [5, 8, 19])




    # CLASSIFIER
    # OPEN LOOP CALIBRATION
    class1 = []
    class2 = []
    # for trial_ii, trial_nr in enumerate(range(len(data_total['neural']['raw']))[params.trials.n_norm:params.trials.n_norm + params.trials.n_tot]):
    for trial_ii, trial_nr in enumerate(range(n_trials)):
        # data_trial = data_total['neural']['raw'][trial_nr][1][2]
        # arr0 = buffer_to_array(data_trial)
        arr0 = methods.get_trial_data(trial_ii, data_total)
        arr = signal.resample(arr0, arr0.shape[1] // params.sampling.smpl_fct, axis=1)          # downsample
        fs = params.sampling.fs // params.sampling.smpl_fct

        av = np.mean(arr, axis=0)                                               # car: remove common average
        arr2 = arr - av
        arr2_lp, taps = methods.low_pass_filter(arr2, params)

        f3, t3, spectrogram = signal.stft(arr2_lp, fs=fs, window='hamming', nperseg=256, noverlap=203)

        # NORMALIZE
        norm_mat2 = norm_mat.reshape(norm_mat.shape + (1,)).repeat(spectrogram.shape[2], axis=2)
        spectrogram = np.abs(spectrogram) / norm_mat2

        features_trial = spectrogram[features[:, 0], features[:, 1], :]
        if trial_ii < 2:
            class1.append(features_trial)
        else:
            class2.append(features_trial)


    n_bl = 2
    n_cues = 2
    class1 = np.array(class1).reshape(n_bl, -1)
    class2 = np.array(class2).reshape(n_cues, -1)
    # class1 = np.array(class1).flatten()     # length: n_bl x n_features x tbins
    # class2 = np.array(class2).flatten()     # length: n_cues x n_features x tbins


    np.savez('classes.npz', class1=class1, class2=class2)

    # plt.draw()
    # plt.show()
