'''
description: signal processing and decoding methods
author: Ioannis Vlachos
date:   16.07.18

Copyright (c) 2018 Ioannis Vlachos.
All rights reserved.'''

import matplotlib.pyplot as plt
import numpy as np
from scipy import signal

from aux import log


#Plot frequency and phase response
def mfreqz(b, a=1, fs=0):
    w, h = signal.freqz(b,a)
    h_dB = 20 * np.log10 (abs(h))
    plt.subplot(211)
    plt.plot((w * fs) / (2 * np.pi), h_dB)
    plt.ylim(-150, 5)
    plt.ylabel('Magnitude (db)')
    plt.xlabel(r'Normalized Frequency (x$\pi$rad/sample)')
    plt.title(r'Frequency response')

    plt.subplot(212)
    h_Phase = np.unwrap(np.arctan2(np.imag(h), np.real(h)))
    plt.plot(w / max(w), h_Phase)
    plt.ylabel('Phase (radians)')
    plt.xlabel(r'Normalized Frequency (x$\pi$rad/sample)')
    plt.title(r'Phase response')
    plt.subplots_adjust(hspace=0.5)

def low_pass_filter(arr, fs, fc):
    '''fs: sampling frequency, fs: cutoff frequency'''

    fn = fs / 2.0                                   # nyquist rate
    width = 500.0 / fn                              # transition width relative to Nyquist rate
    attenuation = 20.0                              # db, stop band attenuation

    N, beta = signal.kaiserord(attenuation, width)
    filt_coef = signal.firwin(N, fc / fn, window=('kaiser', beta))
    res = signal.lfilter(filt_coef, 1.0, arr)            # FIR filter

    return res, filt_coef


def high_pass_filter(arr, fs, fc):
    '''fs: sampling frequency, fs: cutoff frequency'''

    fn = fs / 2.0                                   # nyquist rate
    width = 1. / fn                              # transition width relative to Nyquist rate
    attenuation = 8.1                              # db, stop band attenuation

    N, beta = signal.kaiserord(attenuation, width)
    filt_coef = signal.firwin(N, [1. / fn, 0.9], window=('kaiser', beta), pass_zero=False)
    res = signal.lfilter(filt_coef, 1.0, arr, axis=1)            # FIR filter
    # res = signal.filtfilt(filt_coef, 1.0, arr, axis=1, padlen=arr.shape[1]-1)

    # log.info(f'filter coeffs: {N}')
    return res, filt_coef


def filter_response(filt_coef, fs, fc, xlim=100):
    '''plot frequency response of filter'''
    w, h = signal.freqz(filt_coef, worN=8000)
    # w, h = signal.freqz(filt_coef, 1, [], fs)

    plt.figure(6)
    plt.clf()
    plt.subplot(211)
    plt.plot((w / np.pi) * fs / 2., 20 * np.log10(np.absolute(h)), linewidth=2)
    plt.ylabel('Amplitude (db)')
    plt.title('Frequency Response')
    plt.legend(['Amplitude Response'])
    plt.xlim(0, xlim)
    plt.ylim(-15, 10)
    plt.grid(True)

    plt.subplot(212)
    plt.plot((w / np.pi) * fs / 2., np.unwrap(np.angle(h)), linewidth=2)
    plt.xlabel('Frequency (Hz)')
    plt.ylabel('Angle (rad)')
    plt.legend(['Phase Response'])
    plt.xlim(0, xlim)
    plt.grid(True)


def validate_norm_mat2(arr1, arr2):
    for ii in range(arr2.shape[2]):
        assert np.array_equal(arr1, arr2[:,:,ii])
    log.info('normalization matrix is ok')
    return None


def get_trial_data(trial_id, data):
    if trial_id == 0:
        res = data[-30000 * (trial_id + 1):, :]
    else:
        res = data[-30000 * (trial_id + 1): -30000 * (trial_id), :]
    print(np.mean(res))
    return res.T
