"""
Configuration script for NSP/nPlayServer.

Sets given channnels to broadcast at a given rate.

Author: Ali Zaidi, Ioannis Vlachos
Date: 01.01.2017
(c) All Rights Reserved

"""

import sys
import time

import cerebus.cbpy as cb
import numpy as np


def main(mychannels=list(range(1, 97)), mygroup=5):
    """
    To configure or unnconfigure channnels on the NSP/nPlayServer

    syntax:

    unconfigured channnels = main(channel_list, sampling_group)

    inputs:
        mychannnels: list of channels to be activated
        mygroup: sampling group to assign channels

    Calling "main()" will configure 96 channels at 30k. Easy peasy

    Examples:

    To turn off all channels:
        unconfigd_chans = main(list(range(1,1)), 1)

    To turn on channels 1:5 at 30k:
        main(list(range(1,6)), 5)

    To turn on channels 1, 4 and 9 at 1k:
        main([1,4,9], 2)

    """

    res = cb.open()
    time.sleep(2)

    # First get a list of all channels that are active but not in my list

    all_groups = list(range(1, 6))
    print("Gathering information from NSP...")
    active_channels = []
    for group in range(1, 6):
        chans = cb.get_sample_group(group)
        for channel in range(len(chans[1])):
            active_channels.append(chans[1][channel]['chan'])
            time.sleep(0.1)

    # print("active channels = {}".format(active_channels))

    # Turn off those channnels

    channels_to_deactivate = np.setdiff1d(active_channels, mychannels)

    if channels_to_deactivate != []:
        print("Deactivating unnecessary channels...")

    for channel in channels_to_deactivate:
        res = cb.set_channel_config(channel, chaninfo={'smpgroup': 0})
        time.sleep(0.2)

    print("Configuring your channels...")
    # Now configure your channels
    for channel in mychannels:
        res = cb.set_channel_config(channel, chaninfo={'smpgroup': mygroup})
        time.sleep(0.2)

    # Check if channels have been configured correctly

    print("Verifying configuration...")
    active_channels = []
    chans = cb.get_sample_group(mygroup)
    for channel in range(len(chans[1])):
            active_channels.append(chans[1][channel]['chan'])

    unconfigured_channels = np.setdiff1d(mychannels, active_channels)
    if unconfigured_channels != []:
        print("unconfigured channels = {}".format(unconfigured_channels))

    if(len(unconfigured_channels) == 0):
        print("SUCCESS!")
        if mychannels == []:
            print("All channels are now off")
        else:
            print("assigned {} channel(s) to group {}".format(len(mychannels), mygroup))
    else:
        print("Failed! {} channels are unconfigured".format(len(unconfigured_channels)))

    res = cb.close()

    return unconfigured_channels


if __name__ == '__main__':

    if len(sys.argv)<2:
        print('Please provide the number of channels to switch on.')
    else:
        n_ch = int(sys.argv[1])
        main(list(range(1, n_ch+1)), 5)     # switch n_ch channels on
