#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Jul  2 18:35:52 2018

@author: ale, modified by Ioannis Vlachos

@version: 0.1

@TODO: in play_stimulus decide when to play the whole string
@TODO: put german corpus -> umlaut????
@TODO: save data
"""
import os
import pickle
import re

from nltk.corpus.reader.plaintext import PlaintextCorpusReader
from nltk.probability import ConditionalFreqDist
#from nltk.corpus.reader.conll import ConllCorpusReader
from nltk.tokenize.regexp import RegexpTokenizer
from nltk.util import everygrams

import aux
from aux import log
from paradigm import Paradigm

import tkinter as tk
from tkinter import simpledialog


class ColorSpeller(Paradigm):
    '''
    Paradigm for free spelling. Letters are grouped in different colors, the 
    user can choose colors and letters with different confirmation modality.
    Moreover a there is a sector with special characters from where the user 
    can delete, insert punctuation, select words from vocabulary and quit.
    If enabled, the paradigm provides also word prediction. Both word prediction
    and words in vocabulary are obtained from the current selection and the 
    previous words.
    '''
    def __init__(self, general_cfd=[], pyttsx_rate=100, audio=True, **kwargs):
        '''
        Initialize the paradigm.
        
        Variables:
            trigger: list[int] (triggers for start and end of stimulus presentation)
            states: list[string] (structure of the grid with the letters)
            method: string (method used for confirmation of the answers)
            mode: string (mode of the paradigm)
            validation_string: string (string used for validation method)
            max_length_vocabulary: int (number of words presented in the vocabulary)
            current_state = list[int] (current row of the states grid and index of the letter)
            current_type = string (type of the current selection: sector, letter or word)
            current_string = string (string currently written)
            current_selection = string (string of the current selection)
            yes, no, nc = int (internal values for yes, no and not classified selections)
            corpora_path: string (path where the used corpora are saved)
            general_user_corpus_name: string (name of the corpus with sentences from the user, extracted before impantation; emails, etc)
            speller_user_corpus_name: string (name of the corpus with sentences from the user, after implantation; speller typed sentances)
            general_cfd: ConditionalFreqDist (cfd from the general corpus)
            vocabulary: list[str] (vocabulary with the proposed words)
            word_prediction: boolean (if True word prediction is enabled)
            general_user_cfd: ConsitionalFreqDist (cfd from the user corpus)
            speller_user_cfd: ConsitionalFreqDist (cfd from the user corpus)
            history: list[(list[int],str,bool,str,str,str)] (history of the stimulus presentation: state, type, decision, string, selection and validation string])
        '''
        super().__init__(**kwargs)

        self.trigger = [1,2]
        
        self.engine.setProperty('rate',pyttsx_rate)
        log.info(f'pyttsx3 rate is: {pyttsx_rate}')

        # read configuration file
        self.config = self.params.paradigms.color

        self.audio_path = self.config.audio_path
        
        # state initialized based on the frequency of the letters
        self.states = self.config.states
        # self.method = self.config.confirmation_methods[self.config.selected_confirmation_method]
        
        self.confirm_yes = self.config.confirmation_num_yes
        self.confirm_no = self.config.confirmation_num_no
        self.opposite_answers = self.config.opposite_answers_methods[self.config.selected_opposite_answers_method]
        
        self.mode = self.config.mode[self.config.selected_mode]
        self.validation_string = self.config.validation_string
        self.max_length_vocabulary = self.config.max_length_vocabulary
        
        self.current_state = [0,0]
        self.current_type = 'sector'

        self.current_string = self.config.init_string
        self.current_selection = self.states[0][0]
        
        # values for yes, no, not confirmed and unclassified
        self.yes = aux.decision.yes
        self.no = aux.decision.no
        self.nc = aux.decision.nc                           #not confirmed
        self.unclassified = aux.decision.unclassified
        self.n_unclassified = 0         # track number of non-classified responses
        self.n_yes = 0 # track number of yes answers
        self.n_no = 0 # track number of no answers

        # word prediction
        self.corpora_path = self.config.corpora_path
        corpus_name = self.config.general_corpus
        self.general_user_corpus_name = self.config.general_user_corpus
        self.speller_user_corpus_name = self.config.speller_user_corpus
        
        if corpus_name[-2:] == '.p':
            if general_cfd == []:
                try:
                    log.info('Loading pickle file')
                    self.general_cfd = pickle.load(open(self.corpora_path+os.sep+corpus_name,'rb'))
                    print('Pickle file loaded!')
                except Exception as e:
                    log.error(e)
            else:
                log.info('Using General cfd from memory.')
                self.general_cfd = general_cfd
        else:
            self.general_cfd = self._createCFD(root=self.corpora_path,corpus_name=corpus_name)
        
        if self.general_user_corpus_name[-2:] == '.p':
            self.general_user_cfd = pickle.load(open(self.corpora_path+os.sep+self.general_user_corpus_name,'rb'))
        elif self.general_user_corpus_name[-4:] == '.txt':
            self.general_user_cfd = self._createCFD(root=self.corpora_path,corpus_name=self.general_user_corpus_name)
        else:
            self.general_user_cfd = []
            print('General_user_cfd not loaded!')

        if self.speller_user_corpus_name[-2:] == '.p':
            self.speller_user_cfd = pickle.load(open(self.corpora_path+os.sep+self.speller_user_corpus_name,'rb'))
        elif self.speller_user_corpus_name[-4:] == '.txt':
            self.speller_user_cfd = self._createCFD(root=self.corpora_path,corpus_name=self.speller_user_corpus_name)
        else:
            self.speller_user_cfd = []
            print('speller_user_cfd not loaded!')

        # create vocabulary
        self.vocabulary = []
        self.probable_word = False
        self._update_words_probability()
        
        self.word_prediction = self.config.word_prediction
        
        self.history = []
        
        # manual updates
        self.string_updated = False
        self.new_string = ''
        self.vocabulary_updated = False
        self.suggestion = ''

        log.info('init ok')
        
    def _createCFD(self,root,corpus_name,max_ngram=3):
        '''
        Create the n-grams conditional frequency distribution using one 
        specific corpus
        
        Parameters:
            root: string (path of the corpus)
            corpus_name: string (name of the corpus)
            max_ngram: int (maximum considered ngrams)
        
        Return:
            cfd: ConditionalFreqDist (created conditional frequency ditribution)
        '''
        # root = root
#        fileid = 'tiger_release_aug07.corrected.16012013.conll09'
#        columntypes = ['ignore','words','ignore','ignore','pos']
#        german_corpus = ConllCorpusReader(root,fileid,columntypes,encoding='utf8')
        fileid = corpus_name
        sent_tok = RegexpTokenizer(pattern='\n?[0-9]+\t',gaps=True,discard_empty=True,flags=56)
        
        raw_corpus = PlaintextCorpusReader(root,fileid,sent_tokenizer=sent_tok)
           
        all_grams = []
        for sen in raw_corpus.sents():
            corpus = [w.upper() for w in sen if w.isalpha()]
            all_grams += everygrams(corpus, max_len=max_ngram)
        
        cfd = ConditionalFreqDist((tuple(a),b) for *a,b in all_grams)
        return cfd

    def present_start(self, audio=True):
        '''
        Play auditorly the start stimulus using the play function defined in
        the abstract class

        Return:
            None
        '''
        if audio:
            stim_path = self.audio_path + os.sep + 'beginning.wav'
            try:
                self._play_sound(stim_path, 0, 0)
            except Exception as e:
                raise e
        return None

    def process_result(self, decision, audio=False):
        ''' 
        Append the last state, type, decision, string and selection to history.
        Chek if the answer is confirmed, if no ask for confirmation, if yes 
        select or skip the current state.
        
        Parameters:
            decision: boolean (value of the last decision)
            
        Return:
            finish: boolean (True if the paradigm is finished, False otherwise)            
            
        '''

        self.history.append([self.current_state,self.current_type,decision,self.current_string,self.current_selection,self.validation_string])
        # Check if the decision is confirmed based on history (and on last decision)
        #confirmed, answer = self._check_confirmation(self.method)
        #log.info(f'confirmed:{confirmed}, answer: {answer}')

        # New confirmation method based on number of answers
        confirmed, answer = self._check_confirmation_new(decision)
        log.info(f'confirmed:{confirmed}, answer:{answer}')
        
        if not confirmed:
            finish = self._confirm_result()
        elif answer is self.yes.value:  #The state is confirmed and the decision is Yes
            # os.system("aplay /kiap/data/speller/Audio/question/feedback/ja.wav -q")
            self._say('Ja',self.trigger[0],self.trigger[1])
            finish = self._select()
            self.n_unclassified = 0
        elif answer is self.no.value: #The state is confirmed and the decision is No
            # os.system("aplay /kiap/data/speller/Audio/question/feedback/nein.wav -q")
            self._say('Nein',self.trigger[0],self.trigger[1])
            finish = self._skip()
            log.info(self.current_type)
            self.n_unclassified = 0
        elif answer is self.unclassified.value:
            os.system("aplay /kiap/data/speller/Audio/question/feedback/unclassified.wav -q")
            # os.system("aplay /kiap/data/speller/Audio/question/feedback/danke.wav -q")
            # self._say('Antwort nicht erkannt',self.trigger[0],self.trigger[1])
            # finish = self._skip()
            if self.n_unclassified < self.config.n_repeat_unclassified: 
                finish = self._confirm_result()         # iv: repeat state for unclassified responses
                self.n_unclassified +=1
            else:
                finish = self._skip()
                self.n_unclassified = 0
            log.warning(f'n_unclassified: {self.n_unclassified}')
        else:
            log.warning('Unclassified decision')
            finish = self._skip()

        log.warning(f'finish: {finish}')
        return finish
    
    def _check_confirmation_new(self, decision):
        ''' 
        This function checks if the answer is confirmed or not
        
        Return:
            confirmed: boolean (True if the decision is confirmed)
            answer: int (yes, no or not classified)
        '''
        if decision == self.yes.value:
            self.n_yes += 1
            # self.n_unclassified = 0 #uncomment to reset unclassified after every classified answer
        elif decision == self.no.value:
            self.n_no += 1
            # self.n_unclassified = 0 #uncomment to reset unclassified after every classified answer
        elif decision == self.unclassified.value:
            self.n_unclassified +=1
            log.warning(f'n_unclassified: {self.n_unclassified}')
            if self.n_unclassified <= self.config.n_repeat_unclassified: 
                confirmed = False
                answer = self.nc.value         # iv: repeat state for unclassified responses
            else:
                confirmed = True
                answer = self.unclassified.value
                self.n_unclassified = 0
                self.n_yes = 0
                self.n_no = 0
            return confirmed, answer
            
        # It works only if confirm_yes and confirm_no are the same or one is the other+1
        if (self.n_yes < self.confirm_yes) and (self.n_no < self.confirm_no):
            confirmed = False
            answer = self.nc.value
        elif self.n_yes == self.n_no:
            if self.opposite_answers == 'last':
                confirmed = True
                answer = decision
                self.n_yes = 0
                self.n_no = 0
                self.n_unclassified = 0
            elif self.opposite_answers == 'majority':
                confirmed = False
                answer = self.nc.value
        else:
            confirmed = True
            answer = decision
            self.n_yes = 0
            self.n_no = 0
            self.n_unclassified = 0
            
        return confirmed, answer
    
#    def _check_confirmation(self, method):
#        
#        ''' 
#        This function calls the correct confirmation method
#        
#        Parameters:
#            method: string (confirmation method)
#        '''
#        if method == 'single':
#            return self._confirmation_single()
#        elif method == 'double_yes':
#            return self._confirmation_double_yes()
#        elif method == 'best_of_three':
#            return self._confirmation_best_of_three()
#        
#    def _confirmation_double_yes(self):
#        ''' 
#        Confirmation method
#        confirmed yes: two consecutive yes
#        confirmed no: single no
#        
#        Return:
#            confirmed: boolean (True if the decision is confirmed)
#            answer: int (yes, no or not classified)
#        '''
#        last_decision = self.history[-1][2]
#        log.error(f'history: {self.history[-1]}, n_uncl: {self.n_unclassified}, last_decision: {last_decision}, ')
#
#
#        if last_decision == self.no.value:  #the last answer was no: No confirmed
#            confirmed = True
#            answer = self.no.value
#            self.n_unclassified = 0
#        elif (len(self.history) >= 2) and (self.history[-1] == self.history[-2]): #last and previous both the same question and both true: Yes confirmed
#            confirmed = True
#            answer = last_decision
#            self.n_unclassified = 0
#        elif (last_decision == self.unclassified.value):
#            self.n_unclassified +=1
#            if self.n_unclassified == 2:
#                confirmed = True
#                answer = last_decision
#                self.n_unclassified = 0
#            else:
#                confirmed = False
#                self.history.pop(-1)
#                answer = last_decision
#        else: #the answer is not confirmed
#            confirmed = False
#            answer = self.nc.value
#            # if last_decision == self.yes.value:
#                # os.system("aplay /kiap/data/speller/Audio/question/feedback/ja.wav -q")
#                # self._say('Ja',self.trigger[0],self.trigger[1])
#
#        # log.error(f'history: {self.history[-1]}, confirmed: {confirmed}, answer: {answer}')
#        return confirmed, answer
#    
#    def _confirmation_single(self):
#        ''' 
#        Confirmation method:
#        confirmation yes: single yes
#        confirmation no: single no
#        
#        Return:
#            confirmed: boolean (True if the decision is confirmed)
#            answer: int (yes, no or not classified)
#        '''
#
#        confirmed = True
#        answer = self.history[-1][2]
#
#        log.info(f'Answer:{answer}, history: {self.history}')
#        return confirmed, answer
#    
#            
#    def _confirmation_best_of_three(self):
#        ''' 
#        Confirmation method:
#        confirmation yes: two yes over three questions
#        confirmation no: two no over three questions
#        
#        Return:
#            confirmed: boolean (True if the decision is confirmed)
#            answer: int (yes, no or not classified)
#        '''
#        last_decisions = [item[2] for item in self.history]
#        # log.error(f'last_decisions: {last_decisions}')
#
#        if len(self.history) < 2:
#            confirmed = False
#            answer = self.nc.value
#
#        elif self.history[-2] == self.history[-1]:
#            confirmed = True
#            # if last_decisions[-1][2] is True:
#            #     answer = self.yes.value
#            # else:
#            #     answer = self.no.value
#            answer = last_decisions[-1]
#        elif len(self.history) > 2 and self.history[-3][:2] == self.history[-2][:2] == self.history[-1][:2]:
#            confirmed = True
#            answer = last_decisions[-1]
#            # numeric_answer = 0
#            # for decision in last_decisions[-3:]:
#            #     if decision == self.yes.value:
#            #         numeric_answer += 1
#            #     else:
#            #         numeric_answer -= 1
#            # if numeric_answer > 0:
#            #     answer = self.yes.value
#            # else:
#            #     answer = self.no.value
#        else:
#            confirmed = False
#            answer = self.nc.value
#
#        # if confirmed == False and last_decisions[-1] == self.no.value:        # tbd whether to include or not
#            # self._say('Ja',self.trigger[0],self.trigger[1])
#        # elif confirmed == False and last_decisions[-1] == self.yes.value:
#            # self._say('Nein',self.trigger[0],self.trigger[1])
#
#        log.error(f'history: {self.history[-1]}, confirmed: {confirmed}, decision: {last_decisions}, answer: {answer}')
#        return confirmed, answer

    def _confirm_result(self):
        '''
        If result is not confirmed just ask the same state
        '''
        finish =self._next_state(self.current_state, self.current_type)

        return finish
        
    def _select(self):
        '''
        Select the state updating the next state with the following one.
        If it the state was a sector, the following will be a letter. If it was
        a letter or a word next could be a word if there is any probable, or 
        again the first sector.
        '''        
        if self.current_type == 'sector':
            next_state = [self.current_state[0],0]
            next_type = 'letter'
        else: #letter or phrase
            selection = self._update_string(self.current_state, self.current_type)
            if selection == 'word':
                if len(self.vocabulary) == 0:
                    self._say('Kein wahrscheinliches Wort',self.trigger[0],self.trigger[1])
                    next_state = [0,0] #Reset to the first state
                    next_type = 'sector'
                else:
                    next_state = [0,0]  
                    next_type = 'word'
            elif selection == '^':
                next_state = [self.current_state[0]+1,0]
                next_state[0] %= len(self.states)
                # if self.current_state[0] == 4:      # IV: if at the last '^', then go to the beginning 
                #     next_state = [0,0]  
                # else:
                #     next_state = [self.current_state[0]+1,0]
                next_type = 'sector'
            elif selection == 'end':
                # finish = True
                return True
            else:               
                self._update_states() #Update the states based on the probability
                probable_word = self._update_words_probability()
                if probable_word and self.word_prediction:
                    next_state = [1,0]
                    next_type = 'word'
                else:
                    next_state = [0,0] #Reset to the first state
                    next_type = 'sector'
        
        finish = self._next_state(next_state, next_type)

        return finish
        
    def _skip(self):
        '''
        Skip the state updating the next state with the following one.
        The following state will be the next sector, letter or word, depending
        on the current state
        '''
        if self.current_type == 'sector':
            next_state = self.current_state[:]
            next_state[0] += 1
            next_state[0] %= len(self.states)
            next_type = 'sector'
        elif self.current_type == 'word' and (
                # self.current_state[0] == 1 # Was a probable word
                self.probable_word == True
                or self.current_state[1] == len(self.vocabulary)-1): # End of vocabulary
            next_state = [0,0]
            next_type = 'sector'# Skip and go back to the sectors
            self.probable_word = False
        else: #letter or selected word
            next_state = self.current_state[:]
            next_state[1] += 1     
            next_state[1] %= len(self.states[next_state[0]][1])
            next_type = self.current_type
            
        finish = self._next_state(next_state, next_type)

        return finish

    
    def _update_string(self,state,state_type):
        '''
        Update the current string. If the string is updated with a word append 
        a space at the end.
        
        Parameters:
            state: list[int] (indexes of the current state)
            state_type: string (type of the current state)
            
        Return:
            selection: string (the text appended to the current string)
        '''
        
        if state_type == 'letter':
            selection = self.states[state[0]][1][state[1]]
            if selection not in ('word', '^', '<', '<<<', '.', 'end'):
                self.current_string += selection
            elif selection == '<':
                self.current_string = self.current_string[:-1]
            elif selection == '<<<':
                all_words = self.current_string.split()[:-1]
                self.current_string = ' '.join(all_words)
                if self.current_string:              
                    self.current_string += ' '
            elif selection == '.':
                self.current_string += '. '
            elif selection == 'end':
                # if self.current_string[-1].isalpha():       #will cause a problem is string is empty
                self.current_string += '.'
            
                
        else: #is a word
            selection = self.vocabulary[state[1]]
            if not self.current_string or self.current_string[-1].isspace():
                self.current_string += selection + ' '
            else: #the word is to complete the current string
                tokens = self.current_string.split()
                tokens[-1] = selection + ' '
                self.current_string = ' '.join(tokens)
        
        return selection
        
    def _update_states(self):
        # Update the states based on the previous selected letters
        # Sort the letters based on the probability based on the current string
        '''
        for letter in letters:
            letter_prob[letter] = calculateprobability(current_string,letter)
        letter_order = [i[0] for i in sorted(enumerate(letter_prob), key=lambda x:x[1])]
        # Sort the sectors summing the probability of the letters inside each sector
        for sector in sectors:
            sector_prob[sector] = sum(prob[letter] for letter in sector)
        sector_order = [i[0] for i in sorted(enumerate(sector_prob), key=lambda x:x[1])]
        
        # Order the letters inside each sector based on the probability
        for letters in sector:
            sort(letters)
        '''
        pass
        
    def _update_words_probability(self,max_ngram=3,confidence=0.5):
        '''
        Update the vocabulary based on the current string and on the 
        conditional frequency distributions. Based on the ngrams of the cfd 
        check if there is some word that can follow the last n-1 words of the
        current string. The function will always try to select words based on 
        higher ngrams. The vocabulary is populated with the most probable words.
        
        Parameters:
            max_ngram: int (higher ngram that will be considered)
            confidence: float (between 0 and 1, level of confidence)
        Return: 
            is_probable: boolean (True if the probability of one word is higher than confidence*all_probabilities)
        
        '''
        self.vocabulary = []
        seed = self.current_string.split()
        
        if not self.current_string or self.current_string[-1].isspace():
            current_word = ''
        else:
            current_word = seed.pop()
        #weights_ngrams = [0.25,0.5,1]
        
        max_length_seed = min(len(seed),max_ngram-1)
        
        most_probable = {}
        
        for j in range(max_length_seed,-1,-1):
            if j > 0:
                seed_key = tuple(seed[-j:])
            else:
                seed_key = tuple([])
            # for cfd in [self.general_cfd, self.general_user_cfd, self.speller_user_cfd]:
            for cfd in [self.speller_user_cfd, self.general_cfd, self.general_user_cfd]:
                num_hapaxes = len(cfd[seed_key].hapaxes())
                if num_hapaxes:
                    non_hapaxes = cfd[seed_key].most_common()[:-num_hapaxes]
                else:
                    non_hapaxes = cfd[seed_key].most_common()
                for (key,val) in non_hapaxes:
                    if key.startswith(current_word):
                        freq = cfd[seed_key].freq(key)
                        if key in most_probable:
                            most_probable[key] += freq
                        else:
                            most_probable[key] = freq
                        
            if len(most_probable) >= self.max_length_vocabulary:
                    break
            else:
                continue
            
        sort_most_prob = sorted(most_probable.items(), key=lambda x: x[1])
        sort_most_prob.reverse()
        
        if not sort_most_prob or sort_most_prob[0][1] < confidence*sum(most_probable.values()):
            is_probable = False
        else:
            is_probable = True
            
        self.vocabulary = [key for key,val in sort_most_prob[:self.max_length_vocabulary]]
        self.probable_word = is_probable
        
        return is_probable
    
    def _next_state(self, next_state, next_type):
        '''
        Update the current_state, type and selection.
        
        Parameters:
            next_state: list[int] (index of the next state)
            next_type: string (next type)
            
        Return:
            finish: boolean (True if the paradigm is finish)
        '''
        if next_type == 'end':
            # finish = True
            return True
        self.current_state = next_state
        self.current_type = next_type
        if self.current_type == 'sector':
            self.current_selection = self.states[self.current_state[0]][0]
        elif self.current_type == 'letter':
            self.current_selection = self.states[self.current_state[0]][1][self.current_state[1]]
        else:
            self.current_selection = self.vocabulary[self.current_state[1]]
        # finish = False
        return False
        #data.write_log(current_state)
        #main.start_trial(params)
        
    def present_stimulus(self,audio=True):
        ''' 
        Play auditorly the current stimulus using the say function defined in 
        the abstract class
        
        Return:
            stimulus: string (the stimulus that is presented)
        '''
        
        stimulus = self.current_selection

        if audio is True:
            
            # if (not self.current_string) or (len(self.history) < 1) or self.history[-1][2] != self.yes.value:
            if (not self.current_string) or (len(self.history) < 1):
                str2s = ''
                stim2s = ''
            
            # elif self.probable_word or ((self.history[-1][1] in {'letter','word'}) and (self.history[-1][1] != self.current_type)):
            elif self.current_string != self.history[-1][3]:
                all_sentences = self.current_string.split('.')
                last_sentence = all_sentences[-1];
                if last_sentence.endswith(' '):
                    all_words = last_sentence
                else: #otherwise do the spelling of last word
                    split_string = last_sentence.split()
                    last_word = split_string.pop()
                    last_word = ' '.join(last_word)
                    prev_words = ' '.join(split_string)
                    all_words = ' '.join([prev_words,last_word])

                    log.info(all_words)
                str2s = 'Der aktuelle Satz ist: ' + all_words +'.'
                stim2s = ''
            else:
                str2s = ''
                stim2s = ''

            if self.current_type == 'word':
                stim2s += 'Meinst du das Wort: '
                
            if stimulus == '<':
                if not self.current_string:
                    stim2s += 'Buchstabe loeschen' # useless: doesn't delete anything
                else:
                    stim2s += 'Buchstabe ' + self.current_string[-1] + ' loeschen'
            elif stimulus == '<<<':
                if not self.current_string:
                    stim2s += 'Wort loeschen' # useless: doesn't delete anything
                else:
                    stim2s += 'Wort ' + self.current_string.split()[-1] + ' loeschen'
            elif stimulus == '.':
                stim2s += 'Punkt'
            elif stimulus == ' ':
                stim2s += 'Neues Wort'
            elif stimulus == '?':
                stim2s += 'Fragezeichen'
            elif stimulus == '^':
                stim2s += 'Zurueck zu den Farben?'
            elif stimulus == 'word':
                stim2s += 'Woerterbuch?'
            elif stimulus == 'end':
                stim2s += 'Programm beenden?'
            else:
                stim2s += stimulus
            
            if str2s:
                t2s = [str2s.lower(),stim2s.lower()]
            else:
                t2s = stim2s.lower()
            # log.error(f'Audio: {t2s}')
            self._say(t2s,self.trigger[0],self.trigger[1])
            
        return stimulus
    
    def _save_string(self, string=None):
        '''
        Save the current string in the speller user corpus
        '''
        if string is None:
            string = self.current_string
        if len(string) == 0:
            return
        root = self.corpora_path
        fileid = self.speller_user_corpus_name
        
        corpus_pattern = '([0-9]+)\t'
        p = re.compile(corpus_pattern)
        
        with open(os.path.join(root,fileid),'r+') as f:
            for sen in string.split("."):
                sen_clean = sen.strip()
                if sen_clean:
                    lines = f.readlines()
                    if len(lines) > 0:
                        last_num = p.search(lines[-1]).group(1)
                    else:
                        last_num = '0'
                    f.write(str(int(last_num)+1)+'\t'+sen_clean)
                    if sen_clean[-2:] != '\n':
                        f.write('\n')
            
    
    def get_current_state(self):
        '''
        Return the current state
        
        Return:
            current_string: str (current selected string)
            current_selection: str (current selection)
        '''
        return self.current_string, self.current_selection
    
    def get_mode(self):
        '''
        Return the mode of the used paradigm
        
        Return:
            mode: string (mode of the paradigm)
        
        '''
        return self.mode
    
    def get_vocabulary(self,N=5):
        '''
        Return the most probable words
        
        Parameters:
            N: int (num of returned words)
        Return:
            vocabulary: list[string] (most probable words in order of probability)
        '''
        vocabulary = self.vocabulary[:N]
        if self.probable_word:
            vocabulary[0] = '*'+vocabulary[0]+'*'
        return vocabulary
    
    def get_current_string(self):
        '''
        Return the current string
        
        Return:
            current_string: string (current string)
        '''
        return self.current_string
    
    def set_current_string(self,string):
        '''
        Manually update the current string
        '''
        
        # Flag for manual updates
        self.string_updated = True
        
        # Save the new string
        self.new_string = string.upper()
        
        return
    
    def suggest(self,suggestion):
        '''
        Manually suggest a word
        '''
        if len(suggestion) == 0:
            return
        
        # Flag for manual updates
        self.vocabulary_updated = True
        
        # Save the suggestion in upper case and without final spaces
        self.suggestion = suggestion.upper().rstrip()
        return
    
    def external_updates(self):
        update_state = False
        finish = False
        # Check if there is an update on the current state
        if self.string_updated:
            # Reset flag
            self.string_updated = False
            # Update string
            self.current_string = self.new_string
            self.new_string = ''
            # Update word probability
            probable_word = self._update_words_probability()
            # Update state
            if probable_word and self.word_prediction:
                next_state = [1,0]
                next_type = 'word'
            else:
                next_state = [0,0] #Reset to the first state
                next_type = 'sector'
            update_state = True
            
        if self.vocabulary_updated:
            # Reset flag
            self.vocabulary_updated = False
            # Add the suggestion to the vocabulary and set it as probable
            self.vocabulary = [self.suggestion]+self.vocabulary
            self.probable_word = True
            # Update state
            next_state = [2,0] #2 to indicate a suggested word
            next_type = 'word'            
            update_state = True
                
        if update_state:
            finish = self._next_state(next_state, next_type)
            # Reset answers
            self.n_yes = 0
            self.n_no = 0
            self.n_unclassified = 0
        
        return finish
    
    def save_log(self):
        '''
        Save the log
        
        '''
        pass
    
    def close(self):
        '''
        Save the current string and close the speller with the super class function
        '''
        # Maybe first ask the user?
        # Pop up the current string and ask if it has to be saved
        root = tk.Tk()
        root.withdraw()
        save_str = simpledialog.askstring(title="Save sentence", 
            prompt="The current sentence is: "+self.current_string, 
            initialvalue=self.current_string)
        # If "Ok" save the entered string
        if save_str is not None:
            self._save_string(save_str)

        # Close using the super method
        super(ColorSpeller, self).close()
        root.destroy()
        return
