#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Jul 27 15:06:04 2018

@author: ale
"""

import os
import random

import aux
from aux import log
from paradigm import Paradigm
import pyaudio
from scipy.io import wavfile
from scipy import signal
import numpy as np
import munch

def norm2freq(norm_rate, alpha=360, beta=80):
    return norm_rate * alpha + beta

def note(freq, t_len, amp=1, rate=44100, mode='sin'):
    n_sample = int(t_len * rate)
    factor = (float(freq) * np.pi * 2 / rate) 
    data = []
    if mode == 'saw':
        data = signal.sawtooth(np.arange(n_sample) * factor, .25) * amp
    else:
        data = np.sin(np.arange(n_sample) * factor) * amp
    fade = int(n_sample/5)
    fade_in = np.arange(0., 1., 1/fade)
    fade_out = np.arange(1., 0., -1/fade)

    data[:fade] = np.multiply(data[:fade], fade_in)
    data[-fade:] = np.multiply(data[-fade:], fade_out)
    return data.astype(np.float32)

def load_wav(wav_file, channels=2, rate=44100):
    r, wf = wavfile.read(wav_file)
    if wf.dtype == 'int16':
        wf = np.float32(wf) / pow(2,15)
    if wf.shape[1] > channels:
        wf = wf[:, :channels]
    elif wf.shape[1] < channels:
        if wf.shape[1] == 1:
            wf = np.repeat(wf, channels, axis=1)
    if r != rate:
        n_s = int(wf.shape[0] * rate / r)
        wf = signal.resample(wf, n_s)
    return wf


RATE = 44100

class Question_Feedback(Paradigm):
    '''
    Paradigm for training and validating the goodness of a signal.
    Questions with two different labels (yes/no) will be presented.
    The paradigm can be run in training or validation mode, i.e. without or
    with feedback.
    '''
    def __init__(self, **kwargs):
        '''
        Initialize the paradigm.

        Variables:
            trigger_train: list[int] (for training mode, triggers for start and end yes questions and start and end no questions)
            trigger_no: list[int] (for validation mode, triggers for start and end no questions and start and end feedback)
            trigger_yes: list[int] (for validation mode, triggers for start and end yes questions and start and end feedback)
            prefix_yes: string (prefix of audio files for yes questions)
            prefix_no: string (prefix of audio files for no questions)
            audio_path: string (path where the audio files are stored)
            num_stim: int (number of stimuli presented for each state)
            mode: string (mode of the paradigm)
            list_of_stimuli: list[string] (stimuli that will be presented)
            current_stimulus: string (current stimulus)
            history: list[(string,boolean)] (history of the stimulus presentation: stimulus, decision)
        '''
        super().__init__(**kwargs)
        
        # read configuration file
        self.trigger_train = [1,2,3,4]
        self.trigger_no = [21,22,23,24]
        self.trigger_yes = [11,12,13,14]
        self.start_triger = [110, 111]
        self.prefix_yes = '001'
        self.prefix_no = '002'

        self.audio_fb_target = audio_fb_target
        config = self.params.paradigms.Question_feedback

        self.audio_path = config.audio_path

        self.num_stim = config.number_of_stim

        self.mode = config.mode[config.selected_mode]

        self.list_of_stimuli = self._create_list(config.symmetrical)
        # try:
        self.current_stimulus = self.list_of_stimuli.pop(0)

        self.params = aux.load_config()
        self.yes = aux.decision.yes
        self.no = aux.decision.no
        # except:
            # raise Exception()
            # self.__del__()
            # return None
            # pass


        self.history = []

    def _create_list(self,symmetrical_questions):
        '''
        Create the list of stimuli selecting random questions balanced between yes and no

        Parameters:
            symmetrical_questions: boolean (True if yes and no questions with the same file name but different prefix)

        Return:
            list_of_selected: list[str] (List of selected questions)
        '''
        try:
            all_files_yes = [file for file in os.listdir(self.audio_path) if file.startswith(self.prefix_yes)]
            all_files_no = [file for file in os.listdir(self.audio_path) if file.startswith(self.prefix_no)]
        except Exception as e:
            raise e

        random.seed()
        if symmetrical_questions:
            random.shuffle(all_files_yes)
            nofiles = [self.prefix_no+yesfile[3:] for yesfile in all_files_yes[:self.num_stim]]
            list_of_selected = all_files_yes[:self.num_stim]+nofiles
        else:
            random.shuffle(all_files_yes)
            random.shuffle(all_files_no)
            list_of_selected = all_files_yes[:self.num_stim] + all_files_no[:self.num_stim]

        random.shuffle(list_of_selected)

        # list_of_selected = ['001_2038.wav', '002_2004.wav']*10      # jannis: only for testing, remove afterwards !!!!
        # list_of_selected = ['001_11172.wav', '002_11172.wav']*10      # jannis: only for testing, remove afterwards !!!!
        return list_of_selected


    def process_result(self, decision=None, audio=True):
        '''
        Append the last stimulus and decision to history.
        Give feedback: if training only a beep, if validation the decision
        Update the current_stimulus variable

        Parameters:
            decision: boolean (value of the last decision)

        Return:
            finish: boolean (True if the paradigm is finished, False otherwise)

        '''
        self.history.append([self.current_stimulus, decision])
        if self.mode == 'Training':
            beep_wav = self.audio_path+os.sep+'feedback'+os.sep+'beep.wav'
            danke_wav = self.audio_path+os.sep+'feedback'+os.sep+'danke.wav'
            try:
                self._play_audio(danke_wav,self.trigger_train[2],self.trigger_train[3])
            except Exception as e:
                log.error(f'playing BEEP sound exception: {e}')
            log.debug('BEEP done')

        else:
            if True:
                ja_wav = self.audio_path+os.sep+'feedback'+os.sep+'ja.wav'
                nein_wav = self.audio_path + os.sep + 'feedback'+os.sep+'nein.wav'
                beep_wav = self.audio_path+os.sep+'feedback'+os.sep+'beep.wav'
                unclassified_wav = self.audio_path+os.sep+'feedback'+os.sep+'unclassified2.wav'
                danke_wav = self.audio_path+os.sep+'feedback'+os.sep+'danke.wav'
                if audio:
                    if decision==self.yes.value:
                        self._play_audio(ja_wav, self.trigger_yes[2],self.trigger_yes[3])
                    elif decision==self.no.value:
                        self._play_audio(nein_wav, self.trigger_no[2],self.trigger_no[3])
                    else:
                        log.warning(f'decision: {decision}')
                        self._play_audio(danke_wav, self.trigger_train[2],self.trigger_train[3])

            else:
                hit_wav = self.audio_path+os.sep+'feedback'+os.sep+'hit.wav'
                miss_wav = self.audio_path+os.sep+'feedback'+os.sep+'miss.wav'
                '''
                if (decision and current_stimulus[:3] == self.prefix_yes) or (!decision and current_stimulus[:3] == self.prefix_no):
                    self._play_audio(hit_wav,self.trigger_hit[0],self.trigger_hit[1])
                else:
                    self._play_audio(miss_wav,self.trigger_miss[0],self.trigger_miss[1])
                '''
        if self.list_of_stimuli:
            self.current_stimulus = self.list_of_stimuli.pop(0)
            finish = False
        else:
            finish = True

        return finish

    def present_start(self, audio=True):
        '''
        Play auditorly the start stimulus using the play function defined in
        the abstract class

        Return:
            None
        '''
        if audio:
            stim_path = self.audio_path.split('/question')[0] + os.sep + 'beginning.wav'
            try:
                self._play_sound(stim_path, self.start_triger, self.start_triger)
            except Exception as e:
                raise e
        return None

    def present_stimulus(self, audio=True):
        '''
        Play auditorly the current stimulus using the play function defined in
        the abstract class

        Return:
            current_stimulus: string (the stimulus that is presented)
        '''
        if audio:
            stim_path = self.audio_path+os.sep+self.current_stimulus
            if self.current_stimulus[:3] == self.prefix_yes:
                self._play_audio(stim_path,self.trigger_yes[0],self.trigger_yes[1])
            else:
                self._play_audio(stim_path,self.trigger_no[0],self.trigger_no[1])
        log.debug('stimulus end')
        return self.current_stimulus

    def get_current_state(self):
        '''
        Return the current state

        Return:
            quest_type: str (If yes or no question)
            current_stimulus: str (current question)

        '''
        if self.current_stimulus[:3] == self.prefix_yes:
            quest_type = 'Yes question'
            self.current_response = 'yes'
        else:
            quest_type = 'No question'
            self.current_response = 'no'
        return quest_type, self.current_stimulus
    
    def get_mode(self):
        '''
        Return the mode of the used paradigm
        
        Return:
            mode: string (mode of the paradigm)
        
        '''
        return self.mode
    
    def save_log(self):
        '''
        Save the log
        
        '''
        pass
